(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

module B = Objects

let state = function
    Mast.QUEUE(cls,qo,nm,vis,shr,a) -> Ast.QUEUE(cls,qo,nm,vis,a)
  | Mast.PROCESS(cls,nm,_,vis,shr,a) -> Ast.PROCESS(cls,nm,Ast.NO_PREV,vis,a)

let uop = function
    Mast.NOT -> Ast.NOT
  | Mast.COMPLEMENT -> Ast.COMPLEMENT

let bop = function
    Mast.PLUS -> Ast.PLUS
  | Mast.MINUS -> Ast.MINUS
  | Mast.TIMES -> Ast.TIMES
  | Mast.DIV -> Ast.DIV
  | Mast.MOD -> Ast.MOD
  | Mast.AND -> Ast.AND
  | Mast.OR -> Ast.OR
  | Mast.BITAND -> Ast.BITAND
  | Mast.BITOR -> Ast.BITOR
  | Mast.EQ -> Ast.EQ
  | Mast.NEQ -> Ast.NEQ
  | Mast.LT -> Ast.LT
  | Mast.GT -> Ast.GT
  | Mast.LEQ -> Ast.LEQ
  | Mast.GEQ -> Ast.GEQ
  | Mast.LSHIFT -> Ast.LSHIFT
  | Mast.RSHIFT -> Ast.RSHIFT

let rec expr = function
    Mast.INT(n,a) -> Ast.INT(n,a)
  | Mast.VAR(i,a) -> Ast.VAR(i,a)
  | Mast.FIELD(e,i,a) -> Ast.FIELD(expr e,i,a)
  | Mast.BOOL(b,a) -> Ast.BOOL(b,a)
  | Mast.UNARY(u,e,a) -> Ast.UNARY(uop u,expr e,a)
  | Mast.BINARY(b,e1,e2,a) -> Ast.BINARY(bop b,expr e1,expr e2,a)
  | Mast.INDR(e,a) -> Ast.INDR(expr e,a)
  | Mast.SELECT(a) -> Ast.SELECT(None,a)
  | Mast.EMPTY(id,a) -> Ast.EMPTY(id,[],B.UNK,true,a)
  | Mast.PRIM(f,args,a) -> Ast.PRIM(expr f,List.map expr args,a)
  | Mast.SCHEDCHILD(e,a) -> Ast.SCHEDCHILD(expr e,[],a)
  | Mast.SRCONSCHED(a) -> Ast.SRCONSCHED(a)
  | Mast.IN(e,i,a) -> Ast.IN(expr e,i,[],B.UNK,true,a)
  | _ -> raise (Error.Error "mast2ast.ml, expr: unexpected expression")

let is_const = function
    Mast.CONST -> Ast.CONST
  | Mast.VARIABLE-> Ast.VARIABLE
  | Mast.DEFCONST -> Ast.DEFCONST

let vardecl (Mast.VARDECL(_,t,n,b,a)) = Ast.VARDECL(t,n,b,a)

let valdef = function
    Mast.VALDEF(decl,e,isc,a) -> Ast.VALDEF(vardecl decl,expr e,is_const isc,a)
  | Mast.SYSDEF(decl,isc,a) -> Ast.SYSDEF(vardecl decl,is_const isc,a)
  | Mast.DUMMYDEF(i,e,a) -> Ast.DUMMYDEF(i,expr e,a)

let typedef = function
    Mast.ENUMDEF(nm,ids,a) -> Ast.ENUMDEF(nm,ids,a)
  | Mast.RANGEDEF(nm,first,last,a) -> Ast.RANGEDEF(nm,expr first,expr last,a)

let extfun (Mast.FUNDECL(ty,id,arg_tys,a)) = Ast.FUNDECL(ty,id,arg_tys,a)

let direction = function
    Mast.INC -> Ast.INC
  | Mast.DEC -> Ast.DEC
  | Mast.RANDOM -> Ast.RANDOM

let state_end = function
    Mast.HEAD -> Ast.HEAD
  | Mast.TAIL -> Ast.TAIL
  | Mast.EITHER -> Ast.EITHER

let rec stmt = function
    Mast.IF(tst,thn,els,a) -> Ast.IF(expr tst,stmt thn,stmt els,a)
  | Mast.NEXT(a) ->
      raise (Error.Error "mast2ast.ml, stmt: unexpected super")
  | Mast.FOR(p,l,dir,body,a) -> Ast.FOR(p,l,[],direction dir,stmt body,a)
  | Mast.SWITCH(tst,sc,None,a) ->
      Ast.SWITCH(expr tst,List.map seq_case sc,None,a)
  | Mast.SWITCH(tst,sc,Some def,a) ->
      Ast.SWITCH(expr tst,List.map seq_case sc,Some (stmt def),a)
  | Mast.SEQ(defs,stms,a) -> Ast.SEQ(List.map valdef defs,List.map stmt stms,a)
  | Mast.RETURN(None,a) -> Ast.RETURN(None,a)
  | Mast.RETURN(Some e,a) -> Ast.RETURN(Some (expr e),a)
  | Mast.MOVE(src,dst,a,b,attr) ->
      Ast.MOVE(expr src,dst,None,[],None,a,state_end b,attr)
  | Mast.MOVEFWD(src,y,a) -> Ast.MOVEFWD(expr src,[],state_end y,a)
  | Mast.ASSIGN(l,r,a) -> Ast.ASSIGN(expr l,expr r,false,a)
  | Mast.DEFER(a) -> Ast.DEFER(a)
  | Mast.PRIMSTMT(fn,args,a) -> Ast.PRIMSTMT(expr fn, List.map expr args, a)
  | Mast.ERROR(s,a) -> Ast.ERROR(s,a)
  | Mast.BREAK(a) -> Ast.BREAK(a)

and seq_case (Mast.SEQ_CASE(ids,sis,stm,a)) = Ast.SEQ_CASE(ids,sis,stmt stm,a)

let event_name (Mast.EVENT_NAME(idl,b,a)) = Ast.EVENT_NAME(idl,b,a)

let event_decl (Mast.EVENT(nm,specnms,param,overrideable,stm,a)) =
  Ast.EVENT(event_name nm,List.map event_name specnms,vardecl param,stmt stm,a)

let function_decl (Mast.FUNDEF(ty,id,params,body,inl,a)) =
  let inl =
    match inl with
      Mast.INLINE -> Ast.INLINE
    | Mast.NO_INLINE -> Ast.NO_INLINE in
  (Ast.FUNDEF(ty,id,List.map vardecl params,stmt body,inl,a))

let key = function
    Mast.KEY(low,high) -> Ast.KEY(low,high)
  | Mast.NOKEY -> Ast.NOKEY

let order = function
    Mast.HIGHEST -> Ast.HIGHEST
  | Mast.LOWEST -> Ast.LOWEST

let oc = function
    Mast.CRIT_ID(ky,ordr,modl,var,attr) ->
      Ast.CRIT_ID(key ky,order ordr,var,attr)
  | Mast.CRIT_TEST(ky,ordr,tst,thn,els,attr) ->
      Ast.CRIT_TEST(key ky,order ordr,expr tst,expr thn,expr els,attr)
  | Mast.IMPORT(_,ln) ->
      raise
	(Error.Error
	   (Printf.sprintf "%s: unexpected ordering criteria import"
	      (B.loc2c ln)))

let pos = function
    Mast.ON -> Ast.ON
  | Mast.BEFORE -> Ast.BEFORE
  | Mast.AFTER -> Ast.AFTER

let aspect = function
    Mast.ASPECT(srcs,dsts,param,ps,body,a) ->
      Ast.ASPECT(srcs,[],dsts,[],vardecl param,pos ps,stmt body,[],a)

let admit = function
    Mast.NO_ADMIT(_) -> None
  | Mast.IMPORT_ADMIT(mn,a) ->
      raise (Error.Error "unexpected import admit")
  | Mast.ADMIT(locals,crit_params,crits,att,det,a) ->
      let attach_detach = function
	  None -> None
	| Some (param,code) -> Some(vardecl param, stmt code) in
      Some(Ast.ADMIT(List.map valdef locals,List.map vardecl crit_params,
		     List.map
		       (function crit ->
			 Ast.RETURN(Some crit,Ast.get_exp_attr crit))
		       (List.map expr crits),
		     attach_detach att,attach_detach det,a))
