(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* A foreach that is increasing or decreasing should only be applied to the
sorted state, and only when this state is SORTED, not partially sorted or
dynamically sorted *)

module B = Objects
module CS = Class_state

let error (loc,str) =
  (Error.update_error();
   match loc with
     Some lstr -> Printf.printf "%s: %s\n" lstr str
   | None -> Printf.printf "%s\n" str)

let rec do_check dir_allowed = function
    Ast.IF(tst,thn,els,a) -> do_check dir_allowed thn; do_check dir_allowed els
  | Ast.FOR_WRAPPER(label,stmts,attr) ->
      List.iter (do_check dir_allowed) stmts
  | Ast.FOR(id,state_id,_,Ast.RANDOM,stmt,attr) -> do_check dir_allowed stmt
  | Ast.FOR(id,state_id,states,_,stmt,attr) ->
      if dir_allowed
      then
	if List.for_all
	    (function state_info -> Some state_info = !CS.selected)
	    states
	then do_check dir_allowed stmt
	else error (Some(B.loc2c attr), "ordered for must be on sorted state")
      else
	error
	  (Some(B.loc2c attr),
	   "ordered for requires statically sorted select queue")
  | Ast.SWITCH(exp,cases,default,attr) ->
      List.iter
	(function Ast.SEQ_CASE(pat,_,stmt,attr) -> do_check dir_allowed stmt)
	cases;
      (match default with None -> () | Some stmt -> do_check dir_allowed stmt)
  | Ast.SEQ(decls,stmts,attr) -> List.iter (do_check dir_allowed) stmts
  | Ast.SAFEMOVEFWD(exp,_,stm,state_end,attr) -> do_check dir_allowed stm
  | _ -> ()

(* --------------------- Handlers and functions --------------------- *)

let process_handlers dir_allowed handlers =
  List.iter
    (function (Ast.EVENT(nm,specd,param,stmt,attr)) ->
      do_check dir_allowed stmt)
    handlers

let process_ifunctions dir_allowed ifunctions =
  List.iter
    (function Ast.FUNDEF(ret,nm,params,stmt,inl,attr) ->
      do_check dir_allowed stmt)
    ifunctions

let process_aspects dir_allowed = function
    Ast.ASPECTINFO(aspects) ->
      List.iter
	(function (transition,(bef,aft)) ->
	  List.iter
	    (function Ast.ASPECT(_,_,_,_,_,_,body,_,_) ->
	      do_check dir_allowed body)
	    (bef@aft))
	aspects
  | _ -> ()

let check_foreach
    ((dynamic_sort,crit_info,
     (Ast.SCHEDULER(nm,default,cstdefs,enums,procdecls,fundecls,valdefs,states,
		    criteria,admit,trace,handlers,ifunctions,functions,
		      aspects,attr))) as x)=
  let dir_allowed = dynamic_sort = Compile_select.SORTED in
  process_handlers dir_allowed handlers;
  process_ifunctions dir_allowed ifunctions;
  process_aspects dir_allowed aspects;
  x
