(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: compile_misc.ml,v 1.16 2004/11/11 16:15:00 julia Exp $ *)

(* various simple compilation things:
   - Convert class references to state references
   - Compile in expressions to state_of(expr) == states, where states is
     the | of constants
   - Compile p1 > p2 etc to a call to greater
   - Call to empty converted to a call to empty_proc or empty_queue
   - Drop the nowhere state
   - Convert references to system process fields to ->attr->field_name
   - adjust switches where the test expression is a process/schedule
   - convert next expression to appropriate field access
   - add reference count information to new, end, attach, and detach
*)

module B = Objects
module CS = Class_state

(* -------------------- Miscellaneous operations -------------------- *)

let rec lookup_id x = function
    [] -> raise (Error.Error "internal error 1")
  | ((y,v)::r) -> if B.ideq(x,y) then v else lookup_id x r

let error attr str =
  raise(Error.Error(Printf.sprintf "%s: %s" (B.loc2c attr) str))

(* ------------------------- State constants ------------------------ *)

let make_state_cst id =
  B.mkId(Printf.sprintf "%s_STATE" (String.uppercase (B.id2c id)))

let make_decl id n attr =
  let attr = B.updty attr B.INT in
  Ast.VALDEF(Ast.VARDECL(B.EMPTY,id,false,attr),
	     Ast.BINARY(Ast.LSHIFT,Ast.INT(1,attr),Ast.INT(n,attr),attr),
	     Ast.DEFCONST,attr)

(* needed because a process in an unknown state might be allowed to be
in NOWHERE *)
let make_zdecl id n attr =
  let attr = B.updty attr B.INT in
  Ast.VALDEF(Ast.VARDECL(B.EMPTY,id,false,attr),Ast.INT(0,attr),Ast.DEFCONST,
	     attr)

let rec make_state_csts n = function
    [] -> []
  | Ast.QUEUE(CS.NOWHERE,queue_typ,id,vis,attr)::xs ->
      (make_zdecl (make_state_cst id) n attr) :: (make_state_csts n xs)
  | Ast.PROCESS(CS.NOWHERE,id,previd,vis,attr)::xs ->
      (make_zdecl (make_state_cst id) n attr) :: (make_state_csts n xs)
  | Ast.QUEUE(clsname,queue_typ,id,vis,attr)::xs ->
      (make_decl (make_state_cst id) n attr) :: (make_state_csts (n+1) xs)
  | Ast.PROCESS(clsname,id,previd,vis,attr)::xs ->
      (make_decl (make_state_cst id) n attr) :: (make_state_csts (n+1) xs)

let drop_nowhere states =
  List.fold_left
    (function prev ->
      function
	  Ast.QUEUE(CS.NOWHERE,queue_typ,id,vis,attr) -> prev
	| Ast.PROCESS(CS.NOWHERE,id,previd,vis,attr) -> prev
	| state -> state::prev)
    [] states

(* --------------------------- Expressions -------------------------- *)

let choose_gtr_prim states1 states2 =
  let common_states = Aux.intersect states1 states2 in
  if List.exists
      (function
	  CS.STATE(_,_,CS.QUEUE(CS.FIFO),_)
	| CS.STATE(_,_,CS.QUEUE(CS.SELECT(CS.LIFODEFAULT)),_) 
	| CS.STATE(_,_,CS.QUEUE(CS.SELECT(CS.FIFODEFAULT)),_) -> true
	| _ -> false)
      common_states
  then "greater_lifo_fifo"
  else "greater"

let mk_greater exp1 states1 exp2 states2 attr =
  let battr = B.updty attr B.BOOL in
  let vattr = B.updty attr B.VOID in
  if (!B.hls)
  then
    let iattr = B.updty attr B.SCHEDULER_INST in
    let sched_inst = Ast.VAR(B.mkId("scheduler_instance"),iattr) in
    Ast.PRIM(Ast.VAR(B.mkId("greater"),vattr),[sched_inst;exp1;exp2],battr)
  else
    (* check if both can be in the same queue and this queue can be fifo or
       lifo *)
    let prim = choose_gtr_prim states1 states2 in
    Ast.PRIM(Ast.VAR(B.mkId(prim),vattr),[exp1;exp2],battr)

let mk_geq exp1 states1 exp2 states2 attr =
  let battr = B.updty attr B.BOOL in
  let vattr = B.updty attr B.VOID in
  if (!B.hls)
  then
    let iattr = B.updty attr B.SCHEDULER_INST in
    let sched_inst = Ast.VAR(B.mkId("scheduler_instance"),iattr) in
    Ast.UNARY(Ast.NOT,Ast.PRIM(Ast.VAR(B.mkId("greater"),vattr),
			       [sched_inst;exp2;exp1],
			       attr),
	      battr)
  else
    let prim = choose_gtr_prim states1 states2 in
    Ast.UNARY(Ast.NOT,Ast.PRIM(Ast.VAR(B.mkId(prim),vattr),[exp2;exp1],
			       attr),
	      battr)

let rec orify op attr = function
    [] -> Printf.printf "%s: " (B.loc2c attr);
	raise (Error.Error "internal error 2")
  | [x] -> x
  | x::xs -> Ast.BINARY(op,x,orify op attr xs,attr)

let sysdecls = ref ([] : B.identifier list)

let collect_sys_fields procdecls =
  List.fold_left
    (function prev ->
      function
	  Ast.VARDECL(_,id,true,_) -> id :: prev
	| _ -> prev)
    [] procdecls

let rec compile_exp = function
    Ast.FIELD(exp,fld,attr) ->
      if (Aux.member fld (!sysdecls))
      then
	Ast.FIELD(Ast.PRIM(Ast.VAR(B.mkId "attr",attr),[compile_exp exp],attr),
		  fld,attr)
      else Ast.FIELD(compile_exp exp,fld,attr)

  | Ast.UNARY(uop,exp,attr) -> Ast.UNARY(uop,compile_exp exp,attr)

  | Ast.BINARY(bop,exp1,exp2,attr) ->
      let exp1 = compile_exp exp1
      and exp2 = compile_exp exp2 in
      if (B.ty(Ast.get_exp_attr exp1)) = (B.sched_select B.PROCESS B.SCHEDULER)
	  && List.mem bop [Ast.LT;Ast.GT;Ast.LEQ;Ast.GEQ]
      then error attr "unexpected binary operator - expected pbinary"
      else Ast.BINARY(bop,exp1,exp2,attr)

  | Ast.PBINARY(bop,exp1,states1,exp2,states2,attr) ->
      let exp1 = compile_exp exp1
      and exp2 = compile_exp exp2 in
      if (B.ty(Ast.get_exp_attr exp1)) = (B.sched_select B.PROCESS B.SCHEDULER)
      then
	(match bop with
	  Ast.LT -> mk_greater exp2 states2 exp1 states1 attr
	| Ast.GT -> mk_greater exp1 states1 exp2 states2 attr
	| Ast.LEQ -> mk_geq exp2 states2 exp1 states1 attr
	| Ast.GEQ -> mk_geq exp1 states1 exp2 states2 attr
	| _ -> error attr "unexpected binary operator - expected <,>,>=,>=")
      else
	error attr
	  "unexpected pbinary operator - expected proc/sched comparison"

    | Ast.INDR(exp,attr) -> Ast.INDR(compile_exp exp,attr)
	  
    | Ast.EMPTY(id,states,x,y,attr) ->
	let mk_fname str = Ast.VAR(B.mkId str, B.updty attr B.VOID) in
	orify Ast.AND attr
	  (List.map
	     (function
		 CS.STATE(id,_,CS.PROC,_) ->
		   Ast.PRIM(mk_fname "empty_proc",[Ast.VAR(id,attr)],attr)
	       | CS.STATE(id,_,CS.QUEUE(CS.SELECT(_)),_) ->
		   let sz = !CS.selected_size in
		   if sz > 1
		   then Ast.PRIM(mk_fname "empty_array",[],attr)
		   else Ast.PRIM(mk_fname "empty_queue",[Ast.VAR(id,attr)],
				 attr)
	       | CS.STATE(id,_,CS.QUEUE(_),_) ->
		   Ast.PRIM(mk_fname "empty_queue",[Ast.VAR(id,attr)],attr))
	     states)

    | Ast.PRIM(f,args,attr) -> Ast.PRIM(f,List.map compile_exp args,attr)

    | Ast.SCHEDCHILD(arg,procs,attr) ->
	Ast.SCHEDCHILD(compile_exp arg,procs,attr)

    | Ast.IN(_,id,[states],_,_,attr) as exp -> exp

    | Ast.IN(exp,id,states,_,_,attr) ->
	let mk_fname str = Ast.VAR(B.mkId str, B.updty attr B.VOID) in
	Ast.BINARY(Ast.BITAND,
		   Ast.PRIM(mk_fname "state_of",[exp],attr),
		   orify Ast.BITOR attr
		     (List.map
			(function CS.STATE(id,_,_,_) ->
			  Ast.VAR(make_state_cst id,B.updty attr B.INT))
			states),
		   attr)

    | exp -> exp

(* -------------------------- Declarations ------------------------- *)

let compile_decl = function
    Ast.VALDEF(decl,exp,isconst,attr) ->
      Ast.VALDEF(decl,compile_exp exp,isconst,attr)

  | decl -> decl

(* --------------------------- Statements -------------------------- *)

let state_id (CS.STATE(id,_,_,_)) = id

let rec compile_stmt = function
    Ast.IF(exp,stmt1,stmt2,attr) ->
      Ast.IF(compile_exp exp,compile_stmt stmt1,compile_stmt stmt2,attr)
	
    | Ast.FOR(id,state_id,x,dir,stmt,attr) ->
	Ast.FOR(id,state_id,x,dir,compile_stmt stmt,attr)
	  
    | Ast.FOR_WRAPPER(label,stmts,attr) ->
	Ast.FOR_WRAPPER(label,List.map compile_stmt stmts,attr)
	  
    | Ast.SWITCH(exp,cases,default,attr) ->
	let expty = B.ty (Ast.get_exp_attr exp) in
	let is_proc_sched =
	  (match expty with
	    B.PROCESS -> (B.check_proc "process reference"; true)
	  | B.SCHEDULER -> (B.check_virt "scheduler reference"; true)
	  | _ -> false) in
	let int_attr = B.updty attr B.INT in
	if is_proc_sched
	then
	  let mk_fname str = Ast.VAR(B.mkId str, B.updty attr B.VOID) in
	  Ast.SWITCH(Ast.PRIM(mk_fname "state_of",[compile_exp exp],
			      B.updty attr expty),
		     List.map
		       (function Ast.SEQ_CASE(idlist,x,stmt,y) ->
			 Ast.SEQ_CASE(List.map state_id x,x,compile_stmt stmt,
				      y))
		       cases,Aux.app_option compile_stmt default,attr)
	else Ast.SWITCH(exp,List.map
			  (function Ast.SEQ_CASE(idlist,x,stmt,y) ->
			    Ast.SEQ_CASE(idlist,x,compile_stmt stmt,y))
			  cases,Aux.app_option compile_stmt default,attr)
    | Ast.SEQ(decls,stmts,attr) ->
	Ast.SEQ(List.map compile_decl decls,List.map compile_stmt stmts,attr)
	  
    | Ast.RETURN(None,attr) -> Ast.RETURN(None,attr)
	  
    | Ast.RETURN(Some(exp),attr) -> Ast.RETURN(Some(compile_exp exp),attr)
	  
    | Ast.MOVE(exp,state,x,y,z,a,b,attr) ->
	Ast.MOVE(compile_exp exp,state,x,y,z,a,b,attr)
	       
    | Ast.MOVEFWD(exp,x,state_end,attr) ->
	Ast.MOVEFWD(compile_exp exp,x,state_end,attr)

    | Ast.SAFEMOVEFWD(exp,x,stm,state_end,attr) ->
	raise
	  (Error.Error
	     "compile_misc: unexpected safe move forward statement\n")
	       
    | Ast.ASSIGN(expl,expr,sorted_fld,attr) ->
	Ast.ASSIGN(compile_exp expl,compile_exp expr,sorted_fld,attr)

    | Ast.PRIMSTMT(f,args,attr) ->
	Ast.PRIMSTMT(f,List.map compile_exp args,attr)

    | Ast.ASSERT(exp,attr) -> Ast.ASSERT(compile_exp exp,attr)

    | x -> x

(* ------------------- Check for the initial event ------------------ *)

let check_initial attr nm stmt =
  if List.exists (Ast.event_name_equal nm) (!Events.init_events)
  then
    let new_id = B.fresh_id ()
    and battr = B.updty attr B.BOOL in
    let set = Ast.ASSIGN(Ast.VAR(new_id,battr),
			 Ast.BOOL(true,battr),false,attr) in
    Some(Ast.VALDEF(Ast.VARDECL(B.BOOL,new_id,false,battr),
		    Ast.BOOL(false,battr),
		    Ast.VARIABLE,battr),
	 Ast.SEQ([],
	   [Ast.IF(Ast.VAR(new_id,battr),
		   Ast.SEQ([],
			   [Ast.ERROR("initial event called more than once",
				      attr)],
			   attr),
		   Ast.addBLOCK attr set stmt,
		   attr)],
		 attr))
  else None

(* ------------------------- Reference count ------------------------ *)

let call s attr = Ast.PRIMSTMT(Ast.VAR(B.mkId(s),attr),[],attr)

(* only works if there are separate handlers for new and end, but the event
types rather require this... *)
let check_ref_ct_hd nm stmt attr =
  match !B.schedtype with
    B.VIRTSCHED -> stmt
  | B.PROCSCHED ->
      (match Ast.event_name2c2 nm with
	Some "process.new" ->
	  Ast.addBLOCK attr (call "increment_ref_count" attr) stmt
      | Some "process.end" ->
	  Ast.addBLOCK attr (call "decrement_ref_count" attr) stmt
      | _ -> stmt)

let check_ref_ct_fn nm stmt attr =
  match B.id2c nm with
    "attach" -> stmt (* taken care of in pre_type.ml *)
  | "detach" ->
      Ast.addBLOCK attr (call "decrement_ref_count" attr) stmt
  | _ -> stmt

(* --------------------- Handlers and functions --------------------- *)

(* parser should check that event names are ok *)
let compile_handlers handlers =
  List.map
    (function (Ast.EVENT(nm,specd,param,stmt,attr)) ->
      Ast.EVENT(nm,specd,param,check_ref_ct_hd nm (compile_stmt stmt) attr,
		attr))
    handlers

let compile_ifunctions ifunctions =
  List.map
    (function Ast.FUNDEF(ret,nm,params,stmt,inl,attr) ->
      Ast.FUNDEF(ret,nm,params, check_ref_ct_fn nm (compile_stmt stmt) attr,
		 inl,attr))
    ifunctions

(* -------------------------- entry point --------------------------- *)

let compile_misc
    (Ast.SCHEDULER(nm,default,
		   cstdefs,enums,procdecls,fundecls,valdefs,states,criteria,
		   admit,trace,handlers,ifunctions,functions,aspects,attr)) =
  let state_csts = make_state_csts 0 states in
  sysdecls := collect_sys_fields procdecls;
  let newhandlers = compile_handlers handlers in
  Ast.SCHEDULER(nm,default,cstdefs@state_csts,enums,procdecls,fundecls,
		valdefs,
		List.rev (drop_nowhere states),
		criteria,admit,trace,newhandlers,
		compile_ifunctions ifunctions, compile_ifunctions functions,
		aspects,attr)
