{
(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

                                                                      
(* $Id: event_lexer.mll,v 1.3 2004/07/06 15:21:48 julia Exp $ *)
open Event_parser
open Format
open Lexing

type error =
  | Illegal_character of char
  | Unterminated_comment

(* Error defines error types and positions*)
exception Error of error * int * int

(* To store the position of the beginning of a comment and bitpat *)
let comment_start_pos = ref 0

(* Report Error  *)

let report_error fmt = function
  | Illegal_character c  -> fprintf fmt "Illegal character (%s)" (Char.escaped c)
  | Unterminated_comment -> fprintf fmt "Comment not terminated"


}

(* define rules *)

let id = ['a'-'z' 'A'-'Z'] ['a'-'z' 'A'-'Z' '0'-'9' '_']*
let decimal_literal = ['0'-'9']+

rule token = parse
    	|   	"["             { LBRACK        }
    	|   	"]"             { RBRACK        }
    	|   	"("             { LPAREN        }
    	|   	")"             { RPAREN        }
    	|   	"{"             { LBRACE        }
    	|   	"}"             { RBRACE        }
	|	","		{ COMMA		}
	|	"::"		{ COLONCOLON	}
	|	"="		{ EQUAL		}
	|	"in"		{ IN      	}
	|	"src"		{ SRC  		}
	|	"tgt"		{ TGT  		}
	|	"p"		{ P    		}
	|	"q"		{ Q    		}
	|	"x"		{ X    		}
	|	"arg"		{ ARG  		}
	|	"-"     	{ HYPHEN	}
	|	"!"     	{ BANG		}
	|	"application_events"  { APPEV	}
	|	"initial_events"      { INITEV	}
	|	"system_events"       { SYSEV	}
	|	"top_level_events"    { TOPLEV	}
	|	"scheduler_events"    { SCHEDLEV}
	|	"event_types"         { TYPES 	}
	|	"quasi_event_types"   { QUASITYPES }
	|	"vs_interface"        { VSINTERFACE }
	|	"automaton"           { AUTOMATON  }
	|	"init_automaton"      { IAUTOMATON }
	|	"final"               { FINAL      }
	|	"interface"     { INTERFACE	}
	|	":"		{ COLON		}
	|	";"		{ SEMIC		}
	|	"->"		{ RIGHTARROW	}
	|	"-u>"		{ UNINTARROW	}
	|	"."		{ DOT       	}
	|	"*"		{ TIMES     	}
    	|   eof                 { EOF           }
    	|   [' ' '\t' ]
        	{ token lexbuf }    (* skip whitespace      *)
    	|   '\n' | '\r' | "\r\n"
		{ 
	  	Misc.lineno := !Misc.lineno+1;
          	Misc.cr_index := (lexeme_start lexbuf) + 1;
	  	Misc.incr lexbuf;
          	token lexbuf  (* skip whitespace *)
		}    
    	|   "//" [^'\n' '\r']*       
		{ token lexbuf }    (* skip // comment      *)
    	|   "/*"
        	{ 
          	comment_start_pos := Lexing.lexeme_start lexbuf;
          	comment lexbuf; 
	  	token lexbuf 
		}
    	| decimal_literal
		{ INTEGER(int_of_string(lexeme lexbuf)) }
    	| id                 
		{ IDENT (lexeme lexbuf) }
    	|   _
        { 
	  raise (Error(Illegal_character ((Lexing.lexeme lexbuf).[0]),
		       Lexing.lexeme_start lexbuf,
		       Lexing.lexeme_end lexbuf))
	}

and comment = parse
    | "*/"                  
	{ () } 
    | '\n' | '\r' | "\r\n"
        { 
	  Misc.lineno := !Misc.lineno+1; 
          Misc.cr_index := (lexeme_start lexbuf) + 1;
	  Misc.incr lexbuf;
	  comment lexbuf 
	}
    | eof
        { 
          raise (Error (Unterminated_comment, 
			!comment_start_pos,
			!comment_start_pos + 2))
        }
    | _  
        { comment lexbuf }











