(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: ast2c.ml,v 1.14 2004/11/11 16:15:00 julia Exp $ *)
module L = List

module B = Objects
module CS = Class_state
module G = Genaux

open Format
open Ioctl

type phase = HANDLERS | INTERFACE | INTERFACE_LABEL of string | ATTACH | DETACH
let current_phase = ref HANDLERS

let event_param = ref(B.mkId "")

let break_labels = ref ([] : B.identifier option list)

(* --------------------------- Expressions -------------------------- *)

let list_of_const_val_enum =
  ref ["SchedState.NO_STATE"; "SchedState.RUNNING"; "SchedState.READY";
	"SchedState.BLOCKED"; "SchedState.TERMINATED"; "sched_type.PS";
	"sched_type.VS"]

let rec is_const_val_enum nm_val l =
	match l with
   	[] -> false
 	| t::q -> (nm_val = t) || is_const_val_enum nm_val q

let rec pp_ucexp = function
    Ast.INT(n,attr) -> print_int n
    | Ast.VAR(id,attr) ->
    	(match (B.ty attr) with
	B.ENUM(name) ->
	  if ((!B.csharp) &&
	      (is_const_val_enum ((B.id2c(name))^"."^(B.id2c(id)))
		 !list_of_const_val_enum))
	  then (G.pp_id name; print_string "."; G.pp_id id;)
	  else G.pp_id id
	| _ -> G.pp_id id)
    | Ast.FIELD(Ast.VAR(id,_) as exp,fld,_) as x ->
	let not_target _ =
	  if (G.is_struct false id)
	  then (G.pp_struct_id2 id; print_string "."; G.pp_id fld)
	  else (G.pp_struct_id2 id; print_string "->"; G.pp_id fld) in
	(match B.ty(Ast.get_exp_attr(exp)) with
	  B.EVENT ->
	    let str = B.id2c fld in
	    if str = "target"
	    then print_string "tgt"
	    else if str = "type"
	    then print_string "event"
	    else not_target()
	| B.TIME -> G.pp_struct_id2 id; print_string "."; G.pp_id fld
	| _ -> not_target())

    | Ast.FIELD(exp,fld,attr) ->
	(match B.ty(Ast.get_exp_attr(exp)) with
	  B.TIME -> pp_exp exp; print_string "."; G.pp_id fld
	| _ -> pp_exp exp; print_string "->"; G.pp_id fld)

    | Ast.BOOL(true,attr) -> print_string "true"

    | Ast.BOOL(false,attr) -> print_string "false"

    | Ast.UNARY(Ast.NOT,exp,attr) ->
	(print_string "!"; print_paren exp)

    | Ast.UNARY(Ast.COMPLEMENT,exp,attr) ->
	(print_string "~"; print_paren exp)

    | Ast.BINARY(bop,exp1,exp2,attr) ->
	(match bop with
	  Ast.OR | Ast.AND -> open_box 0
	| _ -> ());
	(if (!B.csharp) then
    	  (if (B.type2ctype (B.ty (Ast.get_exp_attr exp1))="SchedState") then
	    print_string "(int)";
	   print_paren exp1; pp_bop bop;
	   if (B.type2ctype (B.ty (Ast.get_exp_attr exp2))="SchedState") then
	     print_string "(int)";
	   print_paren exp2)
	else
	  (print_paren exp1; pp_bop bop; print_paren exp2));
	(match bop with
	  Ast.OR | Ast.AND -> close_box()
	| _ -> ())

    | Ast.PBINARY(bop,exp1,states1,exp2,states2,attr) ->
	raise (Error.Error "internal error: unexpected pbinary")

    | Ast.INDR(exp,attr) ->
	print_string "*"; print_string "("; pp_exp exp; print_string ")"

    | Ast.SELECT(_,attr) ->
	print_string "select()"

    | Ast.EMPTY(id,_,tv,_,attr) ->
	(print_string "empty("; G.pp_id id; print_string ")")

    | Ast.PRIM(f,args,attr) as exp ->
	let ext =
	  (match f with
	    Ast.VAR(id,_) -> Aux.member id (!G.external_functions)
	  | _ -> false (* assumed not to be external *)) in
	let mk_cast exp =
	  if ext
	  then
	    (match B.ty(Ast.get_exp_attr exp) with
	      B.PROCESS -> Some "(struct bossa_struct *)"
	    | B.SCHEDULER ->
		B.check_virt "scheduler return type";
		Some "(struct scheduler_struct *)"
	    | _ -> None)
	  else None in
	(match mk_cast exp with Some x -> print_string x | None -> ());
	let print_app _ =
	  print_paren f; print_string "(";
	  open_box 0;
	  G.print_between G.pp_comma
	    (function exp ->
	      match mk_cast exp with
	      	Some x -> print_string x ; print_paren exp
	      | None -> pp_exp exp)
	    args;
	  close_box(); print_string ")" in
	if (!B.csharp) then
	(match f with
	  Ast.VAR(id,_) ->
	  (match (B.id2c(id)) with
	    "next_exists" -> pp_ucexp (L.hd args); print_string "->sched"
	  | "empty_proc" ->
	      (print_string "((";
	       pp_ucexp (L.hd args);
	       print_string ") == null)";)
	  | "empty_queue" ->
	      (print_string "list_empty(ref ";
	       print_paren (L.hd args);
	       print_string ")";)
	  | "state_of" ->
	      (print_string "("; print_paren (L.hd args);
	       print_string ")->state ";)
	  | "attr" ->
	      print_string "((task_struct *)((char *)(";
	      print_paren (L.hd args);
	      G.print_nl_string
		")-(ulong)((int)(&((task_struct *)0)->bossa)/2)))"
	  | "select_sort" ->
	      (print_string "select_sort(ref ";
	       print_paren (L.hd args);
	       print_string ")")
	  | "select_partially_sorted" ->
	      (print_string "select_fifo_partially_sorted(&(";
	       pp_ucexp (L.hd args); G.print_nl_string("));");)
	  | _ -> print_app())
	| _ -> print_app())
	else print_app()
	    
    | Ast.SCHEDCHILD(arg,procs,attr) ->
	(print_string "next("; open_box 0; pp_exp arg; close_box();
	 print_string ")")
	  
    | Ast.SRCONSCHED(attr) ->
	open_box 0;
	print_string "src_next_list != src_next_list->next &&";
	force_newline();
	print_string "src_next_list->sched->ops == &EXPORTS";
	close_box()

    | Ast.ALIVE(_,_,attr) -> print_string "target_alive"
	  
    | Ast.AREF(exp1,exp2,_) ->
	(print_string "&";
	 pp_exp exp1; print_string "["; pp_exp exp2; print_string "]")
	  
    | Ast.IN(exp,id,[_],_,_,attr) ->
	(* used when there is only one state possible *)
	if (!B.csharp) then
	  (print_string "("; pp_exp exp; print_string ")->state == ";
	   G.pp_state_cst id)
	else
	  (print_string "state_of("; pp_exp exp; print_string ") == ";
	   G.pp_state_cst id)

    | Ast.IN(_,_,_,_,_,attr) ->
	raise
	  (Error.Error
	     (Printf.sprintf "only one state expected for IN in %s"
		(B.loc2c attr)))
	    
    | Ast.MOVEFWDEXP(exp,state_end,attr) ->
	print_string "forwardImmediate(";
	open_box 0; pp_exp exp; print_string ", ";
	G.pp_id (!event_param);
	G.pp_comma_nl();
	G.print_nl_string "src_next_list->next,";
	print_string "tgt_next_list->next";
	close_box();
	print_string ")"
	  
and pp_exp exp =
  pp_ucexp exp;
  
  match B.com (Ast.get_exp_attr exp) with
    None -> ()
  | Some(str) -> print_string " /* "; print_string str; print_string " */"
	
(* eventually, compare the precedence of the expression with the precedence
   of the context, as done for devil *)
and print_paren exp =
  match exp with
    Ast.INT(n,attr) -> pp_exp exp

    | Ast.VAR(id,attr) -> pp_exp exp

    | Ast.FIELD(exp1,fld,attr) -> pp_exp exp

    | Ast.BOOL(v,attr) -> pp_exp exp

    | Ast.UNARY(uop,uexp,attr) -> pp_exp exp

    | Ast.BINARY(bop,exp1,exp2,attr) ->
	(print_string "("; pp_exp exp; print_string ")")
 
    | Ast.PBINARY(bop,exp1,states1,exp2,states2,attr) ->
	raise (Error.Error "internal error: unexpected pbinary")

    | Ast.INDR(exp,attr) -> pp_exp exp

    | Ast.SELECT(_,attr) -> pp_exp exp

    | Ast.EMPTY(id,_,_,_,attr) -> pp_exp exp

    | Ast.PRIM(f,args,attr) -> pp_exp exp

    | Ast.SCHEDCHILD(arg,procs,attr) -> pp_exp exp

    | Ast.SRCONSCHED(attr) -> pp_exp exp

    | Ast.ALIVE(_,_,attr) -> pp_exp exp

    | Ast.AREF(exp1,exp2,attr) -> pp_exp exp

    | Ast.IN(exp1,id,_,_,_,attr) ->
	(print_string "("; pp_exp exp; print_string ")")

    | Ast.MOVEFWDEXP(exp1,state_end,attr) -> pp_exp exp

and pp_bop = function
  | Ast.AND -> print_space(); print_string "&&"; print_space()
  | Ast.OR -> print_space(); print_string "||"; print_space()
  | op -> print_string (" "^(Ast.bop2c op)^" ")

(* -------------------------- Declarations ------------------------- *)

let need_comma = ref false

let pp_static_timer_structures id =
  let snm = B.id2c id in
  G.print_nl_string ("static struct timer_list _"^snm^"_timer;");
  G.print_nl_string ("static struct bossa_timer_data _"^snm^"_timer_data;")

let pp_decl global (Ast.VARDECL(ty,id,imported,attr)) =
  need_comma := false;
  (if global
  then
    match ty with
      B.TIMER(_) -> pp_static_timer_structures id
    | _ -> ());
  if not (B.id2c id = "current") then
  (if imported
  then
    (if not (!B.csharp) then
     begin
       if global then print_string "extern ";
       print_string (B.type2c(ty)); need_comma := true;
     end)
  else
    (if global then print_string "static ";
     G.pp_type ty; need_comma := true);
   if not imported || not (!B.csharp)
   then begin print_string " "; G.pp_struct_id(id); need_comma := true end)

let pp_decl_cs global (Ast.VARDECL(ty,id,imported,attr)) =
  need_comma := false;
  if not (B.id2c id = "current") then
  if not imported
  then begin print_string "    "; G.pp_struct_id(id); need_comma := true end

let rec pp_def global = function
    Ast.VALDEF(decl,exp,isconst,attr) ->
      (open_box 2;
       if isconst=Ast.CONST then ((print_string "const"; print_space());
         pp_decl global decl; print_string " = ";
         pp_exp exp; print_string ";")
       else if isconst=Ast.DEFCONST
       then
	 if (!B.csharp)
	 then (
	   print_string "private const uint ";
	   pp_decl false decl;
	   print_string " = ";
	   pp_exp exp;
	   print_string ";")
	 else (
	   print_string "#define ";
	   pp_decl false decl;
	   print_string "  ";
	   pp_exp exp)
       else
	 (if (!B.csharp) && global
	 then print_string "private ";
	  pp_decl global decl;
	  if not (!B.csharp)
	  then (print_string " = "; pp_exp exp)
	  else
	    if not global
	    then (print_string " = "; pp_exp exp);
	  print_string ";");
       close_box();
       force_newline())
  | Ast.UNINITDEF(decl,attr) ->
      (pp_decl global decl;
       if (!need_comma) then print_string ";"; force_newline())
  | Ast.SYSDEF(decl,isconst,attr) ->
      (pp_decl global decl;
       if (!need_comma) then print_string ";"; force_newline())
  | Ast.DUMMYDEF(id,exp,attr) ->
      (print_string "???"; print_space(); G.pp_id(id);
       print_string " = "; pp_exp exp; print_string ";";
       force_newline())

let rec pp_def_cs global = function
    Ast.VALDEF(decl,exp,isconst,attr) ->
      (open_box 2;
       pp_decl_cs global decl;
       print_string " = ";
       pp_exp exp; print_string ";";
       close_box();force_newline())
  | Ast.UNINITDEF(decl,attr) ->
      (pp_decl_cs global decl;
       if (!need_comma) then print_string ";"; force_newline())
  | Ast.SYSDEF(decl,isconst,attr) ->
      (pp_decl_cs global decl;
       if (!need_comma) then print_string ";"; force_newline())
  | Ast.DUMMYDEF(id,exp,attr) ->
      (print_string "???"; print_space(); G.pp_id(id);
       print_string " = "; pp_exp exp; print_string ";";
       force_newline())

let pp_defs global defs =
  List.iter (pp_def global) defs;
  G.nl_ifnonempty(defs)

(* --------------------------- Statements -------------------------- *)

let print_casename idlist =
  List.iter
    (function id ->
      print_string "case "; G.pp_state_cst id; G.print_nl_string ":")
    idlist

let print_caseid idlist =
  List.iter
    (function id ->
      print_string "case "; G.pp_id id; G.print_nl_string ":")
    idlist


let rec pp_stmt = function
   Ast.IF(exp,stmt1,stmt2,attr) ->
      ((match exp with
       	Ast.BINARY(Ast.BITAND,_,_,_) -> (print_string "if (("; pp_exp exp; print_string ")!=0")
	| _ -> (print_string "if ("; pp_exp exp;));
       print_string ") "; pp_seq(stmt1);
       match stmt2 with
	 Ast.SEQ([],[],_) -> ()
       | _ -> (force_newline();
	       print_string "else "; pp_seq(stmt2)))

(*------foreach -----*)
    | Ast.FOR(id,state_id,states,dir,stmt,attr) ->
	pp_for id (function _ -> pp_seq stmt)
	  (function ptr -> pp_seq3 stmt ptr id)
	  (move_possible stmt) states dir

    | Ast.FOR_WRAPPER(label,[Ast.FOR(_,_,_,_,_,_) as s],a) ->
	Aux.push None break_labels;
	pp_stmt s;
	Aux.pop break_labels

    | Ast.FOR_WRAPPER(label,stms,a) ->
	Aux.push (Some label) break_labels;
	G.print_between force_newline pp_stmt stms;
	Aux.pop break_labels;
	force_newline();
	G.pp_id label; print_string ":;"

    | Ast.SWITCH(exp,cases,default,attr) ->
	let is_proc_sched =
	  (match B.ty (Ast.get_exp_attr exp) with
	    B.PROCESS -> (B.check_proc "process reference"; true)
	  | B.SCHEDULER -> (B.check_virt "scheduler reference"; true)
	  | _ -> false) in
	(print_string "switch ("; pp_exp exp;
	 print_string ") "; pp_cases cases default is_proc_sched)

    | (Ast.SEQ(decls,stmts,attr) as x) -> pp_seq x

    | Ast.RETURN(None,attr) ->
	(match !current_phase with
	  HANDLERS -> print_string "return;"
	| INTERFACE ->
	    if (!B.csharp)
	    then print_string ("return ("^(!B.type_return)^")SUCCESS;")
	    else print_string ("return SUCCESS;")
	| INTERFACE_LABEL(l) ->
	    print_string "goto "; print_string l; print_string ";"
	| ATTACH | DETACH ->
	    if (!B.csharp)
	    then print_string ("return ("^(!B.type_return)^")SUCCESS;")
	    else print_string ("return SUCCESS;")
	)
    | Ast.RETURN(Some(exp),attr) ->
	(open_box 2; print_string "return ";
	 if (!B.csharp) then print_string ("("^(!B.type_return)^")(");
	 pp_exp exp;
	 if (!B.csharp) then print_string ")"; print_string ";"; close_box())

    | Ast.MOVE(exp,state,_,_,dst,_,_,attr) ->
	(* only used in blast mode *)
	(match dst with
	  None -> raise (Error.Error "destination not known")
	| Some(CS.STATE(_,_,CS.PROC,_)) -> print_string "move_proc("
	| Some(CS.STATE(_,CS.TERMINATED,_,_)) -> print_string "move_nowhere("
	| Some(CS.STATE(_,_,CS.QUEUE(_),_)) -> print_string "move_queue(");
	open_box 0; pp_exp exp; G.pp_comma();
	G.pp_id state; G.pp_comma(); G.pp_state_cst state; close_box();
	print_string ");"

    | Ast.MOVEFWD(exp,_,_,attr) ->
	print_string "forwardImmediate(";
	open_box 0; pp_exp exp; print_string ", "; G.pp_id (!event_param);
	G.pp_comma_nl();
	G.print_nl_string "src_next_list->next,";
	print_string "tgt_next_list->next";
	close_box();
	print_string ");"

    | Ast.SAFEMOVEFWD(exp,x,stm,state_end,attr) ->
      raise
	(Error.Error
	   "ast2c: unexpected safe move forward statement\n")

    | Ast.ASSIGN(expl,expr,sorted_fld,attr) ->
	let ltype = B.ty (Ast.get_exp_attr expl)
	and rtype = B.ty (Ast.get_exp_attr expr) in
	(open_box 2;
	(match expl with
	Ast.FIELD(Ast.VAR(id1, m),id, attr) ->
	  (if ((B.id2c id1="current") && (B.id2c id="need_resched") &&
	       (!B.csharp)) then
	     (print_string "set_current_need_resched(1);"; close_box();)
	   else
	     (pp_exp expl; print_space(); print_string "="; print_space();
	      if not (ltype = rtype)
	      then
	   	(print_string "("; G.pp_type ltype; print_string ")";
	 	 print_paren expr)
	      else pp_exp expr;
	      print_string ";"; close_box()))
	| _ -> (pp_exp expl; print_space(); print_string "="; print_space();
	 	if not (ltype = rtype)
	 	then
	   	  (print_string "("; G.pp_type ltype; print_string ")";
	    	   print_paren expr)
	 	else pp_exp expr;
	 	print_string ";"; close_box()))
	)
	  
    | Ast.DEFER(attr) -> print_string "deferEvent();"

    | Ast.BREAK(attr) ->
	(match Aux.top break_labels with
	  None -> print_string "break;"
	| Some label -> print_string "goto "; G.pp_id label; print_string ";")
	  
    | Ast.PRIMSTMT(f,args,attr) ->
    	if (!B.csharp) then
	  (match f with
	  Ast.VAR(id, m) ->
	    (match (B.id2c(id)) with
	      "update_state" ->
		(pp_ucexp (L.hd args); print_string "->state=(int)";
		 pp_ucexp (L.nth args 1);
		 print_string ";";)
	    |"move_queue_nowhere" ->
		(print_string "fixed (list_head * ptr = &(";
		 pp_ucexp (L.hd args);
		 G.print_nl_string "->list_info))";
		 print_string "\tlist_del(ptr);";)
	    |"move_proc_queue" ->
		(print_string "list_add(&("; pp_ucexp (L.hd args);
		 print_string "->list_info), ref ";
		 pp_ucexp (L.nth args 1); print_string "); ";
		 pp_ucexp (L.hd args);
		 G.print_nl_string " = null;")
	    |"move_queue_partially_sorted_lifo" ->
		( print_string "list_del(&("; pp_ucexp (L.hd args);
		  print_string "->list_info)); insert_psorted_lifo(";
		  pp_ucexp (L.hd args);
		  print_string ", &("; pp_ucexp (L.nth args 1);
		  G.print_nl_string "));";)
	    |"move_proc_partially_sorted_lifo" ->
		( print_string "insert_psorted_lifo(";
		  pp_ucexp (L.hd args); print_string ",&(";
		  pp_ucexp (L.nth args 1); print_string ")); ";
		  pp_ucexp (L.hd args); G.print_nl_string " = null;";)
	    |"move_proc_proc" ->
		( pp_ucexp (L.nth args 1); print_string " = ";
		  pp_ucexp (L.hd args);
		  print_string "; "; pp_ucexp (L.hd args);
		  G.print_nl_string " = null;";)
	    |"move_queue_queue" ->
		( print_string ((G.pp_bossa_struct_cs())^" *temp = ");
		  pp_ucexp (L.hd args);
		  print_string "; list_del(&(temp->list_info));
		    list_add(&(temp->list_info), ref ";
		  pp_ucexp (L.nth args 1); G.print_nl_string ");"; )
	    |"move_queue_proc" ->
		( print_string ((G.pp_bossa_struct_cs())^" *temp = ");
		  pp_ucexp (L.hd args); print_string ";
		    list_del(&(temp->list_info)); ";
		  pp_ucexp (L.nth args 1); G.print_nl_string " = temp;"; )
	    |"move_nowhere_partially_sorted_lifo" ->
		( print_string "insert_psorted_lifo(";
		  pp_ucexp (L.hd args); print_string ", &(";
		  pp_ucexp (L.nth args 1);
		  G.print_nl_string "));";)
	    |"move_nowhere_queue" ->
		( print_string "list_add(&("; pp_ucexp (L.hd args);
		  print_string "->list_info), ref "; pp_ucexp (L.nth args 1);
		  G.print_nl_string ");"; )
	    |"move_proc_nowhere" ->
		( pp_ucexp (L.hd args); G.print_nl_string " = null;";)
	    |_ ->
		(print_paren f; print_string "(";
		 open_box 0;
		 G.print_between G.pp_comma pp_exp args; close_box();
		 print_string ");";
		)
	    )
	  |_ -> (print_paren f; print_string "(";
	  	 open_box 0;
	  	 G.print_between G.pp_comma pp_exp args; close_box();
	  	 print_string ");"))
	else
     	  (if (!B.hls) then
	    (match (f,args) with
	      (Ast.VAR(id, m),[tgt;state;cls]) ->
		(match (B.id2c(id)) with
		  "update_state" ->
		    G.start_block();
		    print_string "struct HLS_VPROC *tgt_udvp = ";
		    pp_ucexp tgt;
		    G.print_nl_string "->udvp;";
		    (match cls with
		      Ast.VAR(id,_) ->
			(match B.id2c id with
			  "RUNNING" ->
			    G.print_nl_string "tgt_udvp->State = VP_Running;";
			    G.print_nl_string "tgt_udvp->TopData = tvp;";
			    print_string "tgt_udvp->Proc = tvp->Proc;";
			| "READY" ->
			    G.print_nl_string "tgt_udvp->State = VP_Ready;";
			    print_string "tgt_udvp->Proc = NO_PROC;";
			| "BLOCKED" ->
			    G.print_nl_string "tgt_udvp->State = VP_Waiting;";
			    print_string "tgt_udvp->Proc = NO_PROC;";
			| "TERMINATED" -> ()
			| _ ->
			    raise
			      (Error.Error "internal error: unexpected class"))
		    | _ -> raise
			  (Error.Error "internal error: unexpected class"));
		    G.end_block();
		    force_newline()
		| _ -> ())
	    | _ -> ());
	   let print_call _ =
	     print_paren f; print_string "(";
	     open_box 0; G.print_between G.pp_comma pp_exp args; close_box();
	     print_string ");" in  
	   (match f with
	     Ast.VAR(id, m) ->
	       (match (B.id2c(id),args,!B.high_res) with
		 ("time_to_ticks",[arg],true) ->
		   print_string "time_to_ticks(&"; pp_exp arg;
		   print_string ");"
	       | ("start_relative_timer",[timer_arg;time_arg],true) ->
		   print_string "start_relative_timer("; pp_exp timer_arg;
		   print_string ", &"; pp_exp time_arg; print_string ");"
	       | ("start_absolute_timer",[timer_arg;time_arg],true) ->
		   print_string "start_absolute_timer("; pp_exp timer_arg;
		   print_string ", &"; pp_exp time_arg; print_string ");"
	       | _ -> print_call())
	   | _ -> print_call()))
	     
    | Ast.ERROR(str,attr) ->
	let einval _ =
	  G.mk_printk (str^"\\n");
	  if (!B.csharp)
	  then G.print_nl_string "set_errno(-_EINVAL);"
	  else G.print_nl_string "errno = -EINVAL;" in
	let einval_attach _ =
	  (* only for process schedulers *)
	  G.print_nl_string ("error_string = \""^str^"\\n\";");
	  if (!B.csharp)
	  then G.print_nl_string "set_errno(-_EINVAL);"
	  else G.print_nl_string "errno = -EINVAL;" in
	(match !current_phase with
	  HANDLERS ->
	    print_string "panic(\"";
	    print_string str;
	    G.print_nl_string ": file %s line %d\",";
	    if (!B.csharp)
	    then
	      (print_string "        \""; print_string (!G.policy_name);
	       print_string ".cs\", 0);";)
	    else print_string "        __FILE__, __LINE__);"
	| INTERFACE
	| INTERFACE_LABEL(_) ->
	    einval();
	    print_string "return FAILURE;"
	| DETACH -> einval();
	    print_string "return FAILURE;"
	| ATTACH -> einval_attach();
	    print_string "return FAILURE;")

    | Ast.ASSERT(exp,_) ->
	(open_box 2; print_string "bossa_assert("; pp_exp exp;
	 print_string ");"; close_box())

(* ----------------------- Start of for loops ---------------------- *)

and pp_for_proc state_id id print_stmt =
  G.start_block();
  G.pp_bossa_struct(); print_string " *";
  G.pp_id(id); print_string " = "; G.pp_id state_id;
  G.pp_semic();
  print_string "if ("; G.pp_id id; print_string ") ";
  G.start_block();
  print_stmt(); G.end_block(); G.end_block()

and pp_for_queue state_id id print_stmt move_possible dir =
  let elem = B.fresh_id() in
  G.start_block();
  G.pp_bossa_struct(); print_string " *";
  G.pp_id(id); G.pp_semic();
  print_string (G.de_structify "struct list_head *");
  G.pp_id elem; G.pp_semic();
  if move_possible
  then
    (let ptr = B.fresh_id() in
    print_string (G.de_structify "struct list_head *");
    G.pp_id(ptr); G.pp_semic();
    if (!B.csharp) then (
      print_string "for ("; G.pp_id elem;
      print_string " = (&state_info.";
      print_string (B.id2c state_id); print_string ")->next, ";
      G.pp_id ptr; print_string " = ";
      G.pp_id elem; print_string "->next; "; G.pp_id elem;
      print_string " != (&state_info.";
      print_string (B.id2c state_id); print_string "); ";
      G.pp_id elem; print_string " = "; G.pp_id ptr;
      print_string ", ";
      G.pp_id ptr; print_string " = ";
      G.pp_id elem; print_string"->next) ")
    else (
      if dir = Ast.DEC
      then print_string "list_for_each_prev_safe("
      else print_string "list_for_each_safe(";
      G.pp_id elem; print_string ", "; G.pp_id ptr;
      print_string ", "; G.pp_id state_id;
      print_string ") ";);
    G.start_block();
    if (!B.csharp) then (
      G.pp_id id;
      print_string " = (("; print_string (G.pp_bossa_struct_cs());
      print_string " *)((char *)(";
      G.pp_id elem; print_string ")-(ulong)((int)(&((";
      print_string (G.pp_bossa_struct_cs());
      print_string " *)0)->list_info)/2)));")
    else (
      G.pp_id id; print_string " = list_entry(";
      G.pp_id elem;
      print_string ", ";
      print_string (G.pp_bossa_struct_cs());
      G.print_nl_string ", list_info);");
    print_stmt();
    G.end_block();
    G.end_block())
  else (
    if (!B.csharp) then (
      print_string "for ("; G.pp_id elem;
      print_string " = state_info."; print_string (B.id2c state_id);
      print_string ".next, prefetch((int*)"; G.pp_id elem;
      print_string "->next) ; "; G.pp_id elem;
      print_string " != &state_info.";
      print_string (B.id2c state_id);
      print_string "; "; G.pp_id elem; print_string " = ";
      G.pp_id elem; print_string "->next, prefetch((int*)";
      G.pp_id elem; print_string "->next)) ";
      ) else (
      if dir = Ast.DEC
      then print_string "list_for_each_prev("
      else print_string "list_for_each(";
      G.pp_id elem; print_string ", "; G.pp_id state_id;
      print_string ") ";
      );
    G.start_block();
    if (!B.csharp) then (
      G.pp_id id; print_string " = ((";
      print_string (G.pp_bossa_struct_cs());
      print_string " *)((char *)(";
      G.pp_id elem; print_string ")-(ulong)((int)(&((";
      print_string (G.pp_bossa_struct_cs());
      print_string " *)0)->list_info)/2)));")
    else (
      G.pp_id id; print_string " = list_entry(";
      G.pp_id elem;
      print_string ", ";
      print_string (G.pp_bossa_struct_cs());
      G.print_nl_string ", list_info);");
    print_stmt();
    G.end_block();
    G.end_block())

and pp_for_array state_id id print_stmt move_possible dir =
  (* this has to test whether a process variable is empty.
     the analysis should determine whether this is possible!!! *)
  let index = B.fresh_id() in
  let sublist = B.fresh_id() in
  let selected_size = !CS.selected_size in
  G.start_block();
  print_string "int "; G.pp_id index; G.pp_semic();
  (* for loop *)
  let (op,min,max) =
    if dir = Ast.DEC
    then ("--",selected_size-1,-1)
    else ("++",0,selected_size) in
  print_string "for (";
  G.pp_id index; print_string " = "; print_int min; print_string "; ";
  G.pp_id index; print_string " != "; print_int max; print_string "; ";
  G.pp_id index; print_string op; print_string ") "; G.start_block();
  print_string (G.de_structify "struct list_head *"); G.pp_id sublist;
  print_string " = "; G.pp_arrayref_id state_id index; G.pp_semic();
  pp_for_queue sublist id print_stmt move_possible dir;
  force_newline();
  G.end_block(); G.end_block()
      
and pp_for id print_stmt ar_print_stmt (*not used*) move_possible states dir =
  List.iter
    (function
	CS.STATE(state_id,_,CS.PROC,_) ->
	  pp_for_proc state_id id print_stmt; force_newline()
      | CS.STATE(_,CS.NOWHERE,_,_) | CS.STATE(_,CS.TERMINATED,_,_) ->
	  (* a bit sloppy to allow these, but convenient for looping over all
	     states *)
	  ()
      | CS.STATE(state_id,_,CS.QUEUE(CS.SELECT(_)),_)
	when !CS.selected_size > 1 ->
	  pp_for_array state_id id print_stmt move_possible dir;
	  force_newline()
      | CS.STATE(state_id,_,CS.QUEUE(_),_) ->
	  pp_for_queue state_id id print_stmt move_possible dir;
	  force_newline())
    states

(* ------------------------ End of for loops ----------------------- *)
	
and pp_seq s =
  let (decls,stmts) =
    match s with
      Ast.SEQ(d,s,_) -> (d,s)
    | _ -> ([],[s]) in
  let stmts = List.rev(flatten stmts) in
  (G.start_block();
   (match decls with
     [] -> ()
   | x::xs -> pp_defs false decls);
   G.print_between force_newline pp_stmt stmts;
   G.end_block())
    
and pp_seq2 s s2 =
  let (decls,stmts) =
    match s with
      Ast.SEQ(d,s,_) -> (d,s)
    | _ -> ([],[s]) in
  let stmts = List.rev(flatten stmts) in
  (G.start_block();
   (match decls with
     [] -> ()
   | x::xs -> pp_defs false decls);
   G.print_between force_newline pp_stmt stmts; force_newline ();
   print_string s2;
   G.end_block())
    
and pp_seq3 s idl idr =
  let (decls,stmts) =
    match s with
      Ast.SEQ(d,s,_) -> (d,s)
    | _ -> ([],[s]) in
  let stmts = List.rev(flatten stmts) in
  (G.start_block();
   (match decls with
     [] -> ()
   | x::xs -> pp_defs false decls);
   G.pp_id idl; print_string " = "; G.pp_id idr; G.print_nl_string "->next;";
   G.print_between force_newline pp_stmt stmts; force_newline ();
   G.end_block())

and flatten stmts =
  List.fold_left
    (function prev ->
      function
	  Ast.SEQ([],s,_) -> (flatten s)@prev
	| s -> s::prev)
    [] stmts

and pp_cases cases default is_proc_sched =
  (G.start_block();
   List.iter
     (function Ast.SEQ_CASE(idlist,_,stmt,_) ->
       ((if is_proc_sched then print_casename else print_caseid) idlist;
	print_string "  "; open_box 0;
	pp_seq2 stmt "break;"; close_box();
	force_newline()))
     cases;
   G.print_nl_string "default:";
   if (!B.csharp) then G.start_block();
   (match default with
     None ->
       G.print_nl_string
	 "  panic(\"unknown state: file: %s, line: %d!.\",";
	 if (!B.csharp) then (print_string "        \""; print_string (!G.policy_name); print_string ".cs\", 0);")
		else print_string "        __FILE__, __LINE__);";
       print_string "return 0;"
   | Some x -> pp_stmt x);
   if (!B.csharp) then (print_string "break;"; G.end_block(););
   G.end_block())

and move_possible = function
    Ast.IF(exp,stm1,stm2,attr) -> move_possible stm1 || move_possible stm2
  | Ast.SWITCH(exp,cases,default,attr) ->
      (List.exists
	(function Ast.SEQ_CASE(pat,states,stm1,attr) -> move_possible stm1)
	cases)
	or
      (match default with
	None -> false
      |	Some x -> move_possible x)
  | Ast.SEQ(decls,stms,attr) -> List.exists move_possible stms
  | Ast.FOR(index,state,state_info,dir,stm1,attr) -> move_possible stm1
  | Ast.FOR_WRAPPER(label,stms,attr) -> List.exists move_possible stms
  | Ast.MOVE(exp,state,srcs,verifsrcs,dst,auto_allowed,state_end,attr) -> true
  | Ast.PRIMSTMT(Ast.VAR(id,_),_,_) -> B.id2c id = "update_state"
  | stm -> false
