(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: bossa2c.ml,v 1.24 2004/11/11 16:15:00 julia Exp $ *)
  
module B = Objects
module CS = Class_state
module G = Genaux
module A = Ast2c
    
open Format
open Ioctl
  
(* ------------------------------- misc ------------------------------ *)
  
let internal_error str =
  raise (Error.Error (Printf.sprintf "bossa2c: %s: internal error" str))

let bitmap_granularity = 32
    
(* ---------------------------- copyright ---------------------------- *)
    
(* Notice de copyright *)
let copyright =
  "/*\n" ^
  "  Copyright (C) 2002 INRIA, Ecole des Mines de Nantes, Univerity of Copenhagen\n" ^
  "  Automatically generated from Bossa specification by the cb compiler\n" ^
  "  --- Do not edit this file ---\n" ^
  "  To compile for use as a module, compile with options \"gcc -O -DMODULE\""^
  "*/\n"
    
(* ----------------------------- checksum ---------------------------- *)

let print_checksum checksum =
  G.print_nl_string ("/* checksum = "^(string_of_int checksum)^" */")
    
(* ----------------------------- exports ----------------------------- *)
    
let make_exports name =
  if (!B.csharp) then
    "private static ___scheduler_operations EXPORTS;\n"^
    "private const string _NAME = \""^name^"\";\n"^
    "const string scheduler_name = _NAME;\n"
  else
    "#ifndef EXPORTS\n"^
    "#define EXPORTS "^(String.lowercase name)^"_sched_ops\n"^
    "#endif\n"^
    "#ifndef NAME\n"^
    "#define NAME \""^(String.lowercase name)^"\"\n"^
    "#endif\n"^
    "#define "^
    (B.sched_select "PROCESS_SCHEDULER" "VIRTUAL_SCHEDULER")
      
      
(* ----------------------- structure prototypes ---------------------- *)
      
let make_structure_prototypes name =
  let local_struct = G.pp_bossa_struct_cs() in
  local_struct^
  ";\n#define PROTOCOL_PROCESS_STRUCT   "^
  (B.sched_select local_struct "struct bossa_struct")^
  "\n#define PROTOCOL_SCHEDULER_STRUCT "^
  (B.sched_select "struct scheduler_struct" local_struct)^
  "\n#define PROTOCOL_BOSSA_STRUCT     "^local_struct^"\n"
							 
(* -------------------------- include files -------------------------- *)
							 
let includefiles _ =
  "/* Compile with options \"gcc -O -D__KERNEL__ \" */\n"^
  "#ifdef __KERNEL__\n"^
  "   #include <linux/sched.h>\n"^
  "   #include <linux/string.h>\n"^
  "   #include <linux/slab.h>\n"^
  "   #include <asm/mmu_context.h>\n"^
  "   #include <linux/interrupt.h>\n"^
  "   #include <linux/smp_lock.h>\n"^
  "   #include <linux/errno.h>\n"^
  "   #include <linux/list.h>\n"^
  "   #include <linux/prefetch.h>\n"^
  "#else\n"^
  "   struct timer_list { int x; };\n"^
  "   #include \"bossa.h\"\n"^
  "   struct task_struct { int pid; void *mm; char *comm; };\n"^
  "   #define kmalloc(x,y) malloc(x)\n"^
  "   #define kfree(x) free(x)\n"^
  "   #define GFP_ATOMIC 0\n"^
  "   #define panic printf\n"^
  "#endif\n"^
  "#define array_ref(x,y,z) x[y].z\n"^
  "#define field_array_ref(x,y,z) x->y[z]\n"^
  "#define sizeof_struct(x) sizeof(struct x)\n"^
  "#define UNBLOCK_INFO_OFFSET 0\n"^
  "#define drop_unblock_res(x) (x&~(1<<UNBLOCK_INFO_OFFSET))\n"^
  "#define get_unblock_res(x) (x&(1<<UNBLOCK_INFO_OFFSET))\n"^
  "#define set_current_need_resched(n) current->need_resched = n\n"^
  "#define SUCCESS 0\n"^
  "#define FAILURE -1\n"^
  (if !B.high_res
  then ("#include <linux/hrtime.h>\n"^
	"extern int bossa_tstojiffie(unsigned long, long, unsigned long *);\n"^
	"#define time_to_ticks(x) (x).tv_sec\n"^
	"#define time_to_jiffies(x) (x).tv_sec\n"^
	"#define time_to_subjiffies(x) (x).tv_nsec\n"^
	"#define make_cycles(x) x\n"^
	"#define make_cycle_time(x,y) {x,y}\n"^
	"static struct timespec global_time_tmp;\n"^
	"static inline struct timespec ticks_to_time(int x) {\n  global_time_tmp.tv_sec = x; global_time_tmp.tv_nsec = 0;\n  return global_time_tmp;\n}\n"^
	"#define tstojiffie(x,y,z) bossa_tstojiffie(x,y,&z)\n"^
	"#define call_normalize_time(x,y,z) normalize_time(x,y,&z)\n")
  else ("#define time_to_ticks(x) x\n"^
	"#define ticks_to_time(x) x\n"))
    
(* -------------------------- useful macros -------------------------- *)
    
let usefulmacros trace =
  if(!B.csharp) then
    "[MethodImplAttribute(MethodImplOptions.InternalCall)]\n"^
    "static private extern uint event_mask0(uint e);\n"^
    "\n"^
    "[MethodImplAttribute(MethodImplOptions.InternalCall)]\n"^
    "static private extern uint event_mask1(uint e);\n"^
    "\n"^
    "[MethodImplAttribute(MethodImplOptions.InternalCall)]\n"^
    "static private extern uint event_mask2(uint e);\n"^
    "\n"^
    "[MethodImplAttribute(MethodImplOptions.InternalCall)]\n"^
    "static private extern uint event_mask3(uint e);\n"
  else
    "#define bossa_assert(x) do{if(!(x)) panic(\"Error in \" #x \"\\n\");} while(0)\n"^
    "#define state_of(e) ((e)->state)\n"^
    "#define update_state(e,s,c) ((e)->state=(s))\n"^
    "#define event_mask0(e) ((e) & 0xff000000)\n"^
    "#define event_mask1(e) ((e) & 0xffff0000)\n"^
    "#define event_mask2(e) ((e) & 0xffffff00)\n"^
    "#define event_mask3(e) ((e) & 0xffffffff)\n"^
    "#define true 1\n"^
    "#define false 0\n"^
    "#define null NULL\n"^
    "typedef int bool;\n"^
    "typedef void * port;\n"^
    (match !B.schedtype with
      B.PROCSCHED -> "static char *error_string;\n"
      (* error string not needed for VS because there is no atomic
	 attach-detach-attach *)
    | B.VIRTSCHED -> "")^
    (match trace with
      Ast.TRACE(n,events,exps,test,attr) ->
	"#define TRACE(x,y,z) trace(x,y,z)\n"
    | _ -> "")
      
(* ----------------------- global declarations ----------------------- *)
      
let pp_defs global defs extra =
  List.iter (A.pp_def global) defs;
  G.nl_ifnonempty(defs);
  if extra && not (!B.csharp)
  then
    (G.print_nl_string "extern spinlock_t bossa_scheduler_lock;";
     G.print_nl_string "extern struct bossa_struct *rts_idle_process;")
      
(* ------------------------ User-declared types ----------------------- *)
    
let pp_typedefs typedefs =
  let enum_function id f idlist =
    if (!B.csharp) then print_string "public ";
    print_string "enum"; print_space(); G.pp_id(id); print_space();
    if (!B.csharp) then print_string ": int ";
    G.start_block();
    G.print_between G.pp_comma f
      idlist;
    G.end_block();
    if (!B.csharp)
    then force_newline()
    else G.pp_semic() in
  List.iter
    (function
	Ast.ENUMDEF(id2,idlist,attr) ->
	  let ctr = ref 0 in
	  enum_function id2
	    (function id ->
	      if (!B.csharp)
	      then A.list_of_const_val_enum :=
		!A.list_of_const_val_enum @ [(B.id2c(id2))^"."^(B.id2c(id))];
	      G.pp_id id; print_string " = ";
	      print_int (!ctr); ctr := (!ctr) + 1)
	    idlist
      | Ast.ENUMVALDEF(id2,idlist,attr) ->
	  enum_function id2
	    (function (id,exp) ->
	      if (!B.csharp)
	      then A.list_of_const_val_enum :=
		!A.list_of_const_val_enum @ [(B.id2c(id2))^"."^(B.id2c(id))];
	      G.pp_id id; print_string " = "; A.pp_exp exp)
	    idlist
      | Ast.RANGEDEF(id,exp1,exp2,attr) ->
	  (print_string "typedef int "; G.pp_id(id); G.pp_semic())
      | Ast.TIMERDEF(id,decls,attr) ->
	  print_string "struct "; G.pp_id id; print_string " ";
	  G.start_block();
	  G.print_between force_newline
	    (function x -> A.pp_decl false x; print_string ";")
	    decls;
	  G.end_block(); G.pp_semic(); force_newline())
    typedefs; G.nl_ifnonempty(typedefs)
    
(* ------------------------ External functions ----------------------- *)
    
let cpt_argt = ref 0
    
let pp_etype_argt param =
  G.pp_etype param;
  if (!B.csharp)
  then
    begin
      cpt_argt := !cpt_argt + 1;
      print_string " argt";
      print_int !cpt_argt;
    end
      
let pp_fundecls fundecls =
  List.iter
    (function Ast.FUNDECL(ret,nm,params,attr) ->
      G.external_functions := nm :: (!G.external_functions);
      if (!B.csharp)
      then
	(G.print_nl_string
	   "[MethodImplAttribute(MethodImplOptions.InternalCall)]";
	 print_string "private unsafe static extern ");
      G.pp_etype ret; print_string " "; print_string (B.id2c(nm));
      print_string " "; 
      if params = []
      then (G.mk_empty_prototype(); G.pp_semic())
      else (print_string "("; G.print_between G.pp_comma pp_etype_argt params;
	    G.print_nl_string ");"))
    fundecls; G.nl_ifnonempty(fundecls)
    
(* ------------------------- persistent timers ----------------------- *)

let count_persistent_timers procdecls =
  let check_vardecl = function
      Ast.VARDECL(B.TIMER(B.PERSISTENT(_)),id,_,_) -> true
    | _ -> false in
  List.length(List.filter check_vardecl procdecls)

let pp_currently_running procdecls =
  match count_persistent_timers procdecls with
    0 -> ()
  | ct -> force_newline(); print_string "int currently_running[";
      print_int ct; print_string "];"

(* ------------------------- process structure ----------------------- *)
    
let pp_com_bstruct com decl_name =
  print_string com; force_newline(); G.pp_bossa_struct(); print_string " *";
  print_string decl_name; print_string ";"; force_newline()
    
let pp_procdecls procdecls =
  if (!B.csharp) then
    print_string "public unsafe struct ";
  G.pp_bossa_struct(); print_string " ";
  G.start_block();
  G.print_nl_string "/* default attributes for Bossa */";
  (match !B.schedtype with
    B.PROCSCHED ->
      G.print_nl_string
	"/* next and previous processes in the current bossa queue */";
      G.mk_public "struct list_head list_info;";
      force_newline();
      G.print_nl_string
	"/* list of virtual schedulers on the way to this process */";
      G.mk_public "struct next_list_struct *next_list;";
      force_newline();
      G.mk_public "int state;";
      force_newline();
  | B.VIRTSCHED ->
      G.print_nl_string
	"/* next and previous processes in the current bossa queue */";
      G.mk_public "struct list_head list_info;";
      force_newline();
      G.mk_public "struct scheduler_operations *ops;";
      force_newline();
(*      G.mk_public "struct event_struct deferred_queue;";
      force_newline();*)
      G.mk_public "int state;";
      force_newline());
  G.print_nl_string "/* Additional process attributes";
  G.print_nl_string " *  specified by the scheduling policy";
  G.print_nl_string " *  in the process = {...} declaration";
  G.print_nl_string "*/";
  
  G.print_between force_newline
    (function x ->
      (match x with
	Ast.VARDECL(ty,id,true,attr) -> print_string "//"
      |	_ -> ());
      if (!B.csharp) then print_string "public ";
      A.pp_decl false x; print_string ";")
    procdecls;
  pp_currently_running procdecls;
  G.end_block(); G.pp_semic(); force_newline()
    
(* ----------------------- operation prototypes ---------------------- *)
    
let top_level_declarations top_level_events name =
  let tgt_type =
    (match !B.schedtype with
      B.PROCSCHED -> G.pp_bossa_struct_cs()
    | B.VIRTSCHED -> "struct bossa_struct") in
  let flat_name l = String.concat "_" (List.map B.id2c l) in
  let _ =
    Events.top_level_event_names_loop
      (function prev -> function last ->
	function specd_name -> function star -> function attr ->
	  G.mk_inline();
	  print_string
	    ("static int "^name^"_"^
	     (flat_name (List.flatten specd_name)));
	  let snm = Ast.event_name2c_ last in
	  if snm = "bossa_schedule"
	  then G.mk_empty_prototype()
	  else if snm = "unblock_timer_persistent"
	  then G.pp_arglist true
	      ["int";"struct timer_list *";"void *";"int";
		(tgt_type^" *");"struct next_list_struct *"]
	  else G.pp_arglist true ["int";(tgt_type^" *");
				  "struct next_list_struct *"];
	  G.pp_semic())
      (function entry_point ->
	let snm = Ast.event_name2c_ entry_point in
	print_string ("static int "^name^"_"^snm^"_entry");
	match snm with
	  "bossa_schedule" -> G.mk_empty_prototype(); G.pp_semic()
	| "unblock_timer_target" | "unblock_timer_notarget" ->
	    G.pp_arglist true
	      ["int";(tgt_type^" *");"struct next_list_struct *"];
	    G.pp_semic()
	| "unblock_timer_persistent" ->
	    G.pp_arglist true
	      ["int";"struct timer_list *";"void *";"int";
		(tgt_type^" *");"struct next_list_struct *"];
	    G.pp_semic()
	| _ -> G.pp_arglist true ["int";(tgt_type^" *")]; G.pp_semic())
      top_level_events in
  ()
    
let pp_prototypes top_level_events nm =
  G.print_nl_string "#ifdef MODULE";
  G.print_nl_string "/************ Routines related to modules *************/";
  G.print_nl_string ("static int __init "^nm^"_init_module(void);");
  G.print_nl_string ("static void __exit "^nm^"_cleanup_module(void);");
  G.print_nl_string "/******************************************************/";
  G.print_nl_string "#endif";
  G.print_nl_string "/******** to know if the module is being used ********/";
  G.print_nl_string ("static inline void increment_ref_count(void);");
  G.print_nl_string ("static inline void decrement_ref_count(void);");
  G.print_nl_string "/******************************************************/";
  force_newline();
  G.print_nl_string ("static void "^nm^"_scheduler_init (void);");
  G.print_nl_string ("static int "^nm^"_ioctl_init (void);");
  print_string ("static int "^nm^"_scheduler");
  G.pp_arglist true ["struct event_struct *e";
		     "struct next_list_struct *src_next_list";
		     "struct next_list_struct *tgt_next_list"];
  G.pp_semic();
  G.print_nl_string ("static int "^nm^
		     "_compute_state(struct next_list_struct *next_list);");
  G.print_nl_string ("static int "^nm^
		     "_ps_attach(struct next_list_struct *next_list);");
  print_string ("static int "^nm^"_ps_detach");
  G.pp_arglist true ["struct next_list_struct *next_list";
		     "struct bossa_struct *_p"];
  G.pp_semic();
  if !B.schedtype = B.VIRTSCHED
  then
    (print_string("static int "^nm^"_attach");
     G.pp_arglist true [(G.pp_bossa_struct_cs()^" *child");"int data[]"];
     G.pp_semic();
     G.print_nl_string("static int "^nm^"_attach_info(char *info[]);");
     G.print_nl_string("static int "^nm^"_detach("^G.pp_bossa_struct_cs()^
		       " *child, bool discard_child);");
     G.print_nl_string
       ("static struct scheduler_struct *\n"^nm^
	"_create_scheduler(struct scheduler_operations *ops, int);"));
  G.print_nl_string("static void "^nm^"_attach_self(void);");
  G.print_nl_string("static PROTOCOL_SCHEDULER_STRUCT *"^nm^
		    "_detach_self(bool is_root);");
  print_string ("static uint "^nm^"_check_process_state");
  G.pp_arglist true
    [(B.sched_select (G.pp_bossa_struct_cs()) "struct bossa_struct")
     ^" *tgt";
     "struct next_list_struct *next";
     "uint classes"];
  G.pp_semic();
  print_string
    ("static void "^nm^"_preempt(void);\n"^
     "static char *\n"^
     "queueToString(struct list_head *start, char *buffer, bool long_format);\n"^
     "static int "^nm^"_do_taskprocinfo(struct next_list_struct *, struct bossa_struct *, char *, bool, bool, bool);\n"^
     "static int "^nm^"_do_schedprocinfo(char *, bool);\n"^
     (if (!B.csharp)
     then
       "static int "^nm^"_list_sched_info(char * page,char ** start,long off,\n"
     else
       "static int "^nm^"_list_sched_info(char * page,char ** start,off_t off,\n")^
     "				 int count,int * eof,void * data);\n"^
     "static int "^nm^"_count_processes(void);\n"^
     "static void "^nm^"_print_processes(void);\n"^
     "static void "^nm^"_set_priority("^
     (B.sched_select (G.pp_bossa_struct_cs()) "struct bossa_struct")^
     " *, struct next_list_struct *, int);\n");
  top_level_declarations top_level_events nm;
  force_newline();
  if not (!B.csharp)
  then
    G.print_nl_string
      "static char *sched_type[4] = {\"Virtual Scheduler\",\"Process Scheduler\",\"VS\",\"PS\"};";
  force_newline()
    
(* ----------------------- operation structure ----------------------- *)
    
let top_level_struct_decls top_level_events name =
  let ty = "(int (*)(int,struct bossa_struct *,struct next_list_struct *))" in
  let uty = "(int (*)(int,struct bossa_struct *,struct next_list_struct *))" in
  let upty = ("(int (*)(int,struct timer_list *,void *,int,"^
	      "struct bossa_struct *,struct next_list_struct *))") in
  let ety = "(int (*)(int,struct bossa_struct *))" in
  let flat_name l = String.concat "_" (List.map B.id2c l) in
  let process_events =
    Events.top_level_event_names_loop
      (function prev -> function last ->
	function specd_name -> function star -> function attr ->
	  let snm = Ast.event_name2c_ last in
	  match snm with
	    "bossa_schedule" ->
	      ("  "^name^"_"^(flat_name (List.flatten specd_name)))
	  | "unblock_timer_persistent" ->
	      ("  "^upty^name^"_"^(flat_name (List.flatten specd_name)))
	  | _ -> ("  "^ty^name^"_"^(flat_name (List.flatten specd_name))))
      (function entry_point ->
	let snm = Ast.event_name2c_ entry_point in
	match snm with
	  "bossa_schedule" -> ("  "^name^"_"^snm^"_entry")
	| "unblock_timer_target" | "unblock_timer_notarget" ->
	    ("  "^uty^name^"_"^snm^"_entry")
	| "unblock_timer_persistent" ->
	    ("  "^upty^name^"_"^snm^"_entry")
	| _ -> ("  "^ety^name^"_"^snm^"_entry"))
      top_level_events in
  (String.concat ",\n"
     (List.map (function (a,b) -> String.concat ",\n" (a @ [b]))
	process_events))^
  (if process_events = [] then "" else ",\n")
    
let operationstructure top_level_events nm default =
  "struct scheduler_operations EXPORTS = {\n"^
  "  "^nm^"_scheduler,\n"^
  (top_level_struct_decls top_level_events nm) ^
  "  \""^nm^"\",\n"^
  (B.sched_select "  PS,\n" "  VS,\n")^
  (if default then "  true,\n" else "  false,\n")^
  "  "^nm^"_scheduler_init,\n"^
  "  "^nm^"_ioctl_init,\n"^
  (B.sched_select "  NULL,\n" ("  "^nm^"_create_scheduler,\n"))^
  "  "^nm^"_check_process_state,\n"^
  "  "^nm^"_preempt,\n"^
  "  "^nm^"_do_taskprocinfo,\n"^
  "  "^nm^"_do_schedprocinfo,\n"^
  "  "^nm^"_list_sched_info,\n"^
  "  "^nm^"_count_processes,\n"^
  "  "^nm^"_print_processes,\n"^
  "  "^nm^"_set_priority,\n"^
  "  "^nm^"_compute_state,\n"^
  "  "^nm^"_ps_attach,\n"^
  "  "^nm^"_ps_detach,\n"^
  (B.sched_select "  NULL,\n" ("  "^nm^"_attach,\n")) ^
  (B.sched_select "  NULL,\n" ("  "^nm^"_attach_info,\n")) ^
  (B.sched_select "  NULL,\n" ("  "^nm^"_detach,\n")) ^
  ("  "^nm^"_attach_self,\n") ^
  ("  "^nm^"_detach_self,\n") ^
  ("  increment_ref_count,\n") ^
  ("  decrement_ref_count,\n") ^
  "  { NULL, NULL }, // filled in at installation with the path to the scheduler\n"^
  "};\n"
    
(* ------------------------------ states ----------------------------- *)
    
let pp_vis = function
    None -> ()
  | Some(CS.PUBLIC) -> ()
  | Some(CS.PRIVATE) -> print_string " /* private */"
	
(* because the empty queue is actually a structure that points to itself,
   and not a null pointer *)
let pp_bossa_struct_queue_decl queue_typ id vis =
  G.mk_public "struct list_head "; G.pp_struct_id(id);
  (match queue_typ with
    CS.SELECT(_) ->
      let sz = !CS.selected_size in
      if sz > 1
      then (print_string "["; print_int sz; print_string "]")
  | _ -> ());
  print_string ";";
  pp_vis vis
    
let pp_bossa_struct_var_decl id vis =
  if (!B.csharp) then print_string "public ";
  G.pp_bossa_struct(); print_string " *"; G.pp_struct_id(id); print_string ";";
  pp_vis vis
    
let pp_states states =
  print_string "/* definition of protocol states */\n";
  if (!B.csharp) then (
    print_string "static bossa_struct *running;\n\n";
    print_string "public unsafe struct ___State_info ")
  else (
    print_string "#ifndef __KERNEL__\n";
    print_string "struct bossa_struct *running;\n";
    print_string "#endif\n";
    print_string "struct state_info ");
  G.start_block();
  G.print_between force_newline
    (function
	Ast.QUEUE(clsname,CS.EMPTY,id,vis,attr) ->
	  pp_bossa_struct_queue_decl CS.EMPTY id None
      | Ast.QUEUE(clsname,queue_typ,id,vis,attr) ->
	  pp_bossa_struct_queue_decl queue_typ id (Some vis)
      | Ast.PROCESS(clsname,id,previd,vis,attr) ->
	  pp_bossa_struct_var_decl id (Some vis);
	  (match previd with
	    Ast.NO_PREV -> ()
	  | Ast.BOSSA_PREV(id) ->
	      force_newline(); pp_bossa_struct_var_decl id None
	  | Ast.RTS_PREV(id) ->
	      ()(*
		   G.print_nl_string "\n#ifndef __KERNEL__";
		   print_string "struct bossa_struct *"; G.pp_id id;
		   print_string ";\n#endif"*)))
    states;
  G.end_block();
  G.pp_semic();  force_newline(); force_newline();
  G.mk_static_private();
  if (!B.csharp)
  then G.print_nl_string "___State_info state_info;"
  else G.print_nl_string "struct state_info state_info __cacheline_aligned;";
  force_newline(); force_newline()
    
(* ------------------------- trace structure ------------------------- *)
    
let pp_trace_struct = function
    Ast.TRACE(n,events,exps,test,attr) ->
      (* structure declaration *)
      print_string "struct trace_struct ";
      G.start_block();
      G.print_nl_string "int jiffies;";
      G.print_nl_string "uint _event;";
      G.print_nl_string "uint event2;";
      (match !B.schedtype with
	B.PROCSCHED ->
	  G.print_nl_string "char comm[16];";
	  G.print_nl_string "int pid;"
      |	B.VIRTSCHED -> G.print_nl_string "char comm[16];");
      G.print_nl_string "int state;";
      G.print_between force_newline
	(function
            Ast.TRACEVAR(fldid,id,attr) ->
	      A.pp_decl false (Ast.VARDECL(B.ty attr, fldid, false, attr));
	      print_string ";"
	  | Ast.TRACEFIELD(id,attr) ->
	      A.pp_decl false (Ast.VARDECL(B.ty attr, id, false, attr));
	      print_string ";"
	  | Ast.TRACEQUEUE(fldid,refid,attr) ->
	      A.pp_decl false (Ast.VARDECL(B.ty attr, fldid, false, attr));
	      print_string ";")
	exps;
      G.end_block(); G.pp_semic();
      force_newline();
      (* variable declarations *)
      print_string "static struct trace_struct trace_buffer[";
      print_int n;
      G.print_nl_string "];";
      (match test with
	None ->
	  G.print_nl_string "static int trace_ctr = "; print_int n;
	  G.pp_semic()
      |	Some expr ->
	  G.print_nl_string "static int trace_ctr = 0;");
      G.print_nl_string "#define EVENT_BOSSA_SCHEDULE -1 // not a real event";
      G.print_nl_string "#define EVENT_PREEMPT -2 // not a real event";
      force_newline();
      force_newline()
  | _ -> ()
	
(* ----------------------- comparison functions ---------------------- *)
	
let pp_functions functions =
  force_newline();
  G.print_between (function _ -> force_newline(); force_newline())
    (function Ast.FUNDEF(ret,nm,params,stmt,inl,attr) ->
      (match inl with
	Ast.INLINE -> G.mk_inline()
      |	_ -> ());
      G.mk_static();
      G.pp_type ret; print_space (); B.type_return := G.pp_type_cs ret;
      G.pp_id(nm);
      (match params with
	[] -> G.mk_empty_prototype()
      | _ -> (print_string "("; open_box 0;
	      G.print_between G.pp_comma (A.pp_decl false) params);
	  close_box(); print_string ")");
      print_string " ";
      A.pp_seq stmt)
    functions;
  force_newline(); force_newline()
    
(* ----------------------- initialize scheduler ---------------------- *)
    
let pp_init_states name states =
  let _i_ = ref 0 in
  _i_ := 0;
  G.mk_static(); print_string "void ";
  print_string name; print_string "_scheduler_init";
  G.mk_empty_prototype(); print_string " ";
  G.start_block();
  G.print_nl_string "/* initialization of states */";
  G.print_between force_newline
    (G.process_state
       (function array -> G.print_nl_string "init_bitmap();")
       (function array ->
	 function index ->
	   if (!B.csharp) then (
	     print_string "fixed (list_head* ptr = &";
	     G.pp_arrayref_id array index; print_string ") {";
	     G.pp_arrayref_id array index; print_string ".next = ";
	     G.pp_arrayref_id array index;
	     print_string ".prev = ptr;}")
	   else (
	     print_string "init_queue("; G.pp_arrayref_id array index;
	     print_string ");"))
       (function qid ->
         if (!B.csharp) then (
	   print_string "fixed (list_head* ptr"; print_int (!_i_);
	   print_string " = &"; G.pp_id qid; print_string ") {";
	   G.pp_id qid; print_string ".next = "; G.pp_id qid;
	   print_string ".prev = ptr"; print_int (!_i_);
	   print_string ";}";_i_:=(!_i_)+1;)
	 else (
	   print_string "init_queue("; G.pp_id qid; print_string ");"))
       (function vid ->
	 if (!B.csharp) then (
	   G.pp_id vid; print_string " = null;";)
	 else (
	   print_string "init_var(&"; G.pp_id vid; print_string ");"))
       (function vid ->
	 if (!B.csharp) then (
	   G.pp_id vid; print_string " = null;";)
	 else (
	   print_string "init_var(&"; G.pp_id vid; print_string ");")))
    states;
  G.end_block(); force_newline(); force_newline()
    
(* ----------------------- check process state ----------------------- *)
    
let pp_check_process_state nm =
  match !B.schedtype with
    B.PROCSCHED ->
      G.mk_static();
      print_string ("uint "^nm^"_check_process_state");
      G.pp_arglist true [(G.pp_bossa_struct_cs())^" *tgt";
			 "struct next_list_struct *next";
			 "uint classes"];
      print_string " ";
      G.start_block();
      print_string "switch(tgt->state) ";
      G.start_block();
      List.iter
	(function
	    (_,CS.STATE(id,CS.NOWHERE,_,_)) -> ()
	  |	(_,CS.STATE(id,cls,_,_)) ->
	      print_string "case ";
	      if (!B.csharp) then print_string "(int) ";
	      G.pp_state_cst id;
	      print_string ": return classes & ";
	      if (!B.csharp) then print_string "(uint) SchedState.";
	      print_string (CS.classic_class2c cls);
	      G.print_nl_string ";")
	(!CS.state_env);
      print_string "default: return 0;";
      G.end_block();
      G.end_block();
      force_newline();
      force_newline()
  | B.VIRTSCHED ->
      G.print_nl_string "static uint";
      print_string (nm^"_check_process_state");
      G.pp_arglist true ["struct bossa_struct *tgt";
			 "struct next_list_struct *next";
			 "uint classes"];
      print_string " ";
      G.start_block();
      print_string
	"return next->sched->ops->check_process_state(tgt,next->next,classes);";
      G.end_block();
      force_newline();
      force_newline()

(* -------------------------- declare bitmap ------------------------- *)

let pp_declare_bitmap _ =
  let sz = !CS.selected_size in
  if sz > 1
  then
    let bitmap_name = (!G.policy_name)^"_bitmap" in
    G.print_nl_string ("#define BITMAP "^bitmap_name);
    let count = (sz + (bitmap_granularity-1)) / bitmap_granularity in
    if count = 1
    then
      begin
	G.print_nl_string "#define WORD_BITMAP";
	G.print_nl_string ("static unsigned long "^bitmap_name^";")
      end
    else
      begin
	G.print_nl_string "#define ARRAY_BITMAP";
	G.print_nl_string ("#define bitmap_size "^(string_of_int count));
	G.print_nl_string
	  ("static unsigned long "^bitmap_name^"["^(string_of_int count)^"];")
      end;
    force_newline()

(* ------------------ create a new child scheduler ------------------ *)
	
let pp_create_scheduler nm states =
  G.print_nl_string "static struct scheduler_struct *";
  print_string (nm^"_create_scheduler");
  G.pp_arglist true ["struct scheduler_operations *ops";"int new_root"];
  print_string " ";
  G.start_block();
  G.print_nl_string ((G.pp_bossa_struct_cs())^" *new;");
  G.print_nl_string "children++;";
  G.print_nl_string
    ("new = ("^(G.pp_bossa_struct_cs())^
     " *)kmalloc(sizeof("^(G.pp_bossa_struct_cs())^"),GFP_KERNEL);");
  G.print_nl_string "new->ops = ops;";
  print_string "if (new_root) ";
  G.start_block();
  print_string "if (root_s_state == RUNNING) ";
  G.start_block();
  List.iter
    (function
	(_,CS.STATE(id,CS.RUNNING,_,_)) ->
	  print_string "new->state = ";
	  G.pp_state_cst id; G.pp_semic();
	  print_string "move_nowhere_proc(new,";
	  G.pp_id id;
	  print_string ");"
      |	_ -> ())
    states;
  G.end_block();
  force_newline();
  print_string "else ";
  G.start_block();
  List.iter
    (function
	(_,CS.STATE(id,_,CS.QUEUE(CS.SELECT(_)),_)) ->
	  print_string "new->state = ";
	  G.pp_state_cst id; G.pp_semic();
	  print_string "move_nowhere_queue(new,";
	  G.pp_id id;
	  print_string ");"
      |	_ -> ())
    states;
  G.end_block();
  force_newline();
  G.print_nl_string "// The following two lines make no sense";
  G.print_nl_string "// set_current_need_resched(1);";
  print_string "// new->ops->preempt();";
  G.end_block();
  force_newline();
  print_string "else ";
  G.start_block();
  List.iter
    (function
	(_,CS.STATE(id,CS.BLOCKED,CS.QUEUE(_),CS.PUBLIC)) ->
	  print_string "new->state = ";
	  G.pp_state_cst id; G.pp_semic();
	  print_string "move_nowhere_queue(new,";
	  G.pp_id id;
	  print_string ");"
      |	(_,CS.STATE(id,CS.BLOCKED,_,CS.PUBLIC)) ->
	  print_string "new->state = ";
	  G.pp_state_cst id; G.pp_semic();
	  print_string "move_nowhere_proc(new,";
	  G.pp_id id;
	  print_string ");"
      |	_ -> ())
    states;
  G.end_block();
  force_newline();
  print_string "return (struct scheduler_struct *)new;";
  G.end_block();
  force_newline();
  force_newline()
    
(* ------------------------ Forward and event ----------------------- *)
    
let pp_create_forward nm =
  print_string "static inline int forwardImmediate";
  G.pp_arglist true [((G.pp_bossa_struct_cs())^" *tgt");
		     "struct event_struct *e";
		     "struct next_list_struct *src_next_list";
		     "struct next_list_struct *tgt_next_list"];
  print_string " ";
  G.start_block();
  print_string "return tgt->ops->handle_event(e,src_next_list,tgt_next_list);";
  G.end_block();
  force_newline();
  force_newline()
    
    
(* ----------------- Begin code for sorting handlers ---------------- *)
    
type switch_line =
    LINE of B.identifier option * Ast.event_name * Ast.event_decl list
	
let get_name n (Ast.EVENT_NAME(l,_,_)) =
  (try Some (List.nth l n) with Failure _ -> None)
    
let classify_handlers n handlers =
  List.fold_left
    (function prev ->
      function Ast.EVENT(nm,_,_,_,_) as x ->
	let id = get_name n nm in
	match prev with
	  [] -> [LINE(id,nm,[x])]
	| (LINE(id1,_,l)::rest) as z ->
	    if id = id1
	    then (LINE(id,nm,x::l))::rest
	    else (LINE(id,nm,[x]))::z)
    [] handlers
    
(* handlers are sorted *)
let rec nest_handlers n event = function
    [Ast.EVENT(_,_,_,s,_)] -> A.pp_seq s
  | handlers ->
      let handler_lists = classify_handlers n handlers in
      let (default,handler_lists) =
	(match handler_lists with
	  ((LINE(None,_,[x]))::rest) -> (Some x,rest)
	| _ -> (None,handler_lists)) in
      let len = 1 + List.length handler_lists in
      if len > 3
      then
	if (!B.csharp)
	then ifify n event handler_lists default
	else switchify n event handler_lists default
      else ifify n event handler_lists default

and switchify n event handler_lists default =
  print_string (Printf.sprintf "switch (event_mask%i(%s)) " n event);
  G.start_block(); switch_lines n event handler_lists default; G.end_block()

and switch_lines n event handler_lists default =
  List.iter
    (function
	LINE(Some id, nm, l) ->
	  print_string "case ";
	  print_string (Ast.event_name2partial_mask n nm);
	  G.print_nl_string ":";
	  (match l with
	    [Ast.EVENT(_,_,_,stmt,_)] -> A.pp_seq(stmt)
	  | _ -> nest_handlers (n+1) event (List.rev l));
	  if (!B.csharp) then G.print_nl_string "break;";
	  force_newline()
      | LINE(None, _, _) ->
	  raise (Error.Error "switch: default found in the middle"))
    handler_lists;
  (match default with
    Some (Ast.EVENT(_,_,_,stmt,_)) ->
      (print_string "default: "; G.start_block(); A.pp_seq(stmt);
       if (!B.csharp) then G.print_nl_string "break;"; G.end_block())
  | None ->
      print_string "default: ";
      G.start_block();
      G.print_nl_string ("panic(\""^(!G.policy_name)^
			 ": unknown event %x\\n\",_event);");
      print_string "return 0;";
      G.end_block())
    
and ifify n event handler_lists default =
  G.start_block();
  let maskval = B.id2c(B.fresh_idx("event")) in
  if not (handler_lists = [])
  then
    begin
      print_string "uint ";
      print_string maskval; print_string " = ";
      print_string (Printf.sprintf "event_mask%i(%s)" n event);
      G.pp_semic()
    end;
  if_lines maskval n event handler_lists default; G.end_block()
    
and if_lines maskval n event handler_lists default =
  List.iter
    (function
	LINE(Some id, nm, l) ->
	  print_string "if ("; print_string maskval; print_string " == ";
	  print_string (Ast.event_name2partial_mask n nm);
	  print_string ") ";
	  (match l with
	    [Ast.EVENT(_,_,_,stmt,_)] -> A.pp_seq(stmt)
	  | _ -> nest_handlers (n+1) event (List.rev l));
	  force_newline(); print_string "else ";
      | LINE(None, _, _) ->
	  raise (Error.Error "if: default found in the middle"))
    handler_lists;
  (match default with
    Some (Ast.EVENT(_,_,_,stmt,_)) -> A.pp_seq(stmt)
  | None ->
      G.print_nl_string ("panic(\""^(!G.policy_name)^
			 ": unknown event %x\\n\",_event);");
      print_string "return 0;")
    
let comp (Ast.EVENT(nm1,s1,_,_,_)) (Ast.EVENT(nm2,s2,_,_,_)) =
  compare (nm1,s1) (nm2,s2)
    
(* parser should check that event names are ok *)
let rec pp_handlers default handlers handler_env top_level_events
    scheduler_events name trace =
  let (attaches,detaches,compute_states,preempts,handlers) =
    List.fold_left
      (function (a,d,c,p,h) ->
	function (Ast.EVENT(name,specd,param,stmt,attr)) as event ->
	  match Ast.event_name2c name with
	    "attach" -> (event::a,d,c,p,h)
	  | "detach" -> (a,event::d,c,p,h)
	  | "compute_state" -> (a,d,event::c,p,h)
	  | "preempt" -> (a,d,c,event::p,h)
	  | _ -> (a,d,c,p,event::h))
      ([],[],[],[],[]) handlers in
  let topenv =
    List.map2 (function x -> function y -> (x,y))
      (if default (* should this be not default? *)
      then
	(List.filter
	   (function nm -> not(Ast.event_name2c2 nm = Some "process.new"))
	   (!Events.top_level_event_names))
      else (!Events.top_level_event_names))
      (!Events.specialized_events) in
  let tophandlers =
    let tlres =
      Events.top_level_event_names_loop
	(function prev -> function last ->
	  function specd_name -> function star -> function attr ->
	    let fspecd_name = List.flatten specd_name in
	    let hname = Ast.EVENT_NAME(fspecd_name,star,attr) in
	    let (_,(concretized_prev,handler_name)) =
	      List.find
		(function (h,c) -> Ast.event_name_equal hname h)
		topenv in
	    let concretized_handler_names =
	      Verifier.handler_name_lookup handler_env handler_name in
	    (prev@[last],
	     List.filter
	       (function Ast.EVENT(nm,specd,_,_,_) ->
		 Ast.event_names_equal specd concretized_prev &&
		 List.exists (Ast.event_name_equal nm)
		   concretized_handler_names)
	       handlers))
	(function _ -> ())
	top_level_events in
    List.map (function (x,_) -> x) tlres in
  let scheduler_events =
    if default
    then scheduler_events
    else (List.map
	    (List.filter
	       (function nm -> not(Ast.event_name2c2 nm = Some "process.new")))
	    scheduler_events) in
  let schedhandlers =
    Aux.make_set
      (List.flatten
	 (List.map
	    (function
		[] -> []
	      |	[scheddef] ->
		  let actual_names =
		    Verifier.handler_name_lookup handler_env scheddef in
		  List.filter
		    (function
			Ast.EVENT(nm,[],_,_,_) ->
			  List.exists (Ast.event_name_equal nm) actual_names
		      | _ -> false)
		    handlers
	      | _ -> internal_error "unexpected multi sched def")
	    scheduler_events)) in
  
  (* the main handler function *)
  let param = match handlers with
    (Ast.EVENT(nm,_,Ast.VARDECL(B.EVENT,id,imported,attr1),stmt,attr) :: _) ->
      id
  | _ ->
      internal_error
	"pp_handlers: handler list must not be empty, only one param allowed"
  in
  Ast2c.event_param := param;
  G.mk_static();
  print_string "int "; print_string name;
  print_string "_scheduler";
  G.pp_arglist true ["struct event_struct *"^(B.id2c param);
		     "struct next_list_struct *src_next_list";
		     "struct next_list_struct *tgt_next_list"];
  print_string " "; G.start_block();
  (match !B.schedtype with
    B.PROCSCHED ->
      G.print_nl_string
	((G.pp_bossa_struct_cs())^" *tgt = ("^(G.pp_bossa_struct_cs())^" *)"^
	 (B.id2c param)^"->target;")
  | B.VIRTSCHED ->
      (match trace with
	Ast.TRACE(n,events,exps,test,attr) ->
	  G.print_nl_string
	    ((G.pp_bossa_struct_cs())^" *tgt = tgt_next_list->sched;")
      | _ -> ()));
  G.print_nl_string ("uint _event = (uint)"^(B.id2c param)^"->type;");
  force_newline();
  if (!B.csharp)
  then G.print_nl_string "prefetch((int*)&state_info);"
  else G.print_nl_string "prefetch(&state_info);";
  let schedhandlers =
    (match !B.schedtype with
      B.PROCSCHED -> schedhandlers
    | B.VIRTSCHED ->
	List.map
	  (function (Ast.EVENT(nm,specd,param,stmt,attr) as x) ->
	    if List.exists (Ast.event_name_equal nm)
		(!Events.top_level_event_names)
	    then
	      let stmt =
		let Ast.VARDECL(_,param_name,_,_) = param in
		let tgt_decl =
		  Ast.VARDECL(B.PROCESS,B.mkId "tgt",false,attr) in
		let tgt_field =
		  Ast.FIELD(Ast.VAR(param_name,attr),B.mkId "target",
			    B.updty attr B.PROCESS) in
		Ast.add_decl_block
		  (Ast.VALDEF(tgt_decl,tgt_field,Ast.VARIABLE,attr))
		  stmt in
	      Ast.EVENT(nm,specd,param,stmt,attr)
	    else x)
	  schedhandlers) in
  nest_handlers 0 "_event" (List.rev (List.sort compare schedhandlers));
  if (!B.csharp) then print_string "return -1;";
  G.end_block(); force_newline(); force_newline();
  
  (* non-handlers *)
  pp_one_handler "int" "compute_state" compute_states
    ["struct next_list_struct *tgt_next_list"] false;
  pp_one_handler "void" "preempt" preempts
    (if (!B.csharp) then [""] else ["void"])  false;
  pp_one_handler "int" "ps_attach" attaches
    ["struct next_list_struct *tgt_next_list"] false;
  pp_one_handler "int" "ps_detach" detaches
    ["struct next_list_struct *tgt_next_list";"struct bossa_struct *tgt"]
    false;

  (* the top-level handlers *)
  
  let tgt_type =
    (match !B.schedtype with
      B.PROCSCHED -> G.pp_bossa_struct_cs()
    | B.VIRTSCHED -> "struct bossa_struct") in
  List.iter
    (function tophandlers ->
      List.iter
	(function (nms,hlist) ->
	  let cur_nm = List.hd (List.rev nms) in
	  let Ast.EVENT_NAME(top_names,_,_) = cur_nm in
	  let actual_snm = Ast.event_name2c2 cur_nm in
	  let actual_snm3 = Ast.event_name2c3 cur_nm in
	  let is_sched = actual_snm = Some "bossa.schedule" in
	  let is_unblock = actual_snm = Some "unblock.preemptive" in
	  let is_timer = actual_snm = Some "unblock.timer" in
	  let is_persistent_timer =
	    actual_snm3 = Some "unblock.timer.persistent" in
	  let all_names =
	    name^"_"^(String.concat "_" (List.map Ast.event_name2c_ nms)) in
	  G.mk_inline(); G.mk_static(); G.print_nl_string "int";
	  print_string all_names;
	  if is_sched
	  then G.mk_empty_prototype()
	  else if is_persistent_timer
	  then G.pp_arglist true ["int _event";
				   "struct timer_list *self";
				   "void *local_data";
				   "int target_alive";
				   (tgt_type^" *tgt");
				  "struct next_list_struct *tgt_next_list"]
	  else G.pp_arglist true ["int _event";(tgt_type^" *tgt");
				  "struct next_list_struct *tgt_next_list"];
	  print_string " "; G.start_block();
	  if is_unblock
	  then pp_unblock hlist name top_names
	  else
	    begin
	      let event_name = 
		if is_timer && !B.schedtype = B.VIRTSCHED
		then unblock_timer_adjustment()
		else "_event" in
	      if (!B.csharp)
	      then G.print_nl_string "prefetch((int*)&state_info);"
	      else G.print_nl_string "prefetch(&state_info);";
	      nest_handlers (List.length top_names) event_name
		(List.rev (List.sort compare hlist))
	    end;
	  G.end_block(); force_newline(); force_newline())
	tophandlers;
      match List.hd tophandlers with
	([entry_point],_) ->
	  let actual_snm = Ast.event_name2c_ entry_point in
	  let actual_snm2 = Ast.event_name2c2 entry_point in
	  let actual_snm3 = Ast.event_name2c3 entry_point in
	  let is_sched = actual_snm2 = Some "bossa.schedule" in
	  let is_timer = actual_snm2 = Some "unblock.timer" in
	  let is_persistent_timer =
	    actual_snm3 = Some "unblock.timer.persistent" in
	  let is_unblock = actual_snm2 = Some "unblock.preemptive" in
	  G.mk_static(); print_string ("int "^name^"_"^actual_snm^"_entry");
	  if is_sched
	  then G.mk_empty_prototype()
	  else if is_persistent_timer
	  then G.pp_arglist true ["int _event";
				   "struct timer_list *self";
				   "void *local_data";
				   "int target_alive";(tgt_type^" *tgt");
				   "struct next_list_struct *next_list"]
	  else if is_timer
	  then G.pp_arglist true ["int _event";(tgt_type^" *tgt");
				   "struct next_list_struct *next_list"]
	  else G.pp_arglist true ["int _event";(tgt_type^" *tgt")];
	  print_string " "; G.start_block();
	  G.print_nl_string "int res;";
	  if is_unblock then G.print_nl_string "int unblock_res;";
	  List.iter
	    (function (nms,_) ->
	      let all_names =
		name^"_"^
		(String.concat "_" (List.map Ast.event_name2c_ nms)) in
	      let is_sched =
		(Ast.event_name2c (List.hd (List.rev nms))) =
		"bossa.schedule" in
	      let call_args =
		if is_sched
		then "()"
		else if is_persistent_timer
		then "(_event,self,local_data,target_alive,tgt,next_list)"
		else if is_timer
		then "(_event,tgt,next_list)"
		else
		  B.sched_select
		    "(_event,tgt,null)"
		    "(_event,tgt,tgt->next_list->next->next)" in
	      if is_sched && not (all_names = (name^"_bossa_schedule"))
	      then
		(print_string "switch (res) ";
		 G.start_block();
		 if (!B.csharp) then
		   (G.print_nl_string
		      "case (int) SchedState.BLOCKED: running = get_rts_idle_process();";
		    G.print_nl_string "break;";
		    G.print_nl_string "case (int) SchedState.READY:")
		 else
		   (G.print_nl_string
		      "case BLOCKED: running = rts_idle_process;";
		    G.print_nl_string "break;";
		    G.print_nl_string "case READY:")
		);
	      if (!B.csharp)
	      then (G.print_nl_string ("res = "^all_names^call_args^";");
		    if is_sched && not (all_names = (name^"_bossa_schedule"))
		    then print_string "break;")
	      else print_string ("res = "^all_names^call_args^";");
	      if is_unblock
	      then
		(force_newline();
		 if (!B.csharp) then (
		   G.print_nl_string
		     "unblock_res = (int)(res&(1<<UNBLOCK_INFO_OFFSET));";
		   print_string
		     "res = (res&(int)(~(1<<UNBLOCK_INFO_OFFSET)));")
		 else (
		   G.print_nl_string "unblock_res = get_unblock_res(res);";
		   print_string "res = drop_unblock_res(res);"));
	      if is_sched && not (all_names = (name^"_bossa_schedule"))
	      then (G.end_block(); force_newline())
	      else force_newline())
	    tophandlers;
	  if is_unblock
	  then
	    (if (!B.csharp)
	    then G.print_nl_string "set_root_s_state((SchedState)res);"
	    else G.print_nl_string "root_s_state = res;";
	     print_string "return unblock_res;")
	  else print_string "return res;";
	  G.end_block();
	  force_newline();
	  force_newline()
      |	_ -> internal_error "unexpected entry point")
    tophandlers

and unblock_timer_adjustment _ =
  G.print_nl_string "uint __event;";
  G.print_nl_string "// ->prev below gets back to the current scheduler";
  G.print_nl_string
    "if (tgt_next_list->prev->sched->ops == &EXPORTS) __event = _event;";
  G.print_nl_string "else __event = 0;";
  "__event"

and unblock_mod_ret var = function
    Ast.IF(exp,stmt1,stmt2,attr) ->
      Ast.IF(exp,unblock_mod_ret var stmt1,unblock_mod_ret var stmt2,attr)
	
  | Ast.FOR(id,state_id,states,dir,stmt,attr) ->
      Ast.FOR(id,state_id,states,dir,unblock_mod_ret var stmt,attr)
	
  | Ast.FOR_WRAPPER(label,stmts,attr) ->
      Ast.FOR_WRAPPER(label,List.map (unblock_mod_ret var) stmts,attr)
	
  | Ast.SWITCH(exp,cases,default,attr) ->
      Ast.SWITCH(exp,
		 List.map
		   (function Ast.SEQ_CASE(ids,states,stmt,attr) ->
		     Ast.SEQ_CASE(ids,states,unblock_mod_ret var stmt,attr))
		   cases,
		 Aux.app_option (unblock_mod_ret var) default,
		 attr)
	
  | Ast.SEQ(decls,stmts,attr) ->
      Ast.SEQ(decls,List.map (unblock_mod_ret var) stmts,attr)
	
  | Ast.RETURN(Some(exp),attr) ->
      Ast.RETURN(Some(Ast.BINARY(Ast.BITOR,var,exp,attr)),attr)
	
  | stmt -> stmt
	
and pp_unblock u name top_names =
  let sv = B.fresh_id() in
  let attr = B.mkAttr 0 in
  let (tv,tvexp) =
    (match !B.schedtype with
      B.PROCSCHED ->
	let var = B.fresh_id() in (var,Ast.VAR(var,B.updty attr B.INT))
    | B.VIRTSCHED ->
	let var = B.mkId "fwdres"
	in (var,Ast.PRIM(Ast.VAR(B.mkId "get_unblock_res", attr),
			 [Ast.VAR(var,B.updty attr B.INT)],
			 attr))) in
  if !B.schedtype = B.PROCSCHED
  then G.print_nl_string ("int "^(B.id2c sv)^" = tgt->state;");
  if (!B.csharp) then (
    G.print_nl_string "int _temp;";
    (match !B.schedtype with
      B.PROCSCHED ->
        print_string "if (";
        G.print_between (function () -> print_string "||")
  	  (function st -> print_string ((B.id2c sv)^" == "); G.pp_state_cst st)
	  (Aux.option_filter
	     (function
	         (_,CS.STATE(id,CS.KERNEL_BLOCKED,_,_))
	       | (_,CS.STATE(id,CS.KERNEL_POLICY_BLOCKED,_,_)) -> Some id
	       | _ -> None)
	     (!CS.state_env));
	G.print_nl_string ") {";
	G.print_nl_string "    _temp = 1;";
	G.print_nl_string "}else{";
	G.print_nl_string "    _temp = 0;";
	G.print_nl_string "}";
	print_string ("int "^(B.id2c tv)^" = ");
        G.print_nl_string ("(_temp) << UNBLOCK_INFO_OFFSET;")
    | B.VIRTSCHED -> (print_string ("int "^(B.id2c tv)^" = "); G.print_nl_string "0;"));
   )else(
    (match !B.schedtype with
      B.PROCSCHED ->
	print_string ("int "^(B.id2c tv)^" = ");
	let options =
	  Aux.option_filter
	    (function
	        (_,CS.STATE(id,(CS.KERNEL_BLOCKED as c),_,_))
	      | (_,CS.STATE(id,(CS.KERNEL_POLICY_BLOCKED as c),_,_)) ->
		  B.check_policy_blocked (CS.class2c c); Some id
	      | (_,CS.STATE(id,CS.BLOCKED,_,_)) ->
		  B.check_blocked "BLOCKED"; Some id
	      | _ -> None)
	    (!CS.state_env) in
	(if options = []
	then raise (Error.Error "no states for UNBLOCK_INFO_OFFSET"));
        print_string "("; open_box 0;
        G.print_between
	  (function () -> print_space(); print_string "||"; print_space())
	  (function st -> print_string ((B.id2c sv)^" == "); G.pp_state_cst st)
	  options;
	close_box();
        G.print_nl_string (") << UNBLOCK_INFO_OFFSET;")
    | B.VIRTSCHED -> () (* fwdres already declared *));
    G.print_nl_string "prefetch(&state_info);";
   );
  let u =
    List.map
      (function Ast.EVENT(a,sp,b,s,c) ->
	Ast.EVENT(a,sp,b,unblock_mod_ret tvexp s,c))
      u in
  nest_handlers (List.length top_names) "_event"
    (List.rev (List.sort compare u))
    
and pp_one_handler ty str handlers args long_args =
  match handlers with
    [Ast.EVENT(_,_,_,stmt,_)] ->
      G.mk_static(); G.print_nl_string ty;
      print_string ((!G.policy_name)^"_"^str);
      G.pp_arglist long_args args;
      print_string " ";
      A.pp_seq stmt;
      force_newline();
      force_newline()
  | _ -> ()
	
(* -------------------------- proc interface ------------------------- *)
	
let rec type2printf id ty print_names =
  let ty_char =
    let rec loop ty =
      (match ty with
	B.ENUM(B.ID(str,_)) -> "d"
      | B.RANGE(_) -> "d"
      | B.SYSTEM(typ) -> loop typ
      | B.INT -> "d"
      | B.BOOL -> "s"
      | B.CYCLES -> "d"
      | B.TIME -> if !B.high_res then "ld.%-8ld" else "ld"
      | B.TIMER(_) -> "x"
      | B.PROCESS -> "d"
      | B.SCHEDULER -> "s"
      | B.EVENT -> "x"
      | B.INDR(t) -> "x"
      | B.STRUCT(B.ID(str,_)) -> "x"
      | B.PORT -> "x"
      | _ -> raise (Error.Error "not printable1")) in
    loop ty in
  "%" ^
  (match id with
    None -> ""
  | Some id ->
      match ty_char with
	"d" | "ld" -> if print_names then "-5.4" else "-5"
      | "ld.%-8ld" -> if print_names then "-11.10" else ""
      |	"x" -> if print_names then "-8.7" else "-8"
      |	_ ->
	  let idlen = String.length (B.id2c id) in
	  Printf.sprintf "-%d.%d" (idlen + 1) idlen) ^
  (if print_names then "s" else ty_char)

let rec type2printf_cs = function
    B.ENUM(B.ID(str,_)) -> str
  | B.RANGE(_) -> "int"
  | B.SYSTEM(typ) -> type2printf_cs typ
  | B.INT -> "int"
  | B.BOOL -> "string"
  | B.CYCLES -> "int"
  | B.TIME -> "int"
  | B.TIMER(_) -> "int*" (* ??? *)
  | B.PROCESS -> "int"
  | B.SCHEDULER -> "string"
  | B.EVENT -> "uint"
  | B.INDR(t) -> "int*"
  | B.STRUCT(B.ID(str,_)) -> (str^"*")
  | B.PORT -> "void*"
  | _ -> raise (Error.Error "not printable2")
	
	
let rec type2val str id = function
    B.SYSTEM(typ) -> type2val str id typ
  | B.BOOL ->
      [Printf.sprintf "(%s%s ? \"true\" : \"false\")" str (B.id2c id)]
  | B.INT | B.CYCLES | B.TIMER(_) | B.PORT | B.RANGE(_) |
    B.ENUM(B.ID(_,_)) | B.EVENT | B.INDR(_) |
    B.STRUCT(B.ID(_,_)) ->
      [Printf.sprintf "%s%s" str (B.id2c id)]
  | B.TIME ->
      if !B.high_res
      then [Printf.sprintf "%s%s.tv_sec" str (B.id2c id);
	     Printf.sprintf "%s%s.tv_nsec" str (B.id2c id)]
      else [Printf.sprintf "%s%s" str (B.id2c id)]
  | B.PROCESS ->
      [Printf.sprintf
	"((%s%s!=null) ? (%s)->pid : -1)" str (B.id2c id)
	(G.mk_attr (str^(B.id2c id)))]
  | B.SCHEDULER ->
      [Printf.sprintf "(%s%s ? %s%s->ops->name : \"NULL\")" str (B.id2c id)
	str (B.id2c id)]
  | _ -> raise (Error.Error "not printable3")

let print_decls_infos decls print_names res =
  print_string "size = sprintf(tmp, \"";
  (* printer directives *)
  List.iter
    (function (Ast.VARDECL(ty,id,imported,attr)) ->
      print_string " ";
      print_string (type2printf (Some id) ty print_names))
    decls;
  print_string "\\n\"";
  (* print data *)
  List.iter
    (function (Ast.VARDECL(ty,id,imported,attr)) ->
      print_string ","; force_newline();
      if print_names
      then print_string ("\""^(B.id2c id)^"\"")
      else
	G.print_between (function _ -> print_string ",") print_string
	  (type2val
	     (if imported
	     then (Printf.sprintf "(%s)->" (G.mk_attr "p"))
	     else "p->")
	     id ty))
    decls;
  print_string ");";
  force_newline();
  print_string ("buffer_extend(tmp,size,buffer,"^res^");")

let toformat n =
  let tabs = n / -8 in
  let rest = n mod -8 in
  let rec loop = function
      0 -> ""
    | 1 -> "\\t"
    | n -> "\\t" ^ (loop (n-1)) in
  (loop tabs) ^ (Printf.sprintf "%%%d" rest)

(* it isn't much use to print a timer, since it is just a pointer *)
let drop_timers l =
  List.filter
    (function
	Ast.VARDECL(B.TIMER(_),id,imported,attr) -> false
      |	_ -> true)
    l

let print_sched_infos decls print_names indent res =
  let do_indent1 =
    match indent with
      None -> ""
    | Some x -> Printf.sprintf "%ss" (toformat x) in
  let do_indent2 =
    match indent with
      None -> ""
    | Some x -> Printf.sprintf ", \"\"" in
  if print_names
  then
    (G.print_nl_string
      ("size = sprintf(tmp, \""^do_indent1^"%-15.15s%-5s\""^do_indent2^
       ", \"Name\", \"Type\");");
     G.print_nl_string ("buffer_extend(tmp,size,buffer,"^res^");"))
  else (
    if not (indent = None)
    then (print_string "if (first) "; G.start_block());
    G.print_nl_string
      ("size = sprintf(tmp, \"%-15.15s%-5s\", p->ops->name, sched_type[2+p->ops->type]);");
    print_string ("buffer_extend(tmp,size,buffer,"^res^");");
    if (indent = None)
    then force_newline()
    else
      (G.end_block(); force_newline(); print_string "else "; G.start_block();
       G.print_nl_string
	 ("size = sprintf(tmp, \""^do_indent1^
	  "%-15.15s%-5s\""^do_indent2^
	  ", p->ops->name, sched_type[2+p->ops->type]);");
       print_string ("buffer_extend(tmp,size,buffer,"^res^");");
       G.end_block();
       force_newline();
       G.print_nl_string "first = false;"));
  print_decls_infos (drop_timers decls) print_names res

let print_task_infos decls print_names indent =
  let do_indent1 =
    match indent with
      None -> ""
    | Some x -> Printf.sprintf "%ss" (toformat x) in
  let do_indent2 =
    match indent with
      None -> ""
    | Some x -> Printf.sprintf ", \"\"" in
  if print_names
  then
    (G.print_nl_string
      ("size = sprintf(tmp, \""^do_indent1^"%-5s%-9.8s%-6s\""^
       do_indent2^", \"Pid\", \"Command\", \"State\");");
     G.print_nl_string "buffer_extend(tmp,size,buffer,total_size);")
  else (
    if not (indent = None)
    then (print_string "if (first) "; G.start_block());
    G.print_nl_string
      ("size = sprintf(tmp, \"%-5d%-9.8s%-6d\",attr(p)->pid,attr(p)->comm,p->state);");
    print_string "buffer_extend(tmp,size,buffer,total_size);";
    if (indent = None)
    then force_newline()
    else
      (G.end_block(); force_newline(); print_string "else "; G.start_block();
       G.print_nl_string
	 ("size = sprintf(tmp, \""^do_indent1^
	  "%-5d%-9.8s%-6d\""^do_indent2^
	  ",attr(p)->pid,attr(p)->comm,p->state);");
       print_string "buffer_extend(tmp,size,buffer,total_size);";
       G.end_block();
       force_newline();
       G.print_nl_string "first = false;"));
  print_decls_infos (drop_timers decls) print_names "total_size"

let make_task_proc_info name decls max_state_len =
  G.mk_static(); print_string ("int "^name^"_do_taskprocinfo");
  G.pp_arglist true ["struct next_list_struct *next_list";
		     "struct bossa_struct *bossa";
		     "char *buffer"; "bool w_header"; "bool long_format";
		     "bool first"];
  print_string " ";
  G.start_block();
  if !B.schedtype = B.PROCSCHED
  then
    G.print_nl_string ((G.pp_bossa_struct_cs())^" * p = ("^
		       (G.pp_bossa_struct_cs())^" *) bossa;");
  G.print_nl_string "char * b = buffer;";
  G.print_nl_string "char tmp[500];";
  G.print_nl_string "int size;";
  force_newline();
  print_string "if (bossa == null) ";
  G.start_block();
  G.mk_printk (name^": Ups! Null process.\\n");
  print_string "return 0;";
  G.end_block();
  force_newline();
  (match !B.schedtype with
    B.PROCSCHED ->
      G.print_nl_string ("if ("^(G.mk_attr "p")^" == null) return 0;")
  | B.VIRTSCHED -> ());
  force_newline();
  (match !B.schedtype with
    B.PROCSCHED ->
      print_string "if (long_format) ";
      G.start_block();
      print_string "if (w_header) ";
      G.start_block();
      if (!B.csharp) then (
	let pp_id_nb_car id =
  	  if G.is_state id then (11 + String.length(B.id2c(id)))
	  else String.length (B.id2c(id))
	in
	let nb_caract = ref 21 in
        print_string("buffer += (sprintf(buffer,\"pid\\\\tcommand\\\\tstate");
        List.iter
 	  (function (Ast.VARDECL(ty,id,imported,attr)) ->
	    if ((!nb_caract) + (pp_id_nb_car id) + 3 > 60) then
	      (G.print_nl_string "\")/2);";
	       print_string "buffer += (sprintf(buffer,\""; nb_caract:=0;);
	    nb_caract := (!nb_caract) + (pp_id_nb_car id) + 3;
	    print_string "\\\\t"; G.pp_id id)
	  decls;
	G.print_nl_string "\")/2);";
        G.print_nl_string "buffer += (sprintf(buffer,\"\\\\n\")/2);";
        G.end_block();
	force_newline();
	G.print_nl_string
	  ("buffer += (sprintf(buffer,\"%d\\\\t%s\\\\t%d\",("^
	   (G.mk_attr "p")^")->pid,attr_comm(p),p->state)/2);");
	List.iter
	  (function (Ast.VARDECL(ty,id,imported,attr)) ->
	    print_string "buffer += (sprintf(buffer, \"";
	    print_string (type2printf (Some id) ty false);
	    print_string("\", ");
	    G.print_between (function _ -> print_string ",") print_string
	      (type2val (if imported then ((G.mk_attr "p")^"->") else "p->")
		 id ty);
	    G.print_nl_string ")/2);";
	  )
	  decls;
	G.print_nl_string "buffer += (sprintf(buffer, \"\\\\n\")/2);")
      else (
	G.print_nl_string("size = sprintf(tmp,\""^name^"\\n\");");
	G.print_nl_string "buffer_extend(tmp,size,buffer,total_size);";
	print_task_infos decls true None;
        G.end_block();
	force_newline();
	print_task_infos decls false (Some max_state_len));
      G.end_block();
      force_newline();
      print_string "else ";
      G.start_block();
      G.print_nl_string
	("size = sprintf(tmp,\"%d \",("^(G.mk_attr "p")^")->pid);");
      print_string "buffer_extend(tmp,size,buffer,total_size);";
      G.end_block();
      force_newline()
  | B.VIRTSCHED ->
      print_string "if (w_header) ";
      G.start_block();
      print_string("size = sprintf(tmp,\""^name^"\\nname\\tstate");
      List.iter
	(function (Ast.VARDECL(ty,id,imported,attr)) ->
	  print_string "\\t"; G.pp_id id)
	decls;
      G.print_nl_string "\\n\");";
      G.print_nl_string "buffer_extend(tmp,size,buffer,total_size);";
      print_string "size = sprintf(tmp,\"%s\t%d";
      List.iter
	(function (Ast.VARDECL(ty,id,imported,attr)) ->
	  print_string "\\t"; print_string (type2printf (Some id) ty false))
	decls;
      print_string "\\n\\n\",next_list->next->sched->ops->name,next_list->next->sched->state";
      List.iter
	(function (Ast.VARDECL(ty,id,imported,attr)) ->
	  print_string ","; force_newline();
	  G.print_between (function _ -> print_string ",") print_string
	    (type2val "next_list->next->sched->" id ty))
	decls;
      G.print_nl_string ");";
      print_string "buffer_extend(tmp,size,buffer,total_size);";
      G.end_block();
      force_newline();
      G.print_nl_string
	"next_list->next->sched->ops->do_taskprocinfo(next_list->next,bossa,buffer,w_header,long_format,true);");
  print_string "return strlen(b); // Should use strnlen ...";
  G.end_block();
  force_newline();
  force_newline();
  if (!B.csharp) then (
    G.print_nl_string "[MethodImplAttribute(MethodImplOptions.InternalCall)]";
    G.print_nl_string
      ("private static extern unsafe char * attr_comm("^
       (G.pp_bossa_struct_cs())^" * s);");
    G.print_nl_string "";
    G.print_nl_string "[MethodImplAttribute(MethodImplOptions.InternalCall)]";
    G.print_nl_string "private static extern unsafe int strlen(char* s);";
    G.print_nl_string "";
    G.print_nl_string "[MethodImplAttribute(MethodImplOptions.InternalCall)]";
    G.print_nl_string "private static extern unsafe int sprintf(char* s1, string s2);";
    G.print_nl_string "";
    G.print_nl_string "[MethodImplAttribute(MethodImplOptions.InternalCall)]";
    G.print_nl_string "private static extern unsafe int sprintf(char* s1, string s2, int i1);";
    G.print_nl_string "";
    G.print_nl_string "[MethodImplAttribute(MethodImplOptions.InternalCall)]";
    G.print_nl_string "private static extern unsafe int sprintf(char* s1, string s2, int pid, char* comm, int state);";
    let rec appartient e l = match l with
      [] -> false
    | t::q -> (e = t) || appartient e q
    in
    let type_fait = ref ["int"] in
    List.iter
      (function (Ast.VARDECL(ty,id,imported,attr)) ->
	if ((appartient (type2printf_cs ty) (!type_fait))=false) then
	  (type_fait := (!type_fait) @ [(type2printf_cs ty)];
	   G.print_nl_string "[MethodImplAttribute(MethodImplOptions.InternalCall)]";
	   print_string "private static extern unsafe int sprintf(char* s1, string s2, ";
	   print_string (type2printf_cs ty);
	   G.print_nl_string " argt);";
	   G.print_nl_string "");
      )
      decls;
    force_newline();
    force_newline();)
      
let make_queueToString name decls max_state_len =
  if (!B.csharp) then (
    G.print_nl_string "static private unsafe char* ";
    print_string
      ("queueToString(ref list_head start, char* buffer, bool long_format) ");
    G.start_block();
    G.print_nl_string ((G.pp_bossa_struct_cs())^" *p;");
    G.print_nl_string "list_head *elem;";
    G.print_nl_string "char tmp[500];";
    G.print_nl_string "int size;";
    force_newline();
    G.print_nl_string
      "if (list_empty(ref start)) { buffer += (sprintf(buffer,\"null\\\\n\")/2);}";
    print_string "else ";
    G.start_block();
    print_string "for (elem = start.next, prefetch((int*)elem->next); elem != &start;";
    G.print_nl_string " elem = elem->next , prefetch((int*)elem->next))";
    G.start_block();
    G.print_nl_string
      ("p = (("^(G.pp_bossa_struct_cs())^" *)((char *)(elem)-(ulong)((int)(&(("
       ^(G.pp_bossa_struct_cs())^" *)0)->list_info)/2)));");
    (match !B.schedtype with
      B.PROCSCHED ->
        print_string (name^"_do_taskprocinfo");
        G.pp_arglist true ["null";"(bossa_struct *)p";"buffer";
			   "elem==&start";"long_format";"true"];
        print_string ";";
        G.end_block();
        force_newline();
        print_string
	  "if (!long_format) { buffer += (sprintf(buffer, \"\\\\n\")/2);}"
    | B.VIRTSCHED ->
        G.print_nl_string
	  "buffer += sprintf(buffer, \"   %s \", p->ops->name);";
        G.end_block();
        force_newline();
        G.print_nl_string "buffer += sprintf(buffer, \"\\\\n\");")
   )else(
    G.print_nl_string "static char *";
    print_string
      ("queueToString(struct list_head *start, char *buffer, bool long_format) ");
    G.start_block();
    G.print_nl_string ((G.pp_bossa_struct_cs())^" *p;");
    G.print_nl_string "struct list_head *elem;";
    G.print_nl_string "char tmp[500];";
    G.print_nl_string "int size;";
    force_newline();
    print_string "if (list_empty(start)) "; G.start_block();
    G.print_nl_string "size = sprintf(tmp,\"null\\n\");";
    print_string "buffer_extend(tmp,size,buffer,buffer);";
    G.end_block(); force_newline();
    print_string "else ";
    G.start_block();
    G.print_nl_string "bool first = true;";
    print_string "list_for_each(elem, start) ";
    G.start_block();
    G.print_nl_string
      ("p = list_entry(elem, "^(G.pp_bossa_struct_cs())^", list_info);");
    (match !B.schedtype with
      B.PROCSCHED ->
        print_string ("buffer += "^name^"_do_taskprocinfo");
        G.pp_arglist true ["NULL";"(struct bossa_struct *)p";"buffer";
			   "elem==start";"long_format";"first"];
	G.print_nl_string ";";
        print_string "first = false;";
        G.end_block();
        force_newline();
        print_string
	  "if (!long_format) "; G.start_block();
	G.print_nl_string "size = sprintf(tmp, \"\\n\");";
	print_string "buffer_extend(tmp,size,buffer,buffer);";
	G.end_block()
    | B.VIRTSCHED ->
	print_sched_infos (drop_timers decls) false (Some max_state_len)
	  "buffer";
        G.end_block();
        force_newline();
        G.print_nl_string "size = sprintf(tmp, \"\\n\");";
	print_string "buffer_extend(tmp,size,buffer,buffer);"));
    G.end_block();
    force_newline();
    print_string "return buffer;";
    G.end_block();
    force_newline();
    force_newline()

let pp_buffer_extend _ =
  G.print_nl_string "static int xcount = 0;";
  G.print_nl_string "static int total_size = 0;";
  force_newline();
  G.print_nl_string "#define buffer_extend(tmp,size,buffer,ret) \\";
  G.print_nl_string "  do { \\";
  G.print_nl_string "    if (!xcount || size + total_size < xcount) { \\";
  G.print_nl_string "      total_size += size; \\";
  G.print_nl_string "      buffer = size + strcpy(buffer,tmp); \\";
  G.print_nl_string "    } \\";
  G.print_nl_string "    else { \\";
  G.print_nl_string "      return ret; \\";
  G.print_nl_string "    } \\";
  G.print_nl_string "  } while(0)";
  force_newline()

let make_do_schedprocinfo name states decls =
  print_string ("static int "^name^"_do_schedprocinfo(char *buffer, bool long_format) ");
  G.start_block();
  if !B.schedtype = B.VIRTSCHED
  then G.print_nl_string ((G.pp_bossa_struct_cs())^" *p;");
  G.print_nl_string "char *orig_buffer = buffer;";
  G.print_nl_string "char tmp[500];";
  G.print_nl_string "int size;";
  force_newline();

  G.print_nl_string
    ("size = sprintf(tmp,\"\\t\\t\\t"^name^
     "\\n\\t\\t\\t%s\\n\\n\",sched_type[EXPORTS.type]);");
  G.print_nl_string "buffer_extend(tmp,size,buffer,total_size);";
  if !B.schedtype = B.VIRTSCHED
  then G.print_nl_string "size = sprintf(tmp,\"Child schedulers:\\n\");"
  else G.print_nl_string "size = sprintf(tmp,\"Child processes:\\n\");";
  G.print_nl_string "buffer_extend(tmp,size,buffer,total_size);";

  let max_state_len =
    -1 *
    List.fold_left
      (function res ->
	function
	    Ast.QUEUE(_,_,id,_,_) ->
	      let len =
		String.length (Printf.sprintf "%s queue: " (B.id2c id)) in
	      if len > res then len else res
	  | Ast.PROCESS(_,id,_,_,_) ->
	      let len =
		String.length (Printf.sprintf "%s process: " (B.id2c id)) in
	      if len > res then len else res)
      0 states in
  
  (match !B.schedtype with
    B.PROCSCHED -> print_task_infos decls true (Some max_state_len)
  | B.VIRTSCHED ->
      print_sched_infos decls true (Some max_state_len) "total_size");
  force_newline();

  let print_label id ty =
    G.print_nl_string ((Printf.sprintf
			  "size = sprintf(tmp, \"%%%ds\",\""
			  max_state_len)
		       ^ (B.id2c id)^" "^ty^": \");");
    G.print_nl_string "buffer_extend(tmp,size,buffer,total_size);" in
  
  G.print_between force_newline
    (G.process_state
       (function array -> print_label array "queue")
       (function array ->
	 function index ->
	   print_string "buffer = queueToString(";
	   G.pp_arrayref_id array index;
	   G.print_nl_string ",buffer,long_format);")
       (function qid ->
	 print_label qid "queue";
	 print_string "buffer = queueToString(";
	 G.pp_id qid;
	 G.print_nl_string ",buffer,long_format);")
       (function vid ->
	 print_label vid (B.sched_select "process" "scheduler");
	 print_string "if (!";
	 G.pp_id vid;
	 print_string ") ";
	 G.start_block();
	 G.print_nl_string "size = sprintf(tmp, \"null\\n\");";
	 print_string "buffer_extend(tmp,size,buffer,total_size);";
	 G.end_block();
	 force_newline();
	 print_string "else ";
	 G.start_block();
	 (match !B.schedtype with
	   B.PROCSCHED ->
	     print_string ("buffer += "^name^
			   "_do_taskprocinfo(NULL,(struct bossa_struct *)");
	     G.pp_id vid;
	     G.print_nl_string ",buffer,false,long_format,true);";
	     print_string
	       "if (!long_format) "; G.start_block();
	     G.print_nl_string "size = sprintf(tmp, \"\\n\");";
	     print_string "buffer_extend(tmp,size,buffer,total_size);";
	     G.end_block()
	 | B.VIRTSCHED ->
	     print_string "p = "; G.pp_id vid; G.print_nl_string ";";
	     print_sched_infos decls false None "total_size");
	 G.end_block())
       (function pvid -> ()))
    states;
  force_newline();
  
(*
  if !B.schedtype = B.VIRTSCHED
  then
    (G.print_nl_string "size = sprintf(tmp, \"\\n\\n\");";
     List.iter
       (G.process_state
	  (function array -> ())
	  (function array ->
	    function index ->
	      print_string "buffer += ";
	      G.pp_arrayref_id array index;
	      G.print_nl_string "->ops->do_schedprocinfo(buffer,false);")
	  (function qid ->
	    print_string "list_for_each(elem, "; G.pp_id qid;
	    print_string ") ";
	    G.start_block();
	    G.print_nl_string
	      ("p = list_entry(elem, "
	       ^(G.pp_bossa_struct_cs())^", list_info);");
	    G.print_nl_string "buffer += p->ops->do_schedprocinfo(buffer,false);";
	    print_string "size = sprintf(tmp, \"\\n\\n\");";
	    G.end_block();
	    force_newline())
	  (function vid ->
	    print_string "if (";
	    G.pp_id vid;
	    G.print_nl_string ") ";
	    G.start_block();
	    print_string "buffer += ";
	    G.pp_id vid;
	    G.print_nl_string "->ops->do_schedprocinfo(buffer,false);";
	    print_string "size = sprintf(tmp, \"\\n\\n\");";
	    G.end_block();
	    force_newline())
	  (function pvid -> ()))
       states); *)
  
  print_string "return strlen(orig_buffer); // Should use strnlen ...";
  G.end_block();
  force_newline();
  force_newline();
  max_state_len

let make_do_schedprocinfo_cs name states =
  print_string ("static public int "^name^"_do_schedprocinfo(char* buffer, bool long_format) ");
  G.start_block();
  if !B.schedtype = B.VIRTSCHED
  then
    (G.print_nl_string ((G.pp_bossa_struct_cs())^" *p;");
     G.print_nl_string (G.de_structify "struct list_head *elem;"));
  G.print_nl_string "char* orig_buffer = buffer;";
  force_newline();
  
  G.print_nl_string ("buffer += (sprintf(buffer, \""^name^"\\\\n\")/2);");
  
  let print_label id ty =
    G.print_nl_string ("buffer += (sprintf(buffer, \""^(B.id2c id)^" "^ty^": \")/2);") in
  
  G.print_between force_newline
    (G.process_state
       (function array -> print_label array "queue")
       (function array ->
	 function index ->
	   print_string "buffer = queueToString(";
	   G.pp_arrayref_id array index;
	   G.print_nl_string ",buffer,long_format);")
       (function qid ->
	 print_label qid "queue";
	 print_string "buffer = queueToString(ref ";
	 G.pp_id qid;
	 G.print_nl_string ",buffer,long_format);")
       (function vid ->
	 print_label vid (B.sched_select "process" "scheduler");
	 print_string "if (";
	 G.pp_id vid;
	 G.print_nl_string "==null) ";
	 G.print_nl_string "  buffer += (sprintf(buffer, \"null\")/2);";
	 print_string "else ";
	 (match !B.schedtype with
	   B.PROCSCHED ->
	     print_string ("buffer += ("^name^
			   "_do_taskprocinfo(null,(bossa_struct *)");
	     G.pp_id vid;
	     G.print_nl_string ",buffer,true,long_format,true)/2);";
	     G.print_nl_string
	       "if (!long_format) buffer += (sprintf(buffer, \"\\\\n\")/2);"
	 | B.VIRTSCHED ->
	     print_string "buffer += \"%s\", ";
	     G.pp_id vid;
	     G.print_nl_string "->ops->name;";
	     G.print_nl_string "buffer += \"\\n\";"))
       (function pvid -> ()))
    states;
  force_newline();
(*
   if !B.schedtype = B.VIRTSCHED
   then
   (G.print_nl_string "size = sprintf(tmp, \"\\n\\n\");";
   List.iter
   (G.process_state
   (function array -> ())
   (function array ->
   function index ->
   print_string "buffer += ";
   G.pp_arrayref_id array index;
   G.print_nl_string "->ops->do_schedprocinfo(buffer,false);")
   (function qid ->
   print_string "list_for_each(elem, "; G.pp_id qid;
   print_string ") ";
   G.start_block();
   G.print_nl_string
   ("p = list_entry(elem, "
   ^G.pp_bossa_struct_cs()^", list_info);");
   G.print_nl_string "buffer += p->ops->do_schedprocinfo(buffer,false);";
   print_string "size = sprintf(tmp, \"\\n\\n\");";
   G.end_block();
   force_newline())
   (function vid ->
   print_string "if (";
   G.pp_id vid;
   G.print_nl_string ") ";
   G.start_block();
   print_string "buffer += ";
   G.pp_id vid;
   G.print_nl_string "->ops->do_schedprocinfo(buffer,false);";
   print_string "size = sprintf(tmp, \"\\n\\n\");";
   G.end_block();
   force_newline())
   (function pvid -> ()))
   states);
 *)
  print_string "return strlen(orig_buffer);";
  G.end_block();
  force_newline();
  force_newline()
    
    
(* --------------------- print trace information --------------------- *)
    
let pp_print_trace_info_decl = function
    Ast.NOTRACE -> ()
  | Ast.TRACE(n,events,exps,test,attr) ->
      print_string "static void print_trace_info(void);"
  | Ast.PRETRACE(_,_,_) -> raise (Error.Error "unexpected pretrace")
    
let pp_print_trace_info handlers = function
    Ast.NOTRACE -> ()
  | Ast.TRACE(n,events,exps,test,attr) ->
      (* event2c function *)
      print_string "static char *event2c(uint event) ";
      G.start_block();
      print_string "switch (event) ";
      G.start_block();
      let event_names =
	Aux.make_set(List.map (function Ast.EVENT(nm,_,_,_,_) -> nm) handlers)
      in
      List.iter
	(function event_name ->
	  if List.exists
	      (function traced_name ->
		Ast.event_name_prefix traced_name event_name)
	      events
	  then
	    (print_string ("case "^(Ast.event_name2mask event_name)^": ");
	     G.print_nl_string
	       ("return \""^(Ast.event_name2c event_name)^"\";")))
	event_names;
      print_string ("default: printk(\""^(!G.policy_name)^
		    ": unknown event %x\",event);");
      G.end_block();
      force_newline();
      print_string "return \"unknown\";";
      G.end_block();
      force_newline();
      force_newline();
      (* state2c function *)
      print_string "static char *state2c(int state) ";
      G.start_block();
      print_string "switch (state) ";
      G.start_block();
      List.iter
	(function
	    (_,CS.STATE(id,CS.NOWHERE,_,_)) -> ()
	  | (_,CS.STATE(id,cls,_,_)) ->
	      print_string "case "; G.pp_state_cst id;
	      print_string ": return ";
	      print_string ("\""^(B.id2c id)^"\"");
	      G.print_nl_string ";")
	(!CS.state_env);
      print_string "default: return \"unknown\";";
      G.end_block();
      G.end_block();
      force_newline();
      force_newline();
      (* print_trace_info function *)
      print_string "static void print_trace_info(void) ";
      G.start_block();
      G.print_nl_string "int i;";
      print_string "for (i=0; i!=trace_ctr; i++) ";
      G.start_block();
      print_string "printk";
      G.pp_arglist true
	(((B.sched_select "\"%d: %s(%x): %-13.13s (%d) state %s"
	     "\"%d: %s(%x): %-13.13s state %s")^
	  (String.concat ""
	     (List.map
		(function
		    Ast.TRACEVAR(fldid,id,attr) ->
		      " "^(B.id2c id)^" "^(type2printf None (B.ty attr) false)
		  | Ast.TRACEFIELD(id,attr) ->
		      " "^(B.id2c id)^" "^(type2printf None (B.ty attr) false)
		  | Ast.TRACEQUEUE(fldid,refid,attr) ->
		      " "^(B.id2c refid)^" "^(type2printf None (B.ty attr) false))
		exps))^"\\n\"")
	 ::
	   ["trace_buffer[i].jiffies";
	     "event2c(trace_buffer[i]._event)";
	     "trace_buffer[i].event2";
	     "trace_buffer[i].comm"] @
	 (B.sched_select ["trace_buffer[i].pid"] []) @
	 ["state2c(trace_buffer[i].state)"] @
	 (Aux.flat_map
	    (function
		Ast.TRACEVAR(fldid,id,attr) ->
		  type2val "trace_buffer[i]." fldid (B.ty attr)
	      | Ast.TRACEFIELD(id,attr) ->
		  type2val "trace_buffer[i]." id (B.ty attr)
	      | Ast.TRACEQUEUE(fldid,refid,attr) ->
		  ["trace_buffer[i]."^(B.id2c fldid)])
	    exps));
      print_string ";";
      G.end_block();
      force_newline();
      print_string "trace_ctr = 0;";
      G.end_block();
      force_newline();
      force_newline()
  | _ -> ()
	
(* ------------------------- print processes ------------------------- *)
	
let pp_print_states name states trace =
  G.mk_static_private();
  print_string ("void "^name^"_print_processes");
  G.mk_empty_prototype(); print_string " ";
  G.start_block();
  (match !B.schedtype with
    B.PROCSCHED ->
      G.print_between force_newline
	(G.process_state
	   (function array -> ())
	   (function array ->
	     function index ->
	       print_string "print_array_queue(\"";
	       G.pp_id array;
	       if (!B.csharp)
	       then print_string "\", ref "
	       else print_string "\", ";
	       G.pp_id index;
	       print_string ", ";
	       G.pp_arrayref_id array index;
	       print_string ");")
	   (function qid ->
	     print_string "print_queue(\""; G.pp_struct_id qid;
	     if (!B.csharp)
	     then print_string "\", ref "
	     else print_string "\", ";
	     G.pp_id qid; print_string ");")
	   (function vid ->
	     print_string "print_var(\""; G.pp_struct_id vid;
	     print_string "\", ";
	     G.pp_id vid; print_string ");")
	   (function pvid -> ()))
	states;
      (match trace with
	Ast.NOTRACE -> ()
      | Ast.TRACE(n,test,events,exps,attr) ->
	  G.print_nl_string "print_trace_info();"
      |	_ -> ())
  | B.VIRTSCHED -> ());
  G.end_block(); force_newline(); force_newline()
    
(* ------------------------- list sched info ------------------------- *)

let pp_list_sched_info name =
  G.mk_static_private(); print_string ("int "^name^"_list_sched_info");
  G.pp_arglist true ["char * page";
		     "char ** start";
		     if (!B.csharp) then "long off" else "off_t off";
		     "int count";
		     "int * eof";
		     "void * data"];
  print_string " ";
  G.start_block();
  G.print_nl_string "int res;";
  G.print_nl_string "ulong flags=0;";
  force_newline();
  G.print_nl_string "spin_lock_irqsave(&bossa_scheduler_lock,flags);";
  G.print_nl_string "xcount = count;";
  G.print_nl_string "total_size = 0;";
  G.print_nl_string ("res = "^name^"_do_schedprocinfo(page, true);");
  G.print_nl_string "xcount = 0;";
  G.print_nl_string "total_size = 0;";
  G.print_nl_string "spin_unlock_irqrestore(&bossa_scheduler_lock,flags);";
  print_string "return res;";
  G.end_block()
    
(* ------------------------- count processes ------------------------- *)
    
let pp_count_states name states =
  G.mk_static_private(); print_string ("int "^name^"_count_processes");
  G.mk_empty_prototype(); print_string " ";
  G.start_block();
  let total = B.id2c(B.fresh_idx "total") in
  G.print_nl_string ("int "^total^" = 0;");
  (match !B.schedtype with
    B.PROCSCHED ->
      G.print_between force_newline
	(G.process_state
	   (function array -> ())
	   (function array ->
	     function index ->
	       if (!B.csharp)
	       then print_string (total^" += count_queue(ref ")
	       else print_string (total^" += count_queue(");
	       G.pp_arrayref_id array index; print_string ");")
	   (function qid ->
	     if (!B.csharp)
	     then print_string (total^" += count_queue(ref ")
	     else print_string (total^" += count_queue(");
	     G.pp_id qid; print_string ");")
	   (function vid ->
	     print_string (total^" += count_var(");
	     G.pp_id vid; print_string ");")
	   (function pvid -> ()))
	states;
      print_string ("return "^total^";"); close_box()
  | B.VIRTSCHED ->
      print_string "return 0;");
  G.end_block(); force_newline(); force_newline()
    
let pp_misc nm states procdecls defs =
  (* attach self *)
  G.print_nl_string
    "/* executed when the scheduler is added to the hierarchy */";
  G.mk_static();
  print_string ("void "^nm^"_attach_self(void) ");
  G.start_block();
  if not (defs = [])
  then G.print_nl_string "/* initialization of timer variables */";
  List.iter
    (function
	Ast.SYSDEF((Ast.VARDECL(B.TIMER(_),id,_,_)),_,attr) ->
	  print_string ((B.id2c id)^" = create_static_bossa_timer");
	  G.pp_arglist true
	    ["NULL";"EXPORTS.next_list.next->next"; (* one past top sched *)
	     "&_"^(B.id2c id)^"_timer";
	     "&_"^(B.id2c id)^"_timer_data";
	     "EVENT_UNBLOCK_TIMER_NOTARGET_"^(String.uppercase(B.id2c id))];
	  G.print_nl_string ";"
      | _ -> ())
    defs;
  G.end_block();
  force_newline();
  force_newline();
  (* detach self *)
  G.mk_static();
  if (!B.csharp)
  then print_string ("scheduler_struct* "^nm^"_detach_self(bool is_root) ")
  else print_string ("PROTOCOL_SCHEDULER_STRUCT *"^nm^"_detach_self(bool is_root) ");
  G.start_block();
  let detach_self _ =
    G.print_nl_string
      "if (EXPORTS.next_list.prev->prev != &(EXPORTS.next_list))";
    if (!B.csharp) then
      G.print_nl_string
	("  EXPORTS.next_list.prev->prev->sched->ops->detach(("^
	 (G.pp_bossa_struct_cs())^" *)EXPORTS.next_list.prev->sched,1);")
    else
      G.print_nl_string
	"  EXPORTS.next_list.prev->prev->sched->ops->detach(EXPORTS.next_list.prev->sched,true);" in
  let destroy_timers() =
    List.iter
      (function
	  Ast.SYSDEF((Ast.VARDECL(B.TIMER(_),id,_,_)),_,attr) ->
	    G.print_nl_string("delete_static_bossa_timer("^(B.id2c id)^");")
	| _ -> ())
      defs in
  let fail els =
    if els then print_string "else ";
    G.start_block();
    G.mk_printk("cannot detach "^nm^"\\n");
    if (!B.csharp) then (
      G.print_nl_string "set_errno(-_EINVAL);";
      print_string "return (scheduler_struct *)FAILURE;")
    else (
      G.print_nl_string "errno = -EINVAL;";
      print_string "return (PROTOCOL_SCHEDULER_STRUCT *)FAILURE;");
    G.end_block() in
  (match !B.schedtype with
    B.VIRTSCHED ->
      print_string "if (children == 0) ";
      G.start_block();
      destroy_timers();
      detach_self();
      print_string "return NULL;";
      G.end_block();
      force_newline();
      print_string "else if (is_root && children == 1) ";
      G.start_block();
      destroy_timers();
      G.print_nl_string (G.de_structify "struct list_head *elem;");
      G.print_nl_string "PROTOCOL_SCHEDULER_STRUCT *child;";
      force_newline();
      (* have to find the child *)
      List.iter
	(G.process_state
	   (function array -> ())
	   (function array ->
	     function index ->
	       G.start_block();
	       print_string "child = ";
	       G.pp_arrayref_id array index;
	       G.print_nl_string ";";
	       G.print_nl_string (nm^"_detach(child,false);");
	       print_string "return child;";
	       G.end_block();
	       force_newline())
	   (function qid ->
	     print_string "list_for_each(elem, "; G.pp_id qid;
	     print_string ") ";
	     G.start_block();
	     G.print_nl_string
	       ("child = list_entry(elem, "
		^(G.pp_bossa_struct_cs())^", list_info);");
	     G.print_nl_string (nm^"_detach(child,false);");
	     print_string "return child;";
	     G.end_block();
	     force_newline())
	   (function vid ->
	     print_string "if (";
	     G.pp_id vid;
	     print_string ") ";
	     G.start_block();
	     print_string "child = ";
	     G.pp_id vid;
	     G.print_nl_string ";";
	     G.print_nl_string (nm^"_detach(child,false);");
	     print_string "return child;";
	     G.end_block();
	     force_newline())
	   (function pvid -> ()))
	states;
      fail false;
      G.end_block();
      force_newline();
      fail true
  | B.PROCSCHED ->
      G.print_nl_string ("if (!is_root && "^nm^
		    "_count_processes() == 0\n#ifdef MODULE");
      G.print_nl_string " && !MOD_IN_USE\n#endif";
      print_string ")";
      G.start_block();
      destroy_timers();
      detach_self();
      print_string "return null;";
      G.end_block();
      force_newline();
      fail true);
  force_newline();
  print_string "return null;";
  G.end_block();
  force_newline();
  force_newline();
  (* set_priority *)
  match !B.schedtype with
    B.PROCSCHED ->
      G.mk_static();
      print_string ("void "^nm^"_set_priority");
      G.pp_arglist false [((G.pp_bossa_struct_cs())^" *p");
		    "struct next_list_struct *nl"; "int niceval"];
      print_string " "; G.start_block();
      if List.exists
	  (function
	      Ast.VARDECL(ty,id,_,attr) ->
		(B.id2c id = "priority" && ty = B.INT))
	  procdecls
      then print_string "p->priority = 20-niceval;"
      else G.mk_printk "scheduler has no priority field\\n";
      G.end_block(); force_newline()
  | B.VIRTSCHED ->
      G.mk_static();
      print_string ("void "^nm^"_set_priority");
      G.pp_arglist false [("struct bossa_struct *p");
			  "struct next_list_struct *next"; "int niceval"];
      print_string " "; G.start_block();
      print_string
	"return next->sched->ops->set_priority(p,next->next,niceval);";
      G.end_block(); force_newline()
	
(* ---------------------------- modules ----------------------------- *)
	
let pp_module_functions nm =
  if (!B.csharp) then (
    let strc = B.sched_select "Process" "Virtual" in
    G.print_nl_string("const string scheduler_string = \""^strc^" Scheduler with "^nm^"\";");
    G.print_nl_string "const string scheduler_version = \"1.0.1\";";
    G.print_nl_string "const string scheduler_copyright = \"Copyright (c) 2002 INRIA, EMN, Univerity of Copenhagen.\";";
    force_newline();
    G.print_nl_string ("static public int Sched"^nm^"_init_module() ");
    G.start_block();
    G.print_nl_string "printk(\"<6>%s - version %s\\\\n\", scheduler_string, scheduler_version);";
    G.print_nl_string "printk(_KERN_INFO + \"%s\\\\n\", scheduler_copyright);";
    G.print_nl_string "// add to all_schedulers";
    G.print_nl_string "bossa_add_scheduler(ref EXPORTS,false);";
    G.print_nl_string "// init devfs thingy";
    G.print_nl_string "EXPORTS.ioctl_init();";
    print_string "return 0;";
    G.end_block();
    force_newline();
    G.print_nl_string "\n";
    G.print_nl_string ("static public void Sched"^nm^"_cleanup_module() ");
    G.start_block();
    G.print_nl_string "// unregister ioctl";
    G.print_nl_string "fixed(next_list_struct* ptr_EXPORTS_next_list = &EXPORTS.next_list)";
    G.print_nl_string "fixed(___scheduler_operations* ptr_EXPORTS = &EXPORTS) {";
    G.print_nl_string "unregister();";
    G.print_nl_string "// remove the scheduler from the parent, if any";
    G.print_nl_string "if (EXPORTS.next_list.prev->prev != ptr_EXPORTS_next_list)";
    print_string "  EXPORTS.next_list.prev->prev->sched->ops->detach((";
    print_string (G.pp_bossa_struct_cs());
    G.print_nl_string " *)EXPORTS.next_list.prev->sched,1);";
    G.print_nl_string "// remove the scheduler from the list of all schedulers";
    G.print_nl_string "bossa_remove_scheduler(ptr_EXPORTS);";
    G.print_nl_string "";
    G.print_nl_string "printk(_KERN_INFO + \"scheduler %s unloaded\\\\n\",EXPORTS.name);";
    print_string "}";
    G.end_block();
    force_newline();
    
   )else(
    G.print_nl_string "/*******************************************************";
    G.print_nl_string " *";
    G.print_nl_string " *	Modification Module";
    G.print_nl_string " *";
    G.print_nl_string " * *****************************************************/";
    G.print_nl_string "#ifdef MODULE";
    G.print_nl_string "   #include <linux/module.h> /* Ncessaire pour tous les modules */";
    G.print_nl_string "   #include <linux/kernel.h> /* Ncessaire pour le alertes noyau */";
    G.print_nl_string "   #include <linux/init.h>   /* Ncessaire pour l'utilisation module_init()/_exit()*/";
    G.print_nl_string "   #include <linux/config.h> /* ?? */";
    G.print_nl_string "static void unregister(void);";
    let strc = B.sched_select "Process" "Virtual" in
    G.print_nl_string
      ("static char scheduler_string[] = \""^strc^" Scheduler with "^nm^"\";");
    G.print_nl_string "static char scheduler_version[] = \"1.0.1\";";
    G.print_nl_string "static char scheduler_copyright[] = \"Copyright (c) 2002 INRIA, Ecole des Mines de Nantes, Univerity of Copenhagen.\";";
    G.print_nl_string "";
    G.print_nl_string "MODULE_AUTHOR(\"Bossa Compiler\");";
    let strc = B.sched_select "process" "virtual" in
    G.print_nl_string ("MODULE_DESCRIPTION(\"Loading of "^strc^" scheduler "^nm^"\");");
    G.print_nl_string "MODULE_LICENSE(\"GPL\");";
    force_newline();
    G.print_nl_string "int bossa_add_scheduler(struct scheduler_operations *u_ops, int stat);";
    G.print_nl_string "void bossa_remove_scheduler(struct scheduler_operations *u_ops);";
    force_newline();
    G.print_nl_string "/******************************************************************************";
    G.print_nl_string (" * "^nm^"_init_module - Scheduler Registration Routine");
    G.print_nl_string " *";
    G.print_nl_string (" * "^nm^"_init_module is the first routine called when the scheduler is");
    G.print_nl_string " * loaded. All it does is add to the list of scheduler and register with the";
    G.print_nl_string " * ioctl subsystem.";
    G.print_nl_string " * ";
    G.print_nl_string " ***************************************************************************/";
    G.print_nl_string ("static int __init "^nm^"_init_module(void) ");
    G.start_block();
    G.print_nl_string "printk(\"<6>%s - version %s\\n\", scheduler_string, scheduler_version);";
    G.print_nl_string "printk(KERN_INFO \"%s\\n\", scheduler_copyright);";
    G.print_nl_string
      ("printk(KERN_INFO \""^nm^"_scheduler_init function at %x\\n\", "
       ^"(unsigned int)"^nm^"_scheduler_init);");
    G.print_nl_string "// add to all_schedulers";
    G.print_nl_string "if (bossa_add_scheduler(&EXPORTS,0)) return EINVAL;";
    G.print_nl_string "// init devfs thingy";
    G.print_nl_string "EXPORTS.ioctl_init();";
    print_string "return 0;";
    G.end_block();
    force_newline();
    force_newline();
    G.print_nl_string ("module_init("^nm^"_init_module);");
    G.print_nl_string "/******************************************************************************";
    G.print_nl_string (" * "^nm^"_cleanup_module - Scheduler Registration Routine");
    G.print_nl_string " *";
    G.print_nl_string (" * "^nm^"_cleanup_module is the first routine called when the scheduler is");
    G.print_nl_string " * unloaded. For instance she does nothing.";
    G.print_nl_string " * ";
    G.print_nl_string " ***************************************************************************/";
    G.print_nl_string ("static void __exit "^nm^"_cleanup_module(void) ");
    G.start_block();
    G.print_nl_string "// unregister ioctl";
    G.print_nl_string "unregister();";
    G.print_nl_string "// remove the scheduler from the parent, if any";
    G.print_nl_string "if (EXPORTS.next_list.prev->prev != &EXPORTS.next_list)";
    G.print_nl_string "  EXPORTS.next_list.prev->prev->sched->ops->detach(EXPORTS.next_list.prev->sched,true);";
    G.print_nl_string "// remove the scheduler from the list of all schedulers";
    G.print_nl_string "bossa_remove_scheduler(&EXPORTS);";
    G.print_nl_string "";
    print_string "printk(KERN_INFO \"scheduler %s unmounted\\n\",EXPORTS.name);";
    G.end_block();
    force_newline();
    force_newline();
    G.print_nl_string ("module_exit("^nm^"_cleanup_module);");
    G.print_nl_string "#endif";
    force_newline();
    G.print_nl_string ("static inline void increment_ref_count(void) {");
    G.print_nl_string "#ifdef MODULE";
    G.print_nl_string "  MOD_INC_USE_COUNT;";
    G.print_nl_string "#endif";
    G.print_nl_string "}";
    force_newline();
    G.print_nl_string ("static inline void decrement_ref_count(void) {");
    G.print_nl_string "#ifdef MODULE";
    G.print_nl_string "  MOD_DEC_USE_COUNT;";
    G.print_nl_string "#endif";
    G.print_nl_string "}";
    force_newline();
    G.print_nl_string "/******* Fin de modification pour insertion modulaire *******/";
   );
  force_newline()
    
let pp_module_functions2 nm =
  if (!B.csharp) then (
    print_string "static void unregister() ";
    G.start_block();
    G.print_nl_string "devfs_unregister_chrdev(major, _BOSSA_DEVICE_NAME);";
    print_string "devfs_unregister(devfs_handle2);";
    G.end_block();
    force_newline();
   ) else (
    G.print_nl_string "#ifdef MODULE";
    print_string "static void unregister(void) ";
    G.start_block();
    G.print_nl_string "devfs_unregister_chrdev(major, BOSSA_DEVICE_NAME);";
    print_string "devfs_unregister(devfs_handle2);";
    G.end_block();
    force_newline();
    G.print_nl_string "#endif"
   )
(* -------------------------- entry point --------------------------- *)
      
let mkdir_if_necessary dirname =
  if Sys.file_exists dirname
  then
    Unix.access dirname [Unix.W_OK; Unix.R_OK; Unix.X_OK]
  else
    Unix.mkdir dirname 0o755
      
      
let bcpp dirname genname
    ((Ast.SCHEDULER(nm,default,cstdefs,enums,procdecls,fundecls,valdefs,states,
		    criteria,_,trace,handlers,ifunctions,functions,_,
		      attr)) as x) =
  let top_level_events = Events.top_level_events() in
  let scheduler_events = Event_types.scheduler_events() in
  G.init_external_functions();
  A.current_phase := A.HANDLERS;
  mkdir_if_necessary dirname;
  let file_ext =
    if (!B.csharp)
    then ".cs"
    else ".c"
  in
  (match genname with
    None -> set_formatter_out_channel (open_out (dirname^"/"^nm^file_ext))
  | Some nm -> set_formatter_out_channel (open_out (dirname^"/"^nm^file_ext)));
  set_margin 80;
  G.policy_name := nm;
  G.print_nl_string copyright;
  let checksum = Checksum.checksum ifunctions in
  print_checksum checksum;
(*  if (!B.csharp) then
    G.print_nl_string (Bossa2csharp.cs_specific nm);*)
  
  G.print_nl_string (make_exports nm);
  if not (!B.csharp) then
    print_string (make_structure_prototypes nm);
(*  if (!B.csharp) then
    G.print_nl_string Bossa2csharp.includefiles_cs
  else*)
    G.print_nl_string (includefiles ());
  if not (!B.csharp) then
    Compile_interface.interface_includes();
  G.print_nl_string (usefulmacros trace);
  open_box 0;
  pp_defs true cstdefs true;
  pp_typedefs enums;
  pp_procdecls procdecls;
  pp_fundecls fundecls;
  pp_print_trace_info_decl trace;
  pp_defs true valdefs false;
  if not (!B.csharp) then
    pp_prototypes top_level_events nm;
  if not (!B.csharp) then
    G.print_nl_string (operationstructure top_level_events nm default);
  pp_states states;
  pp_trace_struct trace;
  if (!B.csharp) then pp_module_functions2 nm;
  pp_module_functions nm;
  if not (!B.csharp) then
    G.print_nl_string "#include \"prim1.c\"\n";
  pp_functions functions;
  pp_check_process_state nm;
  pp_declare_bitmap();
  if not (!B.csharp) then
    G.print_nl_string "#include \"prim.c\"\n";
  if (!B.schedtype = B.VIRTSCHED)
  then
    (G.print_nl_string "#include \"vsprim.c\"\n";
     pp_create_scheduler nm (!CS.state_env);
     pp_create_forward nm);
  pp_init_states nm states;
  B.type_return := "int";
  pp_handlers default handlers (!Verifier.handler_environment)
    top_level_events scheduler_events nm trace;
(*  pp_functions ifunctions;*)
  pp_buffer_extend();
  let max_state_len =
    if (!B.csharp)
    then (make_do_schedprocinfo_cs nm states; 0)
    else make_do_schedprocinfo nm states procdecls in
  make_queueToString nm procdecls max_state_len;
  make_task_proc_info nm procdecls max_state_len;
  pp_print_trace_info handlers trace;
  pp_list_sched_info nm;
  force_newline();
  force_newline();
  pp_print_states nm states trace;
  pp_count_states nm states;
  pp_misc nm states procdecls valdefs;
  close_box();
  force_newline();
  Compile_interface.compile_interface x checksum;
  if not (!B.csharp) then pp_module_functions2 nm;
(*  if (!B.csharp) then Bossa2csharp.cs_specific_end nm valdefs;*)
  B.csharp := true;
  (match genname with
    None ->
      set_formatter_out_channel (open_out (dirname^"/user_stub_"^nm^".cs"))
  | Some nm ->
      set_formatter_out_channel (open_out (dirname^"/user_stub_"^nm)));
  Compile_interface.ucompile_interface x checksum;
  B.csharp := false;
  (match genname with
    None ->
      set_formatter_out_channel (open_out (dirname^"/user_stub_"^nm^".c"))
  | Some nm ->
      set_formatter_out_channel (open_out (dirname^"/user_stub_"^nm)));
  Compile_interface.ucompile_interface x checksum;
  (match genname with
    None ->
      set_formatter_out_channel (open_out (dirname^"/user_stub_"^nm^".h"))
  | Some nm ->
      set_formatter_out_channel(open_out (dirname^"/user_stub_"^nm^".h")));
  Compile_interface.hcompile_interface x
