(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: main.ml,v 1.15 2004/11/11 16:15:00 julia Exp $ *)

(*open Devilast*)
open Lexing
open Format
open Bossa_lexer	(* to get cr_index and lineno *)
open List

let input_name = ref ([] : string list)	(* no filename by default *)
let nocode = ref false
let output_name = ref "out" (* default output name *)
let dirname = ref "c-code"  (* default output directory *)
let genname = ref None      (* default generated code name *)

exception  Err of string ;;


let fatal_error s = Printf.eprintf "%s: %s\n"
    (Filename.basename Sys.argv.(0)) s;
  exit 2


let version =
  "This Software is (C) copyright INRIA - 2001.\n\
    Alpha 0.0.1 (12/10/2001)\n"


let print_version () =
  print_string version;
  exit 0


let print_pos fname line colfrom colto =
  sprintf "File \"%s\", line %d, characters xxx %d-%d:\n"
    fname line colfrom colto


let pr x y = Printf.sprintf "%s.%s" x y

let go input_name output_name dirname genname =
  Objects.init_id();
  let (ast,ty) =
    match input_name with
      [] ->
	let lexbuf = Lexing.from_channel stdin in
	(try
	  Misc.init "stdin";
	  Bossa_lexer.init();
	  Bossa_parser.bossa_spec Bossa_lexer.token lexbuf
	with x ->
	  (Parser_error.report_error lexbuf Format.err_formatter x;
	   exit 2))
    | [real_input_name] ->
	let ch = open_in real_input_name in
	let lexbuf = Lexing.from_channel ch in
	(try
	  Misc.init (Filename.basename real_input_name);
	  Bossa_lexer.init();
	  let res = Bossa_parser.bossa_spec Bossa_lexer.token lexbuf in
	  close_in ch;
	  res
	with x ->
	  (Parser_error.report_error lexbuf Format.err_formatter x;
	   exit 2))
    | _ ->
	let masts_types =
	  List.map
	    (function real_input_name ->
	      let ch = open_in real_input_name in
	      let lexbuf = Lexing.from_channel ch in
	      let name_cell = ref (Objects.MODULE_NAME (Objects.mkId "")) in
	      Objects.module_name := Some name_cell;
	      (try
		Module_misc.init (Filename.basename real_input_name);
		Module_bossa_lexer.init();
		let (res,mod_name) = Module_bossa_parser.module_spec
		    Module_bossa_lexer.token lexbuf in
		close_in ch;
		name_cell := mod_name;
		res
	      with x ->
		(Module_parser_error.report_error lexbuf
		   Format.err_formatter x;
		 exit 2)))
	    input_name in
	let (_,scheduler_type) =
	  try
	    List.find
	      (function (Mast.SCHED(_),_) -> true | (Mast.MODL(_),_) -> false)
	      masts_types
	  with Not_found -> raise (Error.Error "no scheduler found") in
	let (masts,_) = List.split masts_types in
	Error.init_error();
	let merged = Module_builder.builder masts in
	Pp.pretty_print false (pr output_name "md") merged;
	Error.check_errors("module processing");
	(merged, scheduler_type) in
  Objects.schedtype := ty;
  Error.init_error();
  Printf.printf("type checking\n");
  flush stdout;
  let type_checked = Type.type_check ast in
  (*let type_checked1 = Mutate.mutate type_checked in*)
  let type_checked1 = type_checked in
  Pp.pretty_print false (pr output_name "tc") type_checked1;
  Error.check_errors("type checking");
  Error.init_error();
  Printf.printf("verifying\n");
  flush stdout;
  let verified = Verifier.verify type_checked1 in
  Pp.pretty_print true (pr output_name "vr") verified;
  if !Objects.blast
  then
    begin
      Printf.printf "printing for Blast\n";
      let compiled = Compile_misc.compile_misc verified in
      Bossa2c.bcpp dirname genname compiled
    end
  else
    begin
      Printf.printf("postprocessing\n");
      flush stdout;
      let compiled = Compile.compile verified in
      Pp.pretty_print false (pr output_name "end") compiled;
      Printf.printf("C code generating\n");
      flush stdout;
      let genname =
	if Error.any_errors()
	then
	  (match genname with
	    Some x -> Some ("error_"^x)
	  | 	 _ -> Some ("error_"^(List.hd input_name)))
	else genname in
      (*if (!Objects.hls)
      then Bossa2c_hls.bcpp dirname genname compiled
      else*) Bossa2c.bcpp dirname genname compiled;
      Error.check_errors("verification")
    end
      
      
(* Parsing of command-line arguments *)

exception Found_program_name

let set_output s = output_name := s

let anonymous s = input_name := s :: !input_name

let set_dir s = dirname := s
let set_gen s = genname := Some s

let speclist =
  [ (*"-csharp", Arg.Set Objects.csharp,
    "The code will be generated in C#, not in C";

    "-hls", Arg.Set Objects.hls,
    "The code will be generated for the HLS framework";

    "-blast", Arg.Set Objects.blast,
    "The code will be generated for use with Blast";*)

    "-o", Arg.String set_dir,
    "<dirname>  Set the name of the output directory";

    "-hrt", Arg.Set Objects.high_res,
    "Generate code for high-res timers";

    "-code", Arg.String set_gen ,
    "<name> Set the name of generated code";

    "-debug", Arg.String set_output ,
    "<name> Set the name of debugging files";

    "-os", Arg.String Events.set_os ,
    "Select OS";

    "-version", Arg.Unit print_version,
    "Display version info";

    "-m", Arg.Set Objects.mutate,
    "Perform a random mutation after type checking and before verification";

    "-pb", Arg.Set Objects.policy_blocked_types,
    ("Use event types that distinguish between POLICY_BLOCKED "^
     "and KERNEL_BLOCKED")
  ]

let usage =
  Printf.sprintf "Usage: %s [options] <filename> \nOptions are:"
    (Filename.basename Sys.argv.(0))

let main () =
  Arg.parse speclist anonymous usage;
  go !input_name !output_name !dirname !genname


let _ =
  try
    Unix.handle_unix_error main ()
  with
    Sys_error(s)     -> fatal_error s
  | Failure(s)       -> fatal_error s
  | Error.Error(s)   -> fatal_error s
