(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* -------------------- The scheduler structure ------------------- *)
val list_of_const_enum : string list ref

type scheduler =
  SCHEDULER of
    string                (* name *)
    * bool                (* default scheduler *)
    * valdef list         (* constant declarations *)
    * typedef list        (* programmer declared types *)
    * vardecl list        (* fields of the process structure *)
    * fundecl list        (* external function declarations *)
    * valdef list         (* (non-constant) variable declarations *)
    * state list          (* states *)
    * criteria list       (* ordering criteria *)
    * admit option        (* admission criteria *)
    * trace               (* tracing declaration *)
    * event_decl list     (* event handlers *)
    * fundef list         (* interface functions *)
    * fundef list         (* verifier-defined functions *)
    * aspect_info         (* aspects *)
    * Objects.attribute

and typedef =
  ENUMDEF of
    Objects.identifier          (* the name of the declared type *)
    * Objects.identifier list   (* the list of possible values *)
    * Objects.attribute
  | ENUMVALDEF of
    Objects.identifier          (* the name of the declared type *)
    * (Objects.identifier * expr) list (* the list of possible values *)
    * Objects.attribute
  | RANGEDEF of
    Objects.identifier          (* the name of the declared type *)
    * expr                      (* the lower bound *)
    * expr                      (* the upper bound *)
    * Objects.attribute
  | TIMERDEF of
    Objects.identifier          (* the name of the declared type *)
    * vardecl list              (* fields of the timer structure *)
    * Objects.attribute

and vardecl =
  VARDECL of
    Objects.typ                 (* the type of the declared identifier *)
    * Objects.identifier        (* the name of the declared identifier *)
    * bool                      (* true if imported from the system *)
    * Objects.attribute

(* useful to avoid a parser problem *)
and pre_valdef = FN of fundecl | VL of valdef

and fundecl =
  FUNDECL of
    Objects.typ                 (* the return type *)
    * Objects.identifier        (* the name of the declared function *)
    * Objects.typ list          (* the types of the arguments *)
    * Objects.attribute

and valdef =
  VALDEF of
    vardecl               (* the type and name of the declared identifier *)
    * expr                (* the initial value *)
    * is_const            (* CONST if constant, VARIABLE otherwise *)
    * Objects.attribute
  | UNINITDEF of          (* for temporaries introduced by the compiler *)
    vardecl               (* the type and name of the declared identifier *)
    * Objects.attribute         (* can't be constant *)
  | SYSDEF of
    vardecl               (* the type and name of the declared identifier *)
    * is_const            (* CONST if constant, VARIABLE otherwise *)
    * Objects.attribute
  | DUMMYDEF of           (* used by parser, when type not known (a const) *)
    Objects.identifier          (* declared identifier *)
    * expr                (* inital value *)
    * Objects.attribute

  and is_const = CONST | VARIABLE | DEFCONST

and state =
  QUEUE of
    Class_state.classname           (* the class of the state *)
    * Class_state.queue_order       (* FIFO, or SORTED *)
    * Objects.identifier        (* the name of the state *)
    * Class_state.visibility       (* the visibility of the queue *)
    * Objects.attribute
  | PROCESS of
    Class_state.classname           (* the class of the state *)
    * Objects.identifier        (* the name of the state *)
    * prev                      (* the name of the previous value of the state,
			   if any *)
    * Class_state.visibility       (* the visibility of the process variable *)
    * Objects.attribute

and prev =
    NO_PREV | RTS_PREV of Objects.identifier | BOSSA_PREV of Objects.identifier

and criteria =
  CRIT_ID of key * crit_order * Objects.identifier * Objects.attribute
  | CRIT_TEST of key * crit_order * expr * expr * expr * Objects.attribute

and key = KEY of int (* lower limit *) * int (* upper limit *) | NOKEY

and crit_order = HIGHEST | LOWEST

and admit =
  ADMIT of
    valdef list             (* encapsulated variables *)
    * vardecl list          (* criteria parameters - should be read only *)
    * stmt list             (*criteria, initly all returns, to allow inlining*)
    * (vardecl * stmt) option (* attach/new code *)
    * (vardecl * stmt) option (* detach/end code *)
    * Objects.attribute

and trace =
    NOTRACE
  | PRETRACE of int * trace_info list * Objects.attribute
  | TRACE of int * event_name list * classified_exp list * expr option *
	Objects.attribute

  and trace_info =
    TRACEEVENTS of event_name list
  | TRACEEXPS of classified_exp list
  | TRACETEST of expr

  and classified_exp =
    TRACEVAR of Objects.identifier * Objects.identifier * Objects.attribute
  | TRACEFIELD of Objects.identifier * Objects.attribute
  | TRACEQUEUE of Objects.identifier * Objects.identifier * Objects.attribute

and event_decl =
  EVENT of
    event_name            (* the name of the event *)
    * event_name list     (* predecessors for which this event is specialized*)
    * vardecl             (* the event parameters (always the same) *)
    * stmt                (* the body of the event handler *)
    * Objects.attribute

  and event_name =
    EVENT_NAME of
      Objects.identifier list   (* hierarchical event names *)
      * bool                    (* whether the name ends with * *)
      * Objects.attribute

and aspect_info =
    INITASPECTS of aspect_decl list
  | ASPECTINFO of
      ((Class_state.state_info * Class_state.state_info) *
	(aspect_decl list (* before *) * aspect_decl list (* after *))) list
	
and aspect_decl =
  ASPECT of
    Objects.identifier list    (* initial states/classes *)
    * Class_state.state_info list
    * Objects.identifier list  (* final states/classes *)
    * Class_state.state_info list
    * vardecl                  (* the event parameters (always the same) *)
    * position
    * stmt                     (* the body of the event handler *)
    * (Objects.identifier * Objects.attribute) list  (* free variables *)
    * Objects.attribute

and position = ON | BEFORE | AFTER

and fundef =
  FUNDEF of
    Objects.typ                 (* the return type of the function *)
    * Objects.identifier        (* the name of the function *)
    * vardecl list        (* the parameter list of the function *)
    * stmt                (* the body of the function *)
    * inl                 (* true if the function should be inlined *)
    * Objects.attribute

    and inl = INLINE | NO_INLINE

(* ------------------- Statements and expressions ----------------- *)

and stmt =
  IF of
    expr                  (* test expression *)
    * stmt                (* then branch *)
    * stmt                (* else branch *)
    * Objects.attribute
  | FOR of
    Objects.identifier          (* index variable *)
    * Objects.identifier list option (* state or class name *)
    * Class_state.state_info list   (* information about possible states *)
    * direction           (* order in which elements are chosen *)    
    * stmt                (* loop body *)
    * Objects.attribute
  | FOR_WRAPPER of
    Objects.identifier        (* break label *)
    * stmt list         (* list of for loops *)
    * Objects.attribute
  | SWITCH of
    expr                  (* test expression *)
    * seq_case list       (* list of cases *)
    * stmt option         (* default *)
    * Objects.attribute
  | SEQ of
    valdef list           (* list of local variables *)
    * stmt list           (* list of statements *)
    * Objects.attribute
  | RETURN of
    expr option           (* returned expression, if any *)
    * Objects.attribute
  | MOVE of
    expr                        (* moved process *)
    * Objects.identifier        (* destination state *)
    * Class_state.state_info option (* info about state of source, if known *)
    * Class_state.state_info list   (* information about the source collected
				       in the verifier *)
    * Class_state.state_info option (* info about state of target, if known *)
    * bool                (* true if move from a state to itself allowed *)
    * state_end           (* head, tail, or anywhere *)
    * Objects.attribute
  | MOVEFWD of
    expr                  (* moved process *)
    * (Class_state.state_info * Class_state.state_info) list
                          (* information about the sources and dests collected
			     in the verifier *)
    * state_end           (* head, tail, or anywhere *)
    * Objects.attribute
  | SAFEMOVEFWD of        (* if there might be no next scheduler, not in lang*)
    expr                  (* moved process *)
    * (Class_state.state_info * Class_state.state_info) list
                          (* information about the sources and dests collected
			     in the verifier *)
    * stmt                (* default action *)
    * state_end           (* head, tail, or anywhere *)
    * Objects.attribute
  | ASSIGN of
    expr                  (* assigned location *)
    * expr                (* assigned value *)
    * bool                (* can the assignment affect a field of something
			     on a sorted queue *)
    * Objects.attribute
  | DEFER of
    Objects.attribute
  | PRIMSTMT of           (* not in source language *)
    expr                  (* called function *)
    * expr list           (* arguments *)
    * Objects.attribute
  | ERROR of
    string                (* the error message *)
    * Objects.attribute
  | BREAK of              (* break *)
    Objects.attribute
  | ASSERT of             (* not in source language *)
    expr                  (* expression asserted to be true *)
    * Objects.attribute

  and direction = INC | DEC | RANDOM

  and state_end = HEAD | TAIL | EITHER

  and seq_case =
    SEQ_CASE of
      Objects.identifier list   (* list of matching state names *)
      * Class_state.state_info list (* information about all possible states *)
      * stmt              (* case statement *)
      * Objects.attribute

and expr =
  INT        of int * Objects.attribute
  | VAR      of Objects.identifier * Objects.attribute
  | FIELD    of expr * Objects.identifier * Objects.attribute
  | BOOL     of bool * Objects.attribute
  | UNARY    of uop * expr * Objects.attribute
  | BINARY   of bop * expr * expr * Objects.attribute
  | PBINARY  of bop * expr * Class_state.state_info list
	        * expr * Class_state.state_info list
	        * Objects.attribute
  | INDR     of expr * Objects.attribute
  | SELECT   of Class_state.state_info option * Objects.attribute
  | EMPTY    of Objects.identifier * Class_state.state_info list *
	        Objects.testval * bool * Objects.attribute
  | PRIM     of expr * expr list * Objects.attribute
  | SCHEDCHILD of expr * Vobjects.vprocname list * Objects.attribute
  | SRCONSCHED of Objects.attribute
  | ALIVE    of Objects.testval * bool (*insrc*) * Objects.attribute
  | AREF     of expr * expr * Objects.attribute (* only in output *)
  | IN       of expr * Objects.identifier * Class_state.state_info list *
	        Objects.testval * bool * Objects.attribute
  | MOVEFWDEXP of
    expr                  (* moved process *)
    * state_end
    * Objects.attribute

  and uop = NOT | COMPLEMENT
  and bop = PLUS | MINUS | TIMES | DIV | MOD | AND | OR | BITAND | BITOR
    | EQ | NEQ | LT | GT | LEQ | GEQ | LSHIFT | RSHIFT

(* --------------------- Constructor functions -------------------- *)

val mkSCHEDULER : 
    string * bool * valdef list * typedef list * vardecl list *
    pre_valdef list * state list *
    criteria list * admit option * trace * event_decl list *
    fundef list (* interface functions *) *
    fundef list (* functions *) * int -> scheduler
	
val mkENUMDEF : string * string list * int -> typedef
val mkRANGEDEF : string * expr * expr * int -> typedef
val mkTIMERDEF : string * vardecl list * int -> typedef

val mkFUNDECL : Objects.typ * string * Objects.typ list * int -> fundecl

val mkVARDECL : Objects.typ * string * int -> vardecl
val mkSYSVARDECL : Objects.typ * string * int -> vardecl

val mkCONSTDEF : vardecl * expr * int -> valdef
val mkVALDEF : vardecl * expr * int -> valdef
val mkSYSDEF : vardecl * int -> valdef

val mkQUEUE :
    Class_state.classname * Class_state.queue_order * string *
    Class_state.visibility * int -> state
val mkPROCESS :
    Class_state.classname * string * Class_state.visibility * int -> state
val mkPREVPROCESS :
    Class_state.classname * string * 
    Class_state.visibility * string * int -> state

val mkCRIT_ID   : key * crit_order * string * int -> criteria
val mkCRIT_TEST : key * crit_order * expr * expr * expr * int -> criteria

val mkADMIT : valdef list * vardecl list * expr * (vardecl * stmt) option *
    (vardecl * stmt) option * int -> admit

val mkTRACE     : int * trace_info list * int -> trace
val mkTRACEEXP  : string list * int -> classified_exp list

val mkEVENT     : event_name * vardecl * stmt * int -> event_decl

val mkEVENT_NAME : string list * bool * int -> event_name

val mkFUNDEF : Objects.typ * string * vardecl list * stmt * int -> fundef

val mkIF         : expr * stmt * stmt * int -> stmt
val mkFOR        : string * string list * stmt * int -> stmt
val mkFORINC     : string * string * stmt * int -> stmt
val mkFORDEC     : string * string * stmt * int -> stmt
val mkFOR_ANY    : string * stmt * int -> stmt
val mkSWITCH     : expr * seq_case list * int -> stmt
val mkSEQ        : valdef list * stmt list * int -> stmt
val mkRETURN     : expr * int -> stmt
val mkRETURNVOID : int -> stmt
val mkMOVE       : expr * string * int -> stmt
val mkMOVEHD     : expr * string * int -> stmt
val mkMOVETL     : expr * string * int -> stmt
val mkMOVEFWD    : expr * int -> stmt
val mkMOVEHDFWD  : expr * int -> stmt
val mkMOVETLFWD  : expr * int -> stmt
val mkASSIGN     : expr * expr * int -> stmt
val mkDEFER      : int -> stmt
val mkPRIMSTMT   : string * expr list * int -> stmt
val mkERROR      : string * int -> stmt
val mkBREAK      : int -> stmt

val mkSEQ_CASE : string list * stmt * int -> seq_case

val mkINT        : int * int -> expr
val mkVAR        : string * int -> expr
val mkFIELD      : expr * string * int -> expr
val mkBOOL       : bool * int -> expr
val mkUNARY      : uop * expr * int -> expr
val mkBINARY     : bop * expr * expr * int -> expr
val mkINDR       : expr * int -> expr
val mkSELECT     : int -> expr
val mkEMPTY      : string * int -> expr
val mkEMPTYCLASS : Class_state.classname * int -> expr
val mkPRIM       : string * expr list * int -> expr
val mkSCHEDCHILD : expr * int -> expr
val mkSRCONSCHED : int -> expr
val mkALIVE      : int -> expr
val mkIN         : expr * string * int -> expr
val mkINCLASS    : expr * Class_state.classname * int -> expr

val mkCOMPOUND_ASSIGN : expr * bop * expr * int -> stmt

val mk_test_index : unit -> int

(* ------------------------- Useful syntax ------------------------ *)

val mkBLOCK : Objects.attribute -> stmt list -> stmt
val mkDBLOCK : Objects.attribute -> valdef list * stmt list -> stmt
val addBLOCK : Objects.attribute -> stmt -> stmt -> stmt
val postBLOCK : Objects.attribute -> stmt -> stmt list -> stmt
val add_decl_block : valdef -> stmt -> stmt

(* ----------------------- Accessor functions --------------------- *)

val get_exp_attr : expr -> Objects.attribute
val get_stm_attr : stmt -> Objects.attribute

(* ---------------------------- Comments -------------------------- *)

val add_comment : string -> expr -> expr

(* ----------------------- Printing functions --------------------- *)

val event_name2c : event_name -> string
val event_name2c_ : event_name -> string
val event_name2c1 : event_name -> string option
val event_name2c2 : event_name -> string option
val event_name2c3 : event_name -> string option
val event_name_equal : event_name -> event_name -> bool
val event_names_equal : event_name list -> event_name list -> bool
val event_name_prefix : event_name -> event_name -> bool
val event_name2partial_mask : int -> event_name -> string
val event_name2mask : event_name -> string
val bop2c : bop -> string
val equal_except_attributes : stmt -> stmt -> bool

(* --------------------- Special process names --------------------- *)

val is_src_or_tgt : expr -> Objects.identifier ->
  (Vobjects.vprocname * Class_state.state_info list) list *
    Class_state.state_info list
  -> Vobjects.vproc

(* ------------------ Naming nontrivial expressions ---------------- *)

val name_expr : expr -> Objects.attribute -> (valdef list * expr)
val pick_name_expr :
    string -> expr -> Objects.attribute -> (valdef list * expr)

(* ----------------------------- Timers ---------------------------- *)

val collect_timer_names :
    vardecl list -> valdef list -> Objects.identifier list

(* ------------------------- Free variables ------------------------ *)

val make_class_env : (Objects.identifier * Class_state.state_info) list ->
  (Objects.identifier * Class_state.state_info list) list

val collect_exp_all : Objects.identifier list (* env *) ->
      expr ->
      (Objects.identifier * Objects.attribute) list *
      (Objects.identifier * Objects.attribute) list *
      (Class_state.state_info * Objects.attribute) list

val collect_stm_all : Objects.identifier list (* env *) ->
      stmt ->
      (Objects.identifier * Objects.attribute) list *
      (Objects.identifier * Objects.attribute) list *
      (Class_state.state_info * Objects.attribute) list *
      (Objects.identifier * Objects.attribute) list *
      (Objects.identifier * Objects.attribute) list

(* both read and written variables, no fields *)
val collect_stm_vars : Objects.identifier list (* env *) ->
  stmt -> (Objects.identifier * Objects.attribute) list

(* both read and written variables, no fields *)
val collect_exp_vars : Objects.identifier list (* env *) ->
  expr -> (Objects.identifier * Objects.attribute) list

val params2ids : vardecl list -> Objects.identifier list
val valdefs2ids : valdef list -> Objects.identifier list
val no_dup_params : vardecl list -> string -> unit
val no_dup_valdefs : valdef list -> unit

(* ---------------------- Convert exp to string --------------------- *)

val exp2c : expr -> string
