(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

module B = Objects

(* -------------------- Miscellaneous operations -------------------- *)

exception LookupErrException

let rec lookup x = function
    [] -> raise LookupErrException
  | ((y,v)::r) -> if x = y then v else lookup x r

let rec lookup_id x = function
    [] -> raise LookupErrException
  | ((y,v)::r) -> if B.ideq(x,y) then v else lookup_id x r

(* ----------------------------- Classes ---------------------------- *)

type classname = RUNNING | READY | BLOCKED | TERMINATED | NOWHERE
  | POLICY_BLOCKED | KERNEL_BLOCKED | KERNEL_POLICY_BLOCKED

let all_classes in_event_types = 
  match (!B.policy_blocked_types,!B.schedtype) with
    (true,B.PROCSCHED) -> [RUNNING; READY; POLICY_BLOCKED; KERNEL_BLOCKED;
			    KERNEL_POLICY_BLOCKED; TERMINATED]
  | (false,B.PROCSCHED) -> [RUNNING; READY; BLOCKED; TERMINATED]
  | (_,B.VIRTSCHED) ->
      if in_event_types
      then [RUNNING; READY; BLOCKED; TERMINATED]
      else [RUNNING; READY; BLOCKED]

let all_classes_nowhere in_event_types = 
  NOWHERE :: (all_classes in_event_types)

(* no more class automaton
let class_automaton = [(RUNNING,  [BLOCKED; READY; TERMINATED]);
			(READY,   [READY; BLOCKED; TERMINATED]);
			(BLOCKED, [READY; TERMINATED]);
			(NOWHERE, [READY; BLOCKED])]

let vclass_automaton = [(RUNNING, [BLOCKED; READY]);
			(READY, [READY; RUNNING]);
			(BLOCKED, [READY]);
			(NOWHERE, [READY; BLOCKED])]
*)

(* for communicating with the user *)
let class2c = function
    RUNNING -> "RUNNING"
  | READY -> "READY"
  | BLOCKED -> B.check_blocked "BLOCKED"; "BLOCKED"
  | POLICY_BLOCKED -> B.check_policy_blocked "POLICY_BLOCKED"; "POLICY_BLOCKED"
  | KERNEL_BLOCKED ->
      B.check_policy_blocked "KERNEL_BLOCKED";
      (match !B.ps_using_blocked with
	B.IS_USING_BLOCKED -> "BLOCKED"
      |	_ -> "KERNEL_BLOCKED")
  | KERNEL_POLICY_BLOCKED ->
      B.check_policy_blocked "KERNEL_POLICY_BLOCKED"; "KERNEL_POLICY_BLOCKED"
  | TERMINATED -> "TERMINATED"
  | NOWHERE -> "NOWHERE"

(* no error checking, for error messages *)
let safe_class2c = function
    RUNNING -> "RUNNING"
  | READY -> "READY"
  | BLOCKED -> "BLOCKED"
  | POLICY_BLOCKED -> "POLICY_BLOCKED"
  | KERNEL_BLOCKED -> "KERNEL_BLOCKED"
  | KERNEL_POLICY_BLOCKED -> "KERNEL_POLICY_BLOCKED"
  | TERMINATED -> "TERMINATED"
  | NOWHERE -> "NOWHERE"

(* kernel classes, for printing the sched state *)
let classic_class2c = function
    RUNNING -> "RUNNING"
  | READY -> "READY"
  | BLOCKED -> "BLOCKED"
  | POLICY_BLOCKED -> "BLOCKED"
  | KERNEL_BLOCKED -> "BLOCKED"
  | KERNEL_POLICY_BLOCKED -> "BLOCKED"
  | TERMINATED -> "TERMINATED"
  | NOWHERE -> "NOWHERE"

(* string for internal use, eg for making an identifier *)
let class2idc = function
    RUNNING -> "RUNNING"
  | READY -> "READY"
  | BLOCKED -> B.check_blocked "BLOCKED"; "BLOCKED"
  | POLICY_BLOCKED ->
      B.check_policy_blocked "POLICY_BLOCKED";
      "POLICY_BLOCKED"
  | KERNEL_BLOCKED ->
      B.check_policy_blocked "KERNEL_BLOCKED";
      "KERNEL_BLOCKED"
  | KERNEL_POLICY_BLOCKED ->
      B.check_policy_blocked "KERNEL_POLICY_BLOCKED";
      "KERNEL_POLICY_BLOCKED"
  | TERMINATED -> "TERMINATED"
  | NOWHERE -> "NOWHERE"

let class2state = function
    RUNNING -> "RUNNING"
  | READY -> "READY"
  | BLOCKED -> "BLOCKED"
  | _ -> raise (Error.Error("class does not correspond to a scheduler state"))

let id2class = function
    B.ID("RUNNING",_) -> RUNNING
  | B.ID("READY",_) -> READY
  | B.ID("BLOCKED",_) ->
      B.check_blocked "BLOCKED" ; BLOCKED
  | B.ID("POLICY_BLOCKED",_) ->
      B.check_policy_blocked "POLICY_BLOCKED"; POLICY_BLOCKED
  | B.ID("KERNEL_BLOCKED",_) ->
      B.check_policy_blocked "KERNEL_BLOCKED"; KERNEL_BLOCKED
  | B.ID("KERNEL_POLICY_BLOCKED",_) ->
      B.check_policy_blocked "KERNEL_POLICY_BLOCKED"; KERNEL_POLICY_BLOCKED
  | B.ID("TERMINATED",_) -> TERMINATED
  | B.ID("NOWHERE",_) -> NOWHERE
  | id -> raise (Error.Error(Printf.sprintf "identifier %s not a class name"
			       (B.id2c id)))

(* RUNNING > READY > BLOCKED > TERMINATED > NOWHERE *)

(* A total order, with POLICY_BLOCKED < KERNEL_BLOCKED.  This should just be
used for consistency between two lists, not as having some semantics *)
let less_than = function
    (NOWHERE,_) -> false
  | (_,NOWHERE) -> true
  | (RUNNING,_) -> false
  | (READY,RUNNING) -> true
  | (READY,_) -> false
  | (BLOCKED,RUNNING) -> B.check_blocked "BLOCKED" ; true
  | (BLOCKED,READY) -> B.check_blocked "BLOCKED" ; true
  | (BLOCKED,_) -> B.check_blocked "BLOCKED" ; false
  | (POLICY_BLOCKED,RUNNING) -> B.check_policy_blocked "POLICY_BLOCKED" ; true
  | (POLICY_BLOCKED,READY) -> B.check_policy_blocked "POLICY_BLOCKED" ; true
  | (POLICY_BLOCKED,_) -> B.check_policy_blocked "POLICY_BLOCKED" ; false
  | (KERNEL_BLOCKED,RUNNING) -> B.check_policy_blocked "KERNEL_BLOCKED" ; true
  | (KERNEL_BLOCKED,READY) -> B.check_policy_blocked "KERNEL_BLOCKED" ; true
  | (KERNEL_BLOCKED,POLICY_BLOCKED) ->
      B.check_policy_blocked "KERNEL_BLOCKED" ; true
  | (KERNEL_BLOCKED,_) -> B.check_policy_blocked "KERNEL_BLOCKED" ; false
  | (KERNEL_POLICY_BLOCKED,RUNNING) ->
      B.check_policy_blocked "KERNEL_BLOCKED" ; true
  | (KERNEL_POLICY_BLOCKED,READY) ->
      B.check_policy_blocked "KERNEL_BLOCKED" ; true
  | (KERNEL_POLICY_BLOCKED,POLICY_BLOCKED) ->
      B.check_policy_blocked "KERNEL_POLICY_BLOCKED" ; true
  | (KERNEL_POLICY_BLOCKED,KERNEL_BLOCKED) ->
      B.check_policy_blocked "KERNEL_POLICY_BLOCKED" ; true
  | (KERNEL_POLICY_BLOCKED,_) ->
      B.check_policy_blocked "KERNEL_BLOCKED" ; false
  | (TERMINATED,TERMINATED) -> false
  | (TERMINATED,_) -> true

(* ----------------------- State information ------------------------ *)

type state_info =
    STATE of B.identifier * classname * state_type * visibility
and state_type  = PROC | QUEUE of queue_order
and queue_order = FIFO | SELECT of sort_default | EMPTY
and sort_default = LIFODEFAULT | FIFODEFAULT | NODEFAULT
and visibility = PUBLIC | PRIVATE
and share_state = SHARED | UNSHARED

let state2c (STATE(id,_,_,_)) = B.id2c id
let state2class (STATE(_,cls,_,_)) = cls

(* ----------------------------- States ----------------------------- *)

(* all states except NOWHERE *)
let state_env = (ref [] : (B.identifier * state_info) list ref)
let selected = (ref None : state_info option ref)
let selected_size = ref 0 (* number of key array elements *)

let concretize_class cls =
  let res =
    Aux.option_filter
      (function (_,(STATE(_,state_class,_,_) as x)) ->
	if cls = state_class then Some x else None)
      (!state_env) in
  if res = [] && not (cls = POLICY_BLOCKED || cls = KERNEL_POLICY_BLOCKED)
  then
    raise (Error.Error("cannot refer to an empty class "^(safe_class2c(cls))))
  else res
  

(* id should be either a state or a class *)
let concretize id =
  try [lookup_id id (!state_env)]
  with LookupErrException -> concretize_class (id2class id)

let all_states _ =
  Aux.option_filter
    (function
	(_,STATE(_,NOWHERE,_,_)) -> None
      | (_,x) -> Some x)
    (!state_env)

(* ------------------------- States and classes --------------------- *)

let valid_sources dstid dst =
(* no more class automaton *)
  List.map (function (id,x) -> x) (!state_env)
(*
  List.fold_left
    (function prev ->
      function (src,dstlst) ->
	if Aux.member dst dstlst
	then (List.filter
		 (function STATE(id,_,_) -> not(id = dstid))
		 (concretize_class src))
	      @ prev
	else prev)
    []
    (B.sched_select class_automaton vclass_automaton)
*)
