(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* On a particular state transition, if two aspects apply, if one aspect A
reads any variables written by the other B, then B cannot read any
variables written by A, and B must come before A.  This constraint is
independent of whether before or after is specified, and thus if before or
after is specified, then it cannot be that B is specified as after if A is
specified as before.
Aspects cannot cause state transitions. *)

(* If an aspect refers to the name of a state relevant to the transition,
the aspect must be specified as before or after, indicating that it is put
before or after the state transition, respectively.  Otherwise, the aspect
need not be specified as before or after and is put at some position
selected by the compiler. *)

(* An aspect cannot make an assignment that would affect the modified state
change operation, but I don't think that this is possible, because an
aspect can only refer to fields and global variables (besides its own local
variables, but these cannot be the source of side effects), and fields and
global variables cannot contain processes, which are the only things
affected by a state change operation.  Anyway, the aspect is executed after
the process involved in the state change operation has been determined, so
the possibility of conflict would only be in how the programmer might think
about the state change operation, and not in how it is actually carried
out. *)

(* Aspects affect all state transitions in all modules, but may only modify
the variables and fields declared in their own module. *)

module B = Objects
module CS = Class_state

let error ln str =
  Error.update_error();
  Printf.printf "%s: %s \n" (B.loc2c ln) str

(* ---------------- Check for no transition or super ---------------- *)

let rec no_state_transition = function
    Ast.IF(_,st1,st2,ln) -> no_state_transition st1; no_state_transition st2
  | Ast.FOR(_,_,_,_,stmt,ln) -> no_state_transition stmt
  | Ast.FOR_WRAPPER(_,stmts,ln) -> List.iter no_state_transition stmts
  | Ast.SWITCH(_,lsc,st2,ln) -> 
      List.iter (function Ast.SEQ_CASE(_,_,st1,_) -> no_state_transition st1)
	lsc;
      (match st2 with None -> ()
      | Some st -> no_state_transition st)
  | Ast.SEQ(_,lstmt,ln) -> List.iter no_state_transition lstmt
  | Ast.MOVE(_,_,_,_,_,_,_,ln) -> error ln "move not allowed in aspects"
  | Ast.MOVEFWD(_,_,_,ln) -> error ln "move not allowed in aspects"
  | Ast.SAFEMOVEFWD(_,_,stm,_,ln) ->
      error ln "move not allowed in aspects";
      no_state_transition stm
  | _ -> ()

let do_no_state_transition aspects =
  List.iter
    (function Ast.ASPECT(src,src_infos,dst,dst_infos,param,pos,body,_,a) ->
      no_state_transition body)
    aspects

(* ------ Check read and written variables, fields, and states ------ *)

exception Checked

type aspect_info =
    CS.state_info list * CS.state_info list * Ast.position *
      (B.identifier * B.attribute) list *
      (B.identifier * B.attribute) list *
      (Class_state.state_info * B.attribute) list *
      (B.identifier * B.attribute) list *
      (B.identifier * B.attribute) list * Ast.aspect_decl

let grouped =
  (Hashtbl.create(500) :
     (CS.state_info * CS.state_info, aspect_info list ref) Hashtbl.t)

type graph_node =
    Node of aspect_info * graph_node list ref (* parents *) *
	graph_node list ref (* children *)

let overlap l1 l2 =
  let drop_loc l = List.map (function (id,_) -> id) l in
  let l2 = drop_loc l2 in
  List.exists (function x -> List.mem x l2) (drop_loc l1)

let rec build_graph prev = function
    [] -> prev
  | ((Node((_,_,pos,rd_var,rd_fld,_,wr_var,wr_fld,_),parents,children) as n)
     ::rest) ->
      (* if this is an after, make a link back to everyone who is a before *)
      if pos = Ast.AFTER
      then
	List.iter
	  (function 
	      Node((_,_,Ast.BEFORE,_,_,_,_,_,_),p,c) as x ->
		if not(List.memq x !parents)
		then (parents := x :: !parents; c := n :: !c)
	    | _ -> ())
	  prev;
      (* if this is a before, make a link forward to everyone who is an after*)
      if pos = Ast.BEFORE
      then
	List.iter
	  (function 
	      Node((_,_,Ast.AFTER,_,_,_,_,_,_),p,c) as x ->
		if not(List.memq x !children)
		then (children := x :: !children; p := n :: !p)
	    | _ -> ())
	  prev;
      (* make a link back to everyone who writes the variables we read *)
      let sel_var (_,_,_,_,_,_,v,_,_) = v in
      let sel_field (_,_,_,_,_,_,_,f,_) = f in
      let find_writes reads sel =
	List.iter
	  (function 
	      Node(info,p,c) as x ->
		if overlap reads (sel info) && not(List.memq x !parents)
		then (parents := x :: !parents; c := n :: !c))
	  prev in
      find_writes rd_var sel_var;
      find_writes rd_fld sel_field;
      (* make a link forward to everyone who reads the variables we write *)
      let sel_var (_,_,_,v,_,_,_,_,_) = v in
      let sel_field (_,_,_,_,f,_,_,_,_) = f in
      let find_reads writes sel =
	List.iter
	  (function 
	      Node(info,p,c) as x ->
		if overlap writes (sel info) && not(List.memq x !children)
		then (children := x :: !children; p := n :: !p))
	  prev in
      find_reads wr_var sel_var;
      find_reads wr_fld sel_field;
      build_graph (n::prev) rest

exception Cycle of graph_node list

let asp_attr (Ast.ASPECT(_,src_infos,_,dst_infos,param,pos,body,_,a)) = a

let detect_cycles graph =
  let rec loop seen (Node(i,p,c) as n) =
    if List.memq n seen
    then raise (Cycle seen) (* we have a cycle *)
    else
      n ::
      (List.concat (List.map (function child -> loop (n::seen) child) !c)) in
  let rec loop1 = function
      [] -> () (* no cycle *)
    | (n::rest) ->
	let consumed = loop [] n in
	loop1
	  (List.filter (function n1 -> not (List.memq n1 consumed)) rest) in
  try loop1 graph
  with Cycle elements ->
    let Node((_,_,_,_,_,_,_,_,asp),p,c) = List.hd elements in
    failwith
      (Printf.sprintf
	 "cycle involving aspects in %s"
	 (String.concat ", "
	    (List.map
	       (function Node((_,_,_,_,_,_,_,_,asp),p,c) ->
		 B.loc2c (asp_attr asp))
	       elements)))

let disconnect_children (Node(info,p,c) as n) =
  List.iter
    (function (Node(_,p1,_)) ->
      p1 := List.filter (function parent -> not (parent == n)) !p1)
    !c

(* iteratively find everyone who has no parent and put them at the beginning
of the list *)
let rec make_order = function
    [] -> []
  | graph ->
      let (no_parents,with_parents) =
	List.partition (function Node(info,p,c) -> !p = []) graph in
      if no_parents = []
      then failwith "make_order: unexpected graph cycle!";
      List.iter disconnect_children no_parents;
      let sorted_noparents =
	List.sort
	  (function Node((_,_,pos1,_,_,_,_,_,asp),p,c) ->
	    function Node((_,_,pos2,_,_,_,_,_,asp),p,c) ->
	      match (pos1,pos2) with
		(Ast.BEFORE,_) -> -1 (* put it at the beginning *)
	      | (Ast.AFTER,_) -> 1 (* put it at the end *)
	      | (_,Ast.BEFORE) -> 1 (* put it at the beginning *)
	      | (_,Ast.AFTER) -> -1 (* put it at the end *)
	      | _ -> 0)
	  no_parents in
      (List.map (function Node((_,_,pos,_,_,_,_,_,asp),p,c) -> (pos,asp))
	 sorted_noparents) @
      (make_order with_parents)

(* split at the last before and the first after *)
let split_bef_aft (src,dst) l =
  let drop_pos l = List.map (function (_,asp) -> asp) l in
  let rec find_first x prev = function
    [] -> (prev,[])
  | ((pos,asp) as cur)::rest ->
      if x = pos
      then (prev,cur::rest)
      else find_first x (cur::prev) rest in
  let (prevrev,after) = find_first Ast.AFTER [] l in
  let (prev,revbef) = find_first Ast.BEFORE [] prevrev in
  let (bef,either,aft) =
    (drop_pos(List.rev revbef),drop_pos prev,drop_pos after) in
  (* if src is sorted, then put either with after, otherwise put it with
     before *)
  match src with
    CS.STATE(_,_,CS.QUEUE(CS.SELECT(_)),_) -> (bef,either@aft) (* sorted *)
  | _ -> (bef@either,aft)

let check_no_overlap (all_rw : aspect_info list) :
    ((CS.state_info * CS.state_info) *
       (Ast.aspect_decl list * Ast.aspect_decl list)) list =
  (* collect all aspects applying to a given starting and ending state *)
  let collect =
    List.iter
      (function (srcs,dsts,_,_,_,_,_,_,_) as x ->
	List.iter
	  (function src ->
	    List.iter
	      (function dst ->
		let cell =
		  try Hashtbl.find grouped (src,dst)
		  with
		    Not_found ->
		      let cell = ref ([] : aspect_info list) in
		      Hashtbl.add grouped (src,dst) cell;
		      cell in
		cell := x::!cell)
	      dsts)
	  srcs)
      all_rw in
  Hashtbl.fold
    (function transition ->
      function aspects ->
	function rest -> 
	  let nodes =
	    List.map (function a -> Node(a,ref [],ref [])) !aspects in
	  let graph = build_graph [] nodes in
	  detect_cycles graph;
	  (transition, split_bef_aft transition (make_order graph)) :: rest)
    grouped []

(* ------------- Check for the need for before or after ------------- *)

let cls2states class_env cls =
  try List.assoc cls class_env
  with Not_found -> [cls] (* actually a state *)

let check_before_after srcs dsts read_states = function
    Ast.ON ->
      begin
	let modified_states = Aux.union srcs dsts in
	try
	  List.iter
	    (function (state,ln) ->
	      if List.mem state modified_states
	      then
		begin
		  error ln
		    ("referencing a modified state requires before or "^
		     "after aspect");
		  raise Checked
		end)
	    read_states
	with Checked -> ()
      end
  | _ -> ()

(* ------------------------ Checking analyses ----------------------- *)

let idunion l1 l2 =
  let names2 = List.map (function (id,_) -> id) l2 in
  let rec loop = function
      [] -> l2
    | ((id,ln) as cur)::rest ->
	if List.mem id names2 then loop rest else cur::(loop rest) in
  loop l1

let do_checks aspects =
  let all_rw =
    List.map
      (function Ast.ASPECT(s,src_infos,d,dst_infos,param,pos,body,_,a) ->
	let Ast.VARDECL(ty,id,sys,a) = param in
	let (read_var,read_field,read_state,write_var,write_field) =
	  Ast.collect_stm_all [id] body in
	(List.sort compare src_infos,List.sort compare dst_infos,pos,
	 read_var,read_field,read_state,write_var,write_field,
	 Ast.ASPECT(s,src_infos,d,dst_infos,param,pos,body,read_var,a)))
      aspects in
  (* before/after as needed *)
  List.iter
    (function (src_infos,dst_infos,pos,_,_,read_state,_,_,_) ->
      check_before_after src_infos dst_infos read_state pos)
    all_rw;
  (* detect invalid dependencies, return sorted aspects *)
  check_no_overlap all_rw

(* --------------------------- Entry point -------------------------- *)

let process (Ast.SCHEDULER(nm,default,
			   cstdefs,enums,procdecls,fundecls,valdefs,states,
			   criteria,admit,trace,handlers,ifunctions,functions,
			     aspects,attr)) =
  match aspects with
    Ast.INITASPECTS(aspects) ->
      do_no_state_transition aspects;
      let new_aspects = Ast.ASPECTINFO(do_checks aspects) in
      Ast.SCHEDULER(nm,default,
		    cstdefs,enums,procdecls,fundecls,valdefs,states,
		    criteria,admit,trace,handlers,ifunctions,functions,
		      new_aspects,attr)
  | _ -> failwith "expected aspects"
