(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: events.ml,v 1.7 2004/07/30 13:46:09 julia Exp $ *)

module B = Objects
module VB = Vobjects
module CS = Class_state
module ET = Event_types
module ETBC = Event_types_bc (* all BLOCKED classes compressed to BLOCKED *)
(*module ETH = Event_types_hls*)

exception LookupErrException

(* -------------------------- various RTSes ------------------------- *)

type entry = CS.classname * VB.vtype * VB.modified

type type_rule =
    B.isquasi * (entry list * CS.classname list) list * entry list list

let app_events_init_automaton() =
  (*if (!B.hls)
  then ETH.app_events_init_automaton()
  else*)
    if (!B.policy_blocked_types)
    then ET.app_events_init_automaton()
    else ETBC.app_events_init_automaton()

let app_events_automaton() =
  (*if (!B.hls)
  then ETH.app_events_automaton()
  else*)
    if (!B.policy_blocked_types)
    then ET.app_events_automaton()
    else ETBC.app_events_automaton()

let system_events() =
  (*if (!B.hls)
  then ETH.system_event_specifications()
  else*)
    if (!B.policy_blocked_types)
    then ET.system_event_specifications()
    else ETBC.system_event_specifications()

let event_specifications() =
  (*if (!B.hls)
  then ETH.event_specifications()
  else*)
    if (!B.policy_blocked_types)
    then ET.event_specifications()
    else ETBC.event_specifications()

let vs_event_specifications() =
  (*if (!B.hls)
  then ETH.vs_event_specifications()
  else*)
    if (!B.policy_blocked_types)
    then ET.vs_event_specifications()
    else ETBC.vs_event_specifications()

let initial_event_specifications() =
  (*if (!B.hls)
  then ETH.initial_event_specifications()
  else*)
    if (!B.policy_blocked_types)
    then ET.initial_event_specifications()
    else ETBC.initial_event_specifications()

let interface_specifications() =
  (*if (!B.hls)
  then ETH.interface_specifications()
  else*)
    if (!B.policy_blocked_types)
    then ET.interface_specifications()
    else ETBC.interface_specifications()

let vs_interface_specifications() =
  (*if (!B.hls)
  then ETH.vs_interface_specifications()
  else*)
    if (!B.policy_blocked_types)
    then ET.vs_interface_specifications()
    else ETBC.vs_interface_specifications()

let top_level_events() =
  (*if (!B.hls)
  then ETH.top_level_events()
  else*)
    if (!B.policy_blocked_types)
    then ET.top_level_events()
    else ETBC.top_level_events()

(* --------------------- misc ---------------------- *)

let rec lookup x = function
    [] -> raise LookupErrException
  | ((y,v)::r) -> if x = y then v else lookup x r

let rec addto_env nm vl = function
    [] -> [(nm,[vl])]
  | ((n,v) as x)::r ->
      if nm = n
      then
	if Aux.member vl v
	    (* should be checked by config compiler *)
	then raise (Error.Error "repeated Qs not allowed")
	else (n,vl::v)::r
      else x :: (addto_env nm vl r)

(* ------------------------ the event types ------------------------- *)

(* far from the best way to do this... *)
type events = LINUX | BSD | ZERO
let config = ref LINUX
let set_os = function
    "Linux" -> config := LINUX
  | "BSD" -> config := BSD
  | "Zero" -> config := ZERO
  | _ -> raise (Error.Error "unknown OS")

(* ------------------------------ Misc ------------------------------ *)

(* identifiers are state names in the following *)

let init_events = (ref [] : Ast.event_name list ref)
(*
let app_events  = (ref [] : Ast.event_name list ref)
let sys_events  = (ref [] : Ast.event_name list ref)
*)
let init_automaton = (ref [] : (int * Ast.event_name * int) list ref)

let all_event_names = (ref [] : Ast.event_name list ref)

let automaton_start = 0

let drop_attach_detach aut =
  List.filter
    (function
	(src,nm,int,dst) ->
	  match Ast.event_name2c nm with
	    "attach" -> false (* not PS events *)
	  | "detach" -> false
	  | _ -> true)
    aut

let init_automaton _ =
  let aut = app_events_init_automaton() in
  match !B.schedtype with
    B.PROCSCHED -> drop_attach_detach aut
  | B.VIRTSCHED -> aut

let automaton _ =
  let aut = app_events_automaton() in
  match !B.schedtype with
    B.PROCSCHED -> drop_attach_detach aut
  | B.VIRTSCHED -> aut

(* --------------------- top-level event names ---------------------- *)

let top_level_event_names_loop row_fn entry_fn top_level_events =
  List.map
    (function row ->
      let (_,_,res1) =
	List.fold_left
	  (function (prev_events,prev_ids,res) ->
	    function Ast.EVENT_NAME(nm,s,a) as x ->
	      let extended = prev_ids @ [nm] in
	      (prev_events @ [x],
	       extended, (row_fn prev_events x extended s a) :: res))
	  ([],[],[]) row in
      let res2 = entry_fn (List.hd row) in
      (List.rev res1,res2))
    top_level_events

let top_level_event_names = ref ([] : Ast.event_name list)

let get_top_level_event_names top_level_events =
  let tlres =
    top_level_event_names_loop
      (function _ -> function _ ->
	function specd_name -> function star -> function attr ->
	Ast.EVENT_NAME(List.flatten specd_name,star,attr))
      (function entry_point -> ())
      top_level_events in
  List.flatten (List.map (function (x,_) -> x) tlres)

let flatten_specd_name specd (Ast.EVENT_NAME(nm,s,a)) =
  let name = (List.fold_left
		(function whole_name ->
		  function (Ast.EVENT_NAME(specdnm,_,_)) ->
		    whole_name @ specdnm)
		[] specd) @ nm in
  (Ast.EVENT_NAME(name,s,a))

(* filled in later, in verification, when handlers are available *)
let specialized_events =
  ref ([] : (Ast.event_name list * Ast.event_name) list)

(* -------------------------- entry point --------------------------- *)

(* sort by event name, such that the most specific names comes first *)
let init_event_env _ =
  init_events := initial_event_specifications();
(* not currently used
  app_events := application_event_specifications();
  sys_events := system_event_specifications();
*)
  all_event_names :=
    (if !B.schedtype = B.PROCSCHED
    then List.map (function (nm,_) -> nm) (event_specifications())
    else List.map (function (nm,_) -> nm) (vs_event_specifications()));
  top_level_event_names :=
    get_top_level_event_names (top_level_events())
